#ifdef __KERNEL__
  #ifndef __RTC_H__
  #define __RTC_H__

#include <linux/config.h>

#ifdef CONFIG_SANDPOINT_I2C_RTC_PCF8563
   /* I2C bus slave registers. */
#ifdef M_NH230
//ansel 20050325 we need no transformation here, I2C will handle this
#  define RTC_I2C_READ		0x51
#  define RTC_I2C_WRITE		0x51
#else
#  define RTC_I2C_READ		0xa3
#  define RTC_I2C_WRITE		0xa2
#endif

   /* Phillips PCF8563 registers. */
#  define RTC_CONTROL1		0x00	/* Control/Status register 1. */
#  define RTC_CONTROL2		0x01	/* Control/Status register 2. */
#  define RTC_CLOCKOUT_FREQ	0x0d	/* CLKOUT frequency. */
#  define RTC_TIMER_CONTROL	0x0e	/* Timer control. */
#  define RTC_TIMER_CNTDOWN	0x0f	/* Timer countdown. */

   /* BCD encoded clock registers. */
#  define RTC_SECONDS_REG		0x02
#  define RTC_MINUTES_REG		0x03
#  define RTC_HOURS_REG			0x04
#  define RTC_DAY_OF_MONTH_REG		0x05
#  define RTC_WEEKDAY_REG		0x06	/* Not coded in BCD! */
#  define RTC_MONTH_REG			0x07
#  define RTC_YEAR_REG			0x08
#  define RTC_MINUTE_ALARM_REG		0x09
#  define RTC_HOUR_ALARM_REG		0x0a
#  define RTC_DAY_ALARM_REG		0x0b
#  define RTC_WEEKDAY_REG_ALARM_REG	0x0c	/* Not coded in BCD! */

/* Control/Status 2 bits */
#  define RTC_CTRL2_TIE		0x01	/* Timer Interrupt Enable */
#  define RTC_CTRL2_AIE		0x02	/* Alarm Interrupt Enable */
#  define RTC_CTRL2_TF		0x04	/* Timer Flag */
#  define RTC_CTRL2_AF		0x08	/* Alarm Flag */
#  define RTC_CTRL2_TP		0x10	/* Timer Pulse */

/* Timer Control bits */
#  define RTC_TC_TD0		0x01
#  define RTC_TC_TD1		0x02
#  define RTC_TC_TE		0x80	/* Timer Enable */

/* Timer source clock frequency */
#  define RTC_TC_FREQ_MASK	(RTC_TC_TD0 | RTC_TC_TD1)
#  define RTC_TC_FREQ_4096	(0)
#  define RTC_TC_FREQ_64	(RTC_TC_TD0)
#  define RTC_TC_FREQ_1		(RTC_TC_TD1)
#  define RTC_TC_FREQ_1_60	(RTC_TC_TD0 | RTC_TC_TD1)

#endif

#ifdef CONFIG_SANDPOINT_I2C_RTC_PCF8563
  extern unsigned char pcf8563_readreg(int reg);
  extern void pcf8563_writereg(int reg, unsigned char val);
  extern int pcf8563_init(void);
  #  define CMOS_READ(x) pcf8563_readreg(x)
  #  define CMOS_WRITE(val,reg) pcf8563_writereg(reg,val)
  #  define RTC_INIT() pcf8563_init()
#endif



#if 0


/* Conversions to and from the RTC BCD encoded format. */
#define BCD_TO_BIN(x) ((x)=((x)&15) + ((x)>>4)*10)
#define BIN_TO_BCD(x) ((x)=(((x)/10)<<4) + (x)%10)


/*
 * The struct used to pass data via the following ioctl. Similar to the
 * struct tm in <time.h>, but it needs to be here so that the kernel
 * source is self contained, allowing cross-compiles, etc. etc.
 */
struct rtc_time {
	int tm_sec;
	int tm_min;
	int tm_hour;
	int tm_mday;
	int tm_mon;
	int tm_year;
	int tm_wday;
	int tm_yday;
	int tm_isdst;
};


/* ioctl() calls that are permitted to the /dev/rtc interface. */
#define RTC_MAGIC 'p'
#define RTC_RD_TIME     _IOR(RTC_MAGIC, 0x09, struct rtc_time)	/* Read RTC time. */
#define RTC_SET_TIME    _IOW(RTC_MAGIC, 0x0a, struct rtc_time)	/* Set RTC time. */
#define RTC_SET_CHARGE  _IOW(RTC_MAGIC, 0x0b, int)
#define RTC_VLOW_RD     _IOR(RTC_MAGIC, 0x11, int)  /* Voltage Low detector */
#define RTC_VLOW_SET    _IO(RTC_MAGIC, 0x12)        /* Clear voltage low information */
#define RTC_MAX_IOCTL 0x12

#endif 

  #endif  // end of #ifdef {__RTC_H__}
#endif  // end of #if {__KERNEL__}


