/*
 * $Id: physmap.c,v 1.1.1.1 2004/12/23 01:05:30 cvs Exp $
 *
 * Normal mappings of chips in physical memory
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/config.h>

#ifdef CONFIG_MTD_PARTITIONS
#include <linux/mtd/partitions.h>
#endif

#define WINDOW_ADDR CONFIG_MTD_PHYSMAP_START
#define WINDOW_SIZE CONFIG_MTD_PHYSMAP_LEN
#define BUSWIDTH CONFIG_MTD_PHYSMAP_BUSWIDTH

static struct mtd_info *mymtd;

__u8 physmap_read8(struct map_info *map, unsigned long ofs)
{
	return __raw_readb(map->map_priv_1 + ofs);
}

__u16 physmap_read16(struct map_info *map, unsigned long ofs)
{
	return __raw_readw(map->map_priv_1 + ofs);
}

__u32 physmap_read32(struct map_info *map, unsigned long ofs)
{
	return __raw_readl(map->map_priv_1 + ofs);
}

void physmap_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

void physmap_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	__raw_writeb(d, map->map_priv_1 + adr);
	mb();
}

void physmap_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, map->map_priv_1 + adr);
	mb();
}

void physmap_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	__raw_writel(d, map->map_priv_1 + adr);
	mb();
}

void physmap_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio(map->map_priv_1 + to, from, len);
}

struct map_info physmap_map = {
	name: "Physically mapped flash",
	size: WINDOW_SIZE,
	buswidth: BUSWIDTH,
	read8: physmap_read8,
	read16: physmap_read16,
	read32: physmap_read32,
	copy_from: physmap_copy_from,
	write8: physmap_write8,
	write16: physmap_write16,
	write32: physmap_write32,
	copy_to: physmap_copy_to
};

#ifdef CONFIG_MTD_PARTITIONS

#ifdef CONFIG_MTD_CMDLINE_PARTS
static struct mtd_partition *mtd_parts = 0;
static int                   mtd_parts_nb = 0;
#else
static struct mtd_partition physmap_partitions[] = {
/* Put your own partition definitions here */

#ifdef M_NH230
  #define M_KB(n)  (n * 1024)          // Convert to (n * 1KB)
  
  // Please refer to the definition of flash partitions for NH-230.
  {
    name: "Boot-loader",               // It is /dev/mtd0
    size: M_KB(128),                   // 128KB
    offset: 0x300000,                  // offset in flash
    mask_flags:	MTD_WRITEABLE,         // Force Read-Only
  }, 
  {
    name: "Boot CFG",                  // It is /dev/mtd1
    size: M_KB(16),                    // 16KB
    offset: 0,
    mask_flags:	MTD_CAP_ROM,           // R/W as mask nothing
  }, 
  {
    name: "POST Result",               // It is /dev/mtd2
    size: M_KB(8),                     // 8KB
    offset: 0x4000,
    mask_flags:	MTD_WRITEABLE,         // Force Read-Only
  }, 
  {
    name: "Linux CFG",                 // It is /dev/mtd3
    size: M_KB(32),                    // 32KB
    offset: 0x8000,
    mask_flags:	MTD_CAP_ROM,
  }, 
  {
    name: "Reserved",                  // It is /dev/mtd4
    size: M_KB(8),                     // 8KB
    offset: 0x6000,
    mask_flags:	MTD_WRITEABLE,         // Force Read-Only
  }, 
  {
    name: "Linux",                     // It is /dev/mtd5
    size: M_KB(1536),                  // 1536KB
    offset: 0x10000,
    mask_flags:	MTD_CAP_ROM,
  }, 
  {
    name: "RAM disk #2",               // It is /dev/mtd6
    size: M_KB(1472),                  // 1472KB
    offset: 0x190000,
    mask_flags:	MTD_CAP_ROM,
  },
  {
    name: "RAM disk #1 (root)",        // It is /dev/mtd7
    size: M_KB(896),                   // 896KB
    offset: 0x320000,           
    mask_flags:	MTD_CAP_ROM,
  }
#else
  #if 1 //add by andyw top
	{
		name:		"Ramdisk 0x0 -> 0x6fffff",
		size:		0x700000,		
		offset:		0,
		mask_flags:	MTD_CAP_ROM,  /* rw */
	}, {
		name:		"Bootloader at 0x700000",
		size:		0x30000,
		offset:		MTDPART_OFS_APPEND,
		mask_flags:	MTD_CAP_ROM,  /* rw */
	}, {
		name:		"Kernel at 0x730000",
		size:		0xc0000,
		offset:		MTDPART_OFS_APPEND,
		mask_flags:	MTD_CAP_ROM,  /* rw */
	}, {
		name:		"Igateway Config file at 0x7f0000 ",
		size:		MTDPART_SIZ_FULL,
		offset:		MTDPART_OFS_APPEND,
		mask_flags:	MTD_CAP_ROM,  /* rw */
	}
  #endif
  #if 0 //add by andyw bottom
	{
		name:		"Igateway Config file at 0x0-0x40000 ",
		size:		0x40000,
		offset:		0,
		mask_flags:	MTD_CAP_ROM,  /* rw */
	}, {
		name:		"Ramdisk 0x40000 -> 0x700000",
		size:		0x6c0000,		
		offset:		MTDPART_OFS_APPEND,
		mask_flags:	MTD_CAP_ROM,  /* rw */
	}, {
		name:		"Bootloader at 0x700000",
		size:		0x30000,
		offset:		MTDPART_OFS_APPEND,
		mask_flags:	MTD_CAP_ROM,  /* rw */
	}, {
		name:		"Kernel at 0x730000",
		size:		MTDPART_SIZ_FULL,
		offset:		MTDPART_OFS_APPEND,
		mask_flags:	MTD_CAP_ROM,  /* rw */
	}
  #endif
  #if 0 //kernel.org
	{
		name:		"bootROM",
		size:		0x80000,
		offset:		0,
		mask_flags:	MTD_WRITEABLE,  /* force read-only */
	}, {
		name:		"zImage",
		size:		0x100000,
		offset:		MTDPART_OFS_APPEND,
		mask_flags:	MTD_WRITEABLE,  /* force read-only */
	}, {
		name:		"ramdisk.gz",
		size:		0x300000,
		offset:		MTDPART_OFS_APPEND,
		mask_flags:	MTD_WRITEABLE,  /* force read-only */
	}, {
		name:		"User FS",
		size:		MTDPART_SIZ_FULL,
		offset:		MTDPART_OFS_APPEND,
	}
  #endif
#endif  // end of #ifdef {M_NH230}
};

#define NUM_PARTITIONS	(sizeof(physmap_partitions)/sizeof(struct mtd_partition))

#endif
#endif

int __init init_physmap(void)
{
	static const char *rom_probe_types[] = { "cfi_probe", "jedec_probe", "map_rom", 0 };
	const char **type;

       	printk(KERN_NOTICE "physmap flash device: %x at %x\n", WINDOW_SIZE, WINDOW_ADDR);
	physmap_map.map_priv_1 = (unsigned long)ioremap(WINDOW_ADDR, WINDOW_SIZE);

	if (!physmap_map.map_priv_1) {
		printk("Failed to ioremap\n");
		return -EIO;
	}
	
	mymtd = 0;
	type = rom_probe_types;
	for(; !mymtd && *type; type++) {
		mymtd = do_map_probe(*type, &physmap_map);
	}
	if (mymtd) {
		mymtd->module = THIS_MODULE;

		add_mtd_device(mymtd);
#ifdef CONFIG_MTD_PARTITIONS
#ifdef CONFIG_MTD_CMDLINE_PARTS
		mtd_parts_nb = parse_cmdline_partitions(mymtd, &mtd_parts, 
							"phys");
		if (mtd_parts_nb > 0)
		{
			printk(KERN_NOTICE 
			       "Using command line partition definition\n");
			add_mtd_partitions (mymtd, mtd_parts, mtd_parts_nb);
		}
#else
		if (NUM_PARTITIONS != 0) 
		{
			printk(KERN_NOTICE 
			       "Using physmap partition definition\n");
			add_mtd_partitions (mymtd, physmap_partitions, NUM_PARTITIONS);
		}

#endif
#endif
		return 0;
	}

	iounmap((void *)physmap_map.map_priv_1);
	return -ENXIO;
}

static void __exit cleanup_physmap(void)
{
	if (mymtd) {
		del_mtd_device(mymtd);
		map_destroy(mymtd);
	}
	if (physmap_map.map_priv_1) {
		iounmap((void *)physmap_map.map_priv_1);
		physmap_map.map_priv_1 = 0;
	}
}

module_init(init_physmap);
module_exit(cleanup_physmap);


MODULE_LICENSE("GPL");
MODULE_AUTHOR("David Woodhouse <dwmw2@infradead.org>");
MODULE_DESCRIPTION("Generic configurable MTD map driver");
