#ifdef RTL8169_IOCTL_SUPPORT

#include "r8169_ioctl.h"


int rtl8169_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
	struct rtl8169_private *priv = dev->priv;
	void *ioaddr = priv->mmio_addr;

	struct ver_t *pver;
	struct link_t *plink;
	struct mii_t *pmii;

	unsigned char val8;

	switch(cmd)
	{
	case RTL8169_IOCTL_RESET: // Reset
			rtl8169_close (dev);
			rtl8169_open (dev);
			printk("%s:ioctl <RESET>\n", dev->name);
			break;
	
	case RTL8169_IOCTL_GET_VER: // Get SW, HW version
            pver = (struct ver_t *)(ifr->ifr_data);

			switch(priv->mac_version)
			{
				case RTL_GIGA_MAC_VER_E: pver->mac_version = 'E'; break;
				case RTL_GIGA_MAC_VER_D: pver->mac_version = 'D'; break;
				case RTL_GIGA_MAC_VER_B: pver->mac_version = 'B'; break;
				default: pver->mac_version = 'X'; break;
			}

			switch(priv->phy_version)
			{
				case RTL_GIGA_PHY_VER_G: pver->phy_version = 'G'; break;
				case RTL_GIGA_PHY_VER_F: pver->phy_version = 'F'; break;
				case RTL_GIGA_PHY_VER_E: pver->phy_version = 'E'; break;
				case RTL_GIGA_PHY_VER_D: pver->phy_version = 'D'; break;
				default: pver->phy_version = 'X'; break;
			}

            memcpy( pver->drv_version, RTL8169_VERSION, IFNAMSIZ-2 );

			printk("%s:ioctl <Get SW, HW version>\n", dev->name);
			break;

	case RTL8169_IOCTL_GET_LINK: // Get Link status
            plink = (struct link_t *)(ifr->ifr_data);

			val8 = RTL_R8(PHYstatus);
			if( val8 & _1000bpsF ){
				plink->speed = 1000;
				plink->duplex = 1;
			}
			else{
				plink->speed = (val8 & _100bps) ? 100 : 10;
				plink->duplex = (val8 & FullDup) ? 1 : 0;
			}
			printk("%s:ioctl <Get Link status>\n", dev->name);
			break;

	case RTL8169_IOCTL_GET_MII_REG: // Get MII_REG
			pmii = (struct mii_t *)(ifr->ifr_data);
			pmii->get_val = RTL8169_READ_GMII_REG( ioaddr, pmii->get_reg );
			printk("%s:ioctl <Get MII_REG><%d>\n", dev->name, pmii->get_reg);
			break;

	case RTL8169_IOCTL_SET_MII_REG: // Set MII_REG
			pmii = (struct mii_t *)(ifr->ifr_data);
			RTL8169_WRITE_GMII_REG( ioaddr, pmii->set_reg, pmii->set_val );
			printk("%s:ioctl <Set MII_REG><%d,%d>\n", dev->name, pmii->set_reg, pmii->set_val);
			break;

	case RTL8169_IOCTL_GET_MTU: // Get MTU
			ifr->ifr_mtu = priv->curr_mtu_size;
			printk("%s:ioctl <Get MTU><%d>\n", dev->name, ifr->ifr_mtu);
			break;

	default:
			break;
	}

	return 0;
}

#endif //endif #define RTL8169_IOCTL_SUPPORT

