//
// Subject :
//   The IOCTL routine for Realtek RTL8169S (Gigabit Ethernet PCI controller).
// 
//
//

#ifdef RTL8169_IOCTL_SUPPORT

#include "r8169_ioctl.h"

//sp/
//#define SP_R8169_IOCTL_DEBUG
#if      defined(SP_R8169_IOCTL_DEBUG)
	#ifndef	DEBUG
	//# define DEBUG
	#endif
	# define  sp_r8169_ioctlTrace(format,arg...)		printk(KERN_INFO __FILE__": "__FUNCTION__ ": " format "\n" , ## arg)
#else
	# define  sp_r8169_ioctlTrace(format, args...)
#endif

# ifdef M_NH230
int Cap10_100 = 0;
int Cap1000 = 0;

void do_nway(){
	unsigned long ioaddr = g_ioaddr;
	int i, option;
	int	val = RTL8169_READ_GMII_REG( ioaddr, PHY_AUTO_NEGO_REG );

#ifdef RTL8169_HW_FLOW_CONTROL_SUPPORT
		val |= PHY_Cap_PAUSE | PHY_Cap_ASYM_PAUSE ;
#endif //end #define RTL8169_HW_FLOW_CONTROL_SUPPORT

	RTL8169_WRITE_GMII_REG( ioaddr, PHY_AUTO_NEGO_REG, Cap10_100 | ( val&0xC1F ) );	//leave PHY_AUTO_NEGO_REG bit4:0 unchanged
	RTL8169_WRITE_GMII_REG( ioaddr, PHY_1000_CTRL_REG, Cap1000 );

	// Enable auto-negotiation and restart auto-nigotiation
	RTL8169_WRITE_GMII_REG( ioaddr, PHY_CTRL_REG, PHY_Enable_Auto_Nego | PHY_Restart_Auto_Nego );
	udelay(100);

	// wait for auto-negotiation process
	for( i = 10000; i > 0; i-- ){
		//check if auto-negotiation complete
		if( RTL8169_READ_GMII_REG(ioaddr, PHY_STAT_REG) & PHY_Auto_Neco_Comp ){
			udelay(100);
			option = RTL_R8(PHYstatus);
			if( option & _1000bpsF ){
				printk(KERN_INFO "%s: 1000Mbps Full-duplex operation.\n", g_dev->name);
			}
			else{
				printk(KERN_INFO "%s: %sMbps %s-duplex operation.\n", g_dev->name,
						(option & _100bps) ? "100" : "10", (option & FullDup) ? "Full" : "Half" );
			}
			break;
		}
		else{
			udelay(100);
		}// end of if( RTL8169_READ_GMII_REG(ioaddr, 1) & 0x20 )
	}// end for-loop to wait for auto-negotiation process
	
	option = RTL_R8(PHYstatus);
	if( option & _1000bpsF ){
		g_priv->linkstatus = _1000_Full;
	}
	else{
		if(option & _100bps){
			g_priv->linkstatus = (option & FullDup) ? _100_Full : _100_Half;
		}
        else{
			g_priv->linkstatus = (option & FullDup) ? _10_Full : _10_Half;
		}
	}
	printk("g_priv->linkstatus = 0x%02x\n", g_priv->linkstatus);
	
	
}

# endif//#ifdef M_NH230
			


int rtl8169_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
	sp_r8169_ioctlTrace("cmd=%d", cmd);
	
	struct rtl8169_private *priv = dev->priv;
#ifdef M_NH230
	unsigned long ioaddr = priv->ioaddr;
#else
	void *ioaddr = priv->mmio_addr;
#endif

#ifdef M_NH230
#else
	struct ver_t *pver;
#endif
	struct link_t *plink;
	struct mii_t *pmii;

	unsigned char val8;

	switch(cmd)
	{
	case RTL8169_IOCTL_RESET: // Reset
			rtl8169_close (dev);
			rtl8169_open (dev);
			printk("%s:ioctl <RESET>\n", dev->name);
			break;

#ifdef M_NH230
	// We do not support to pass the version of MAC and PHY to upper layers
#else
	case RTL8169_IOCTL_GET_VER: // Get SW, HW version
            pver = (struct ver_t *)(ifr->ifr_data);

			switch(priv->mac_version)
			{
				case RTL_GIGA_MAC_VER_E: pver->mac_version = 'E'; break;
				case RTL_GIGA_MAC_VER_D: pver->mac_version = 'D'; break;
				case RTL_GIGA_MAC_VER_B: pver->mac_version = 'B'; break;
				default: pver->mac_version = 'X'; break;
			}

			switch(priv->phy_version)
			{
				case RTL_GIGA_PHY_VER_G: pver->phy_version = 'G'; break;
				case RTL_GIGA_PHY_VER_F: pver->phy_version = 'F'; break;
				case RTL_GIGA_PHY_VER_E: pver->phy_version = 'E'; break;
				case RTL_GIGA_PHY_VER_D: pver->phy_version = 'D'; break;
				default: pver->phy_version = 'X'; break;
			}

            memcpy( pver->drv_version, RTL8169_VERSION, IFNAMSIZ-2 );

			printk("%s:ioctl <Get SW, HW version>\n", dev->name);
			break;
#endif  // end of #ifdef M_NH230


	case RTL8169_IOCTL_GET_LINK: // Get Link status
            plink = (struct link_t *)(ifr->ifr_data);

			val8 = RTL_R8(PHYstatus);
			if( val8 & _1000bpsF ){
				plink->speed = 1000;
				plink->duplex = 1;
			}
			else{
				plink->speed = (val8 & _100bps) ? 100 : 10;
				plink->duplex = (val8 & FullDup) ? 1 : 0;
			}
			printk("%s:ioctl <Get Link status>\n", dev->name);
#ifdef M_NH230
			printk("%s:ioctl plink->speed=%d plink->duplex=%d\n", dev->name, plink->speed, plink->duplex);
			strcpy(ifr->ifr_name, "ethx");
			
#endif//#ifdef M_NH230
			break;

	case RTL8169_IOCTL_GET_MII_REG: // Get MII_REG
			pmii = (struct mii_t *)(ifr->ifr_data);
			pmii->get_val = RTL8169_READ_GMII_REG( ioaddr, pmii->get_reg );
#ifdef M_NH230

//#include <asm/uaccess.h>

			//printk("%s:ioctl <Get MII_REG><%08x> pmii->get_val=%08x\n", dev->name, pmii->get_reg, pmii->get_val);
			//printk("%s:ioctl <Get MII_REG><%08x>\n", dev->name, pmii->get_reg);
			//printk("%s:ioctl..test1 pmii->get_reg=%d pmii->get_val=%08x\n", pmii->get_reg, pmii->get_val);
			char temp[64];
			int error ;
			//error = copy_from_user(temp, ifr->ifr_data, "5");
			//printk("%s:ioctl..test1 ifr->ifr_data=%s\n", ifr->ifr_data);
			printk("%s:ioctl..test1 ifr->ifr_data=%s\n", temp);
#else
			printk("%s:ioctl <Get MII_REG><%d>\n", dev->name, pmii->get_reg);
#endif
			break;

	case RTL8169_IOCTL_SET_MII_REG: // Set MII_REG
			pmii = (struct mii_t *)(ifr->ifr_data);
			RTL8169_WRITE_GMII_REG( ioaddr, pmii->set_reg, pmii->set_val );
#ifdef M_NH230
			printk("%s:ioctl <Set MII_REG><%08x,%08x>\n", dev->name, pmii->set_reg, pmii->set_val);
#else
			printk("%s:ioctl <Set MII_REG><%d,%d>\n", dev->name, pmii->set_reg, pmii->set_val);
#endif
			break;

	case RTL8169_IOCTL_GET_MTU: // Get MTU
			ifr->ifr_mtu = priv->curr_mtu_size;
			printk("%s:ioctl <Get MTU><%d>\n", dev->name, ifr->ifr_mtu);
			break;
			
#ifdef M_NH230
	case RTL8169_IOCTL_NWAY: 
			printk("RTL8169_IOCTL_NWAY\n");
			Cap10_100 = PHY_Cap_10_Half | PHY_Cap_10_Full | PHY_Cap_100_Half | PHY_Cap_100_Full;
			Cap1000 = PHY_Cap_1000_Full;
			
			do_nway();
			break;
	case RTL8169_IOCTL_10F: 
			printk("RTL8169_IOCTL_10F\n");
			Cap10_100 = PHY_Cap_10_Full | PHY_Cap_10_Half;
			Cap1000 = PHY_Cap_Null;
			
			do_nway();
			break;
	case RTL8169_IOCTL_100F: 
			printk("RTL8169_IOCTL_100F\n");
			Cap10_100 = PHY_Cap_100_Full | PHY_Cap_100_Half | PHY_Cap_10_Full | PHY_Cap_10_Half;
			Cap1000 = PHY_Cap_Null;
			
			do_nway();
			break;
	case RTL8169_IOCTL_1000F: 
			printk("RTL8169_IOCTL_1000F\n");
			Cap10_100 = PHY_Cap_100_Full | PHY_Cap_100_Half | PHY_Cap_10_Full | PHY_Cap_10_Half;
			Cap1000 = PHY_Cap_1000_Full;
			
			do_nway();
			break;
#endif//#ifdef M_NH230



#ifdef M_NH230
	default:
			return (-EOPNOTSUPP);    // command is not supported
#else
	default:
			break;
#endif
	}

	return 0;
}

#endif //endif #define RTL8169_IOCTL_SUPPORT

