/*
 * (C) Copyright 2000
 * Rob Taylor, Flying Pig Systems. robt@flyingpig.com.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <mpc824x.h>
#include <pci.h>

int checkboard (void)
{
	/*TODO: Check processor type */

	puts (	"Board: Sandpoint "
#ifdef CONFIG_MPC8240
		"8240"
#endif
#ifdef CONFIG_MPC8245
  #ifdef M_NH230
  		"8241"
  #else
		"8245"
  #endif  
#endif
		" Unity ##Test not implemented yet##\n");
	return 0;
}

#if 0 	/* NOT USED */
int checkflash (void)
{
	/* TODO: XXX XXX XXX */
	printf ("## Test not implemented yet ##\n");

	return (0);
}
#endif

long int initdram (int board_type)
{
	int		 i, cnt;
	volatile uchar * base= CFG_SDRAM_BASE;
	volatile ulong * addr;
	ulong		 save[32];
	ulong		 val, ret  = 0;

	for (i=0, cnt=(CFG_MAX_RAM_SIZE / sizeof(long)) >> 1; cnt > 0; cnt >>= 1) {
		addr = (volatile ulong *)base + cnt;
		save[i++] = *addr;
		*addr = ~cnt;
	}

	addr = (volatile ulong *)base;
	save[i] = *addr;
	*addr = 0;
   	
	if (*addr != 0) {
		*addr = save[i];
		goto Done;
	}

	for (cnt = 1; cnt <= CFG_MAX_RAM_SIZE / sizeof(long); cnt <<= 1) {
		addr = (volatile ulong *)base + cnt;
		val = *addr;
		*addr = save[--i];

		if (val != ~cnt) {
			ulong new_bank0_end = cnt * sizeof(long) - 1;
			ulong mear1  = mpc824x_mpc107_getreg(MEAR1);
			ulong emear1 = mpc824x_mpc107_getreg(EMEAR1);
			
			mear1 =  (mear1  & 0xFFFFff00) |
			((new_bank0_end & MICR_ADDR_MASK) >> MICR_ADDR_SHIFT);
			emear1 = (emear1 & 0xFFFFff00) |
			((new_bank0_end & MICR_ADDR_MASK) >> MICR_EADDR_SHIFT);
			//mpc824x_mpc107_setreg(MEAR1,  mear1);
			mpc824x_mpc107_setreg(EMEAR1, emear1);

			ret = cnt * sizeof(long);
			goto Done;
		}
	}

	ret = CFG_MAX_RAM_SIZE;
Done:
	return ret;
}

/*
 * Initialize PCI Devices, report devices found.
 */
#ifndef CONFIG_PCI_PNP
static struct pci_config_table pci_sandpoint_config_table[] = {
#ifdef M_NH230
        // Assign to PCI-to-Gigabit Ethernet only
        { PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, PCI_DEVICE_NUMBER_ETHERNET, PCI_ANY_ID, 
          pci_cfgfunc_config_device, { PCI_ENET0_IOADDR, 
          			       PCI_ENET0_MEMADDR, 
          			       PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER }},
#else
	{ PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, 0x0f, PCI_ANY_ID, 
	  pci_cfgfunc_config_device, { PCI_ENET0_IOADDR, 
				       PCI_ENET0_MEMADDR, 
				       PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER }},
	{ PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, PCI_ANY_ID, 0x10, PCI_ANY_ID, 
	  pci_cfgfunc_config_device, { PCI_ENET1_IOADDR, 
				       PCI_ENET1_MEMADDR, 
				       PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER }},
#endif  // end of #ifdef {M_NH230}
	{ }
};
#endif

struct pci_controller hose = {
#ifndef CONFIG_PCI_PNP
	config_table: pci_sandpoint_config_table,
#endif
};


#ifdef M_NH230_NOT_ENABLED
	// Refer to /linux-2.4.1/include/asm-ppc/nh230.h for definition of GPIO pins.
  #define GPIO_PIN_WR_PCI_RESET_DIS    0x02  // bit[1]: low active: assert PCI reset to all devices
  #define GPIO_PIN_WR_STATUS_LED_DIS   0x04  // bit[2]: low active: turn on status LED
  #define GPIO_PIN_PCI_RESET_ON    (GPIO_PIN_WR_STATUS_LED_DIS)
  
  void my_out_8(volatile u8 *addr, int val)
  {
    __asm__ __volatile__("stb%U0%X0 %1,%0; eieio" : "=m" (*addr) : "r" (val));
  }
  
  #define my_writeb(b,addr) my_out_8((volatile u8 *)(addr), (b))
#endif


void pci_init(void)
{
#ifdef M_NH230_NOT_ENABLED
	// 1) Assert PCI Reset# via GPIO pin: reset PCI peripherals to be working before access
	printf("PCI : assert Reset# to PCI peripherals.... ");
	// 1-1) Assert PCI Reset# 200ms
	my_writeb(GPIO_PIN_PCI_RESET_ON, CFG_ERCR4_RCS3_START_ADDRESS);
	printf("A1");
	udelay(400 * 1000);
	printf("A2");
	
	// 1-2) De-assert it 100ms 
	my_writeb(GPIO_PIN_WR_PCI_RESET_DIS | GPIO_PIN_WR_STATUS_LED_DIS, CFG_ERCR4_RCS3_START_ADDRESS);
	printf("A3");
	udelay(200 * 1000);
	printf("Done\n");
#endif

	pci_mpc824x_init(&hose);
}
