/*
 * (C) Copyright 2000
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <watchdog.h>
#include <command.h>
#include <cmd_nvedit.h>
#include <cmd_bootm.h>
#include <malloc.h>
#include <pci.h>
#if defined(CONFIG_BOOT_RETRY_TIME) && defined(CONFIG_RESET_TO_RETRY)
#include <cmd_boot.h>		/* for do_reset() prototype */
#endif

#ifdef CFG_HUSH_PARSER
#include <hush.h>
#endif

/************************************************************************
*  All of functions and variables of POST need are define as below 
*  By George 06/21/2005
************************************************************************/

/* Function that returns a character from the environment */
extern uchar (*env_get_char)(int);

/* Function that returns a pointer to a value from the environment */
/* (Only memory version supported / needed). */
extern uchar *env_get_addr(int);

/* Function that updates CRC of the enironment */
extern void env_crc_update (void);
extern void setenv (char *varname, char *varvalue);
extern int saveenv (void);

void run_nh230_post(void);
void do_run_nh230_post(int cnt);
void WritePOSTResultToFLASH(int flag);
void WritePOSTCntToFLASH(int LoopCnt, int Cnt);
int flash_test(int cnt);
int memory_test(int cnt);
int pci_device_scan(int cnt);
int StringToInt(char *lpsz_str);
int exp10(int x);
static int postcnt (uchar *, int);

static struct pci_device_id RTL8169S[] = {
	{0x10EC, 0x8169},            // Realtek  RTL8169S
	{}
};
static struct pci_device_id IT8212[] = {
	{0x1283, 0x8212},            // ITE IT8212 IDE
	{}
};
static struct pci_device_id IT8211[] = {
	{0x1283, 0x8211},            // ITE IT8211 IDE
	{}
};
static struct pci_device_id EHCI[] = {
	{0x17A0, 0x8084},            // USB 2.0 (EHCI controller)
	{}
};
static struct pci_device_id UHCI[] = {
	{0x17A0, 0x8083},            // USB 1.1 (UHCI controllerb
	{}
};

char ErrMsg[64];
uchar PassMsg[32];

/************************************************************************
************************************************************************/

#define MAX_DELAY_STOP_STR 32

static char * delete_char (char *buffer, char *p, int *colp, int *np, int plen);
static int parse_line (char *, char *[]);
#if defined(CONFIG_BOOTDELAY) && (CONFIG_BOOTDELAY >= 0)
static int abortboot(int);
#endif

#undef DEBUG_PARSER

char        console_buffer[CFG_CBSIZE];		/* console I/O buffer	*/

static char erase_seq[] = "\b \b";		/* erase sequence	*/
static char   tab_seq[] = "        ";		/* used to expand TABs	*/

#ifdef CONFIG_BOOT_RETRY_TIME
static uint64_t endtime = 0;  /* must be set, default is instant timeout */
static int      retry_time = -1; /* -1 so can call readline before main_loop */
#endif

#define	endtick(seconds) (get_ticks() + (uint64_t)(seconds) * get_tbclk())

#ifndef CONFIG_BOOT_RETRY_MIN
#define CONFIG_BOOT_RETRY_MIN CONFIG_BOOT_RETRY_TIME
#endif

#ifdef CONFIG_MODEM_SUPPORT
int do_mdm_init = 0;
extern void mdm_init(void); /* defined in board.c */
#endif


#ifdef M_NH230
  #define ETH_ALEN    6     // Size of 6-byte MAC address in bytes
  #define CHAR_COLON  ':'   // Ethernet address string (xx:xx:xx:xx:xx:xx) separated by 5 colons
  #define M_FAIL      ((char)(-1))  
  
  //
  // Subject :
  //   Convert the given ASCII character into a signed hex-decimal.
  //   e.g.    'A' --> 0x0A,  '9' -->0x09
  //
  // Input :
  //   c_ascii  -  ascii code of character to convert (e.g. 'A')
  //
  // Output :
  //   -1       -  unable to convert
  //   others   -  hex-decimal signed integer (e.g. 0x0A)
  //
  char if_ascii_2_hex (char c_ascii)
  {
    if ('0' <= c_ascii && '9' >= c_ascii)     // '0' .. '9' ?
      return (c_ascii - '0');

    if ('a' <= c_ascii && 'z' >= c_ascii)     // 'a' .. 'z' ?
      return (c_ascii - 'a' + 10);

    if ('A' <= c_ascii && 'Z' >= c_ascii)     // 'A' .. 'Z' ?
      return (c_ascii - 'A' + 10);

    return (M_FAIL);                              // unable to convert
  }


  //
  // Subject :
  //   Convert the given 2-byte character sting (e.g. "AA") to hex-decimal (0xAA).
  // Input :
  //   *lpsz_str : 2-byte character string (e.g. "AA").
  //   *lpsz_hex : 1-byte buffer to store converted result (e.g. 0x0AA)
  // Output :
  //   0  : converted OK
  //   -1 : error
  //
  char if_str_2_hex(char *lpsz_str, unsigned char *lpsz_hex)
  {
    if ( (lpsz_str == NULL) || (lpsz_hex == NULL) )
      return (M_FAIL);    
    
    char c_byte0 = if_ascii_2_hex(lpsz_str[0]), c_byte1 = if_ascii_2_hex(lpsz_str[1]);
    if (c_byte0 == M_FAIL || c_byte1 == M_FAIL)
      return (M_FAIL);                         // Convert error
    
	*lpsz_hex = (c_byte0 << 4 | c_byte1);      // OK
    return (0);
  }

  //
  // Subject :
  //   Convert given MAC address string (xx:xx:xx:xx:xx:xx) into 6-byte Ethernet address.
  // Input :
  //   *lpsz_str : pointer to buffer storing xx:xx:xx:xx:xx:xx
  //   *lpsz_eth : pointer to buffer to store 6-byte Ethernet addresses converted
  // Output :
  //   0      :  converted to 6-byte Ethernet addresses successfully
  //   others :  error
  //
  int if_string_2_eth(char *lpsz_str, char *lpsz_eth)
  {
    if ( (lpsz_str == NULL) || (strlen(lpsz_str) != 17) || (lpsz_eth == NULL) )
      return (M_FAIL);
    

    #define M_IS_COLON(n)  (lpsz_str[n] == CHAR_COLON)
    if ( M_IS_COLON(2) && M_IS_COLON(5) && M_IS_COLON(8) && M_IS_COLON(11) && M_IS_COLON(14) )
      ;                                // Separators are OK
    else
      return (M_FAIL);                     // Separator checking failed 
  
    if (if_str_2_hex(lpsz_str+0,  lpsz_eth+0) != 0)
      return (M_FAIL);
    if (if_str_2_hex(lpsz_str+3,  lpsz_eth+1) != 0)
      return (M_FAIL);
    if (if_str_2_hex(lpsz_str+6,  lpsz_eth+2) != 0)
      return (M_FAIL);
    if (if_str_2_hex(lpsz_str+9,  lpsz_eth+3) != 0)
      return (M_FAIL);
    if (if_str_2_hex(lpsz_str+12, lpsz_eth+4) != 0)
      return (M_FAIL);
    if (if_str_2_hex(lpsz_str+15, lpsz_eth+5) != 0)
      return (M_FAIL);
    return (0);     
  }
  
  //
  // Subject :
  //   Verify whether the given MAC address string is valid ?
  // Input :
  //   *lpsz_MAC_string : string with MAC addresses (xx:xx:xx:xx:xx:xx)
  //
  // Output :
  //   1 : valid
  //   0 : not
  //
  int is_MAC_address_valid(char *lpsz_MAC_str)
  {
    // 1) Validate the given parameters ?
    if (lpsz_MAC_str == NULL) {
      printf("%s:error: NULL pointer found.\n", __FUNCTION__);
      return (0);
    }
    
    // 2) Convert the MAC address string into 6-byte MAC address
    unsigned char sz_MAC[ETH_ALEN+1];
    if (if_string_2_eth(lpsz_MAC_str, sz_MAC) != 0) {  // error ?
      printf("%s:error: cannot convert the give MAC string.\n", __FUNCTION__);
      return (0);
    }  
    
    // 3) Is the first byte not 0x00 ?
    if (sz_MAC[0] != 0x00) {
      printf("%s:error: the given MAC address string is not 00:xx:xx:xx:xx:xx\n", __FUNCTION__);
      return (0);
    }
    
    return (1);                        // OK
  }
#endif

/***************************************************************************
 * Watch for 'delay' seconds for autoboot stop or autoboot delay string.
 * returns: 0 -  no key string, allow autoboot
 *          1 - got key string, abort
 */
#if defined(CONFIG_BOOTDELAY) && (CONFIG_BOOTDELAY >= 0)
# if defined(CONFIG_AUTOBOOT_KEYED)
static __inline__ int abortboot(int bootdelay)
{
	int abort = 0;
	uint64_t etime = endtick(bootdelay);
	struct
	{
		char* str;
		u_int len;
		int retry;
	}
	delaykey [] =
	{
		{ str: getenv ("bootdelaykey"),  retry: 1 },
		{ str: getenv ("bootdelaykey2"), retry: 1 },
		{ str: getenv ("bootstopkey"),   retry: 0 },
		{ str: getenv ("bootstopkey2"),  retry: 0 },
	};
	
	char presskey [MAX_DELAY_STOP_STR];
	u_int presskey_len = 0;
	u_int presskey_max = 0;
	u_int i;


#ifdef M_NH230
	// 1) Check whether had the 6-byte MAC addresses been programmed into flash correctly ?
	//    If not yet, return 1 to abort auto-boot.
	//
	char *s;
        if ((s = getenv(STR_ETHADDR)) == NULL) {  // MAC address not programmed yet ?
          printf("PPCBoot: error: MAC addresses are not programmed yet.\n");
          printf("PPCBoot: please type setenv ethaddr=xx:xx:xx:xx:xx:xx and saveenv to fix\n");
          return (1);                             // return 1 to abort the auto-boot
        }
        else {
          // 1-1) MAC address programmed, then check it
          if (is_MAC_address_valid(s) == 0) {
            printf("PPCBoot: error: MAC addresses [%s] in flash are invalid\n", s);
            printf("PPCBoot: please type setenv ethaddr=xx:xx:xx:xx:xx:xx and saveenv to fix\n");
            return (1);                           // return 1 as not valid MAC address
          }
        }
#endif  // end of #ifdef {M_NH230}

#  ifdef CONFIG_AUTOBOOT_PROMPT
	printf (CONFIG_AUTOBOOT_PROMPT, bootdelay);
#  endif

#  ifdef CONFIG_AUTOBOOT_DELAY_STR
	if (delaykey[0].str == NULL)
		delaykey[0].str = CONFIG_AUTOBOOT_DELAY_STR;
#  endif
#  ifdef CONFIG_AUTOBOOT_DELAY_STR2
	if (delaykey[1].str == NULL)
		delaykey[1].str = CONFIG_AUTOBOOT_DELAY_STR2;
#  endif
#  ifdef CONFIG_AUTOBOOT_STOP_STR
	if (delaykey[2].str == NULL)
		delaykey[2].str = CONFIG_AUTOBOOT_STOP_STR;
#  endif
#  ifdef CONFIG_AUTOBOOT_STOP_STR2
	if (delaykey[3].str == NULL)
		delaykey[3].str = CONFIG_AUTOBOOT_STOP_STR2;
#  endif

	for (i = 0; i < sizeof(delaykey) / sizeof(delaykey[0]); i ++) {
		delaykey[i].len = delaykey[i].str == NULL ?
				    0 : strlen (delaykey[i].str);
		delaykey[i].len = delaykey[i].len > MAX_DELAY_STOP_STR ?
				    MAX_DELAY_STOP_STR : delaykey[i].len;

		presskey_max = presskey_max > delaykey[i].len ?
				    presskey_max : delaykey[i].len;

#  if DEBUG_BOOTKEYS
		printf("%s key:<%s>\n",
		       delaykey[i].retry ? "delay" : "stop",
		       delaykey[i].str ? delaykey[i].str : "NULL");
#  endif
	}

	/* In order to keep up with incoming data, check timeout only
	 * when catch up.
	 */
	while (!abort && get_ticks() <= etime) {
		for (i = 0; i < sizeof(delaykey) / sizeof(delaykey[0]); i ++) {
			if (delaykey[i].len > 0 &&
			    presskey_len >= delaykey[i].len &&
			    memcmp (presskey + presskey_len - delaykey[i].len,
		        	    delaykey[i].str,
				    delaykey[i].len) == 0) {
#  if DEBUG_BOOTKEYS
				printf("got %skey\n",
				       delaykey[i].retry ? "delay" : "stop");
#  endif

#  ifdef CONFIG_BOOT_RETRY_TIME
				/* don't retry auto boot */
				if (! delaykey[i].retry)
					retry_time = -1;
#  endif
				abort = 1;
			}
		}

		if (tstc()) {
			if (presskey_len < presskey_max) {
				presskey [presskey_len ++] = getc();
			}
			else {
				for (i = 0; i < presskey_max - 1; i ++)
					presskey [i] = presskey [i + 1];
				
				presskey [i] = getc();
			}
		}
	}
#  if DEBUG_BOOTKEYS
	if (!abort)
		printf("key timeout\n");
#  endif

	return abort;
}

# else	/* !defined(CONFIG_AUTOBOOT_KEYED) */

static __inline__ int abortboot(int bootdelay)
{
	int abort = 0;

	printf("Hit any key to stop autoboot: %2d ", bootdelay);

#if defined CONFIG_ZERO_BOOTDELAY_CHECK
        /*
         * Check if key already pressed
         * Don't check if bootdelay < 0
         */
	if (bootdelay >= 0) {
		if (tstc()) {	/* we got a key press	*/
			(void) getc();  /* consume input	*/
			printf ("\b\b\b 0\n");
			return 1; 	/* don't auto boot	*/
		}
        }
#endif

	while (bootdelay > 0) {
		int i;

		--bootdelay;
		/* delay 100 * 10ms */
		for (i=0; !abort && i<100; ++i) {
			if (tstc()) {	/* we got a key press	*/
				abort  = 1;	/* don't auto boot	*/
				bootdelay = 0;	/* no more delay	*/
				(void) getc();  /* consume input	*/
				break;
			}
			udelay (10000);
		}

		printf ("\b\b\b%2d ", bootdelay);
	}

	putc ('\n');

	return abort;
}
# endif	/* CONFIG_AUTOBOOT_KEYED */
#endif	/* CONFIG_BOOTDELAY >= 0  */

/****************************************************************************/

void main_loop (void)
{
#ifndef CFG_HUSH_PARSER
	static char lastcommand[CFG_CBSIZE] = { 0, };
	int len;
	int rc = 1;
	int flag;
#endif

#if defined(CONFIG_BOOTDELAY) && (CONFIG_BOOTDELAY >= 0)
	char *s;
	int bootdelay;
#endif
#ifdef CONFIG_PREBOOT
	char *p;
#endif

#if defined(CONFIG_VFD) && defined(VFD_TEST_LOGO)
	ulong bmp = 0;		/* default bitmap */
	extern int trab_vfd (ulong bitmap);

#ifdef CONFIG_MODEM_SUPPORT
	if (do_mdm_init)
		bmp = 1;	/* alternate bitmap */
#endif
	trab_vfd (bmp);
#endif	/* CONFIG_VFD && VFD_TEST_LOGO */

#ifdef CONFIG_MODEM_SUPPORT
	debug ("DEBUG: main_loop:   do_mdm_init=%d\n", do_mdm_init);
	if (do_mdm_init) {
		uchar *str = strdup(getenv("mdm_cmd"));
		setenv ("preboot", str);  /* set or delete definition */
		if (str != NULL)
			free (str);
		mdm_init(); /* wait for modem connection */
	}
#endif  /* CONFIG_MODEM_SUPPORT */

#ifdef CFG_HUSH_PARSER
	ppcboot_hush_start ();
#endif

#ifdef CONFIG_PREBOOT
	if ((p = getenv ("preboot")) != NULL) {
# ifdef CONFIG_AUTOBOOT_KEYED
		int prev = disable_ctrlc(1);	/* disable Control C checking */
# endif

# ifndef CFG_HUSH_PARSER
		run_command (p, 0);
# else
		parse_string_outer(p, FLAG_PARSE_SEMICOLON |
				    FLAG_EXIT_FROM_LOOP);
# endif

# ifdef CONFIG_AUTOBOOT_KEYED
		disable_ctrlc(prev);	/* restore Control C checking */
# endif
	}
#endif /* CONFIG_PREBOOT */

#if defined(CONFIG_BOOTDELAY) && (CONFIG_BOOTDELAY >= 0)
	s = getenv ("bootdelay");
	bootdelay = s ? (int)simple_strtol(s, NULL, 10) : CONFIG_BOOTDELAY;

#if 0
	printf ("### main_loop entered:\n\n");
#endif

# ifdef CONFIG_BOOT_RETRY_TIME
	s = getenv ("bootretry");
	if (s != NULL)
		retry_time = (int)simple_strtoul(s, NULL, 10);
	else
		retry_time =  CONFIG_BOOT_RETRY_TIME;
	if (retry_time >= 0 && retry_time < CONFIG_BOOT_RETRY_MIN)
		retry_time = CONFIG_BOOT_RETRY_MIN;
# endif	/* CONFIG_BOOT_RETRY_TIME */

	s = getenv ("bootcmd");
	if (bootdelay >= 0 && s && !abortboot (bootdelay)) {
# ifdef CONFIG_AUTOBOOT_KEYED
		int prev = disable_ctrlc(1);	/* disable Control C checking */
# endif

# ifndef CFG_HUSH_PARSER

        //Add function for POST
		printf("running post....\n");
		run_nh230_post();
		
		run_command (s, 0);
# else
		parse_string_outer(s, FLAG_PARSE_SEMICOLON |
				    FLAG_EXIT_FROM_LOOP);
# endif

# ifdef CONFIG_AUTOBOOT_KEYED
		disable_ctrlc(prev);	/* restore Control C checking */
# endif
	}
#endif	/* CONFIG_BOOTDELAY */

	/*
	 * Main Loop for Monitor Command Processing
	 */
#ifdef CFG_HUSH_PARSER
	parse_file_outer();
	/* This point is never reached */
	for (;;);
#else
	for (;;) {
#ifdef CONFIG_BOOT_RETRY_TIME
		if (rc >= 0) {
			/* Saw enough of a valid command to
			 * restart the timeout.
			 */
			reset_cmd_timeout();
		}
#endif
		len = readline (CFG_PROMPT);

		flag = 0;	/* assume no special flags for now */
		if (len > 0)
			strcpy (lastcommand, console_buffer);
		else if (len == 0)
			flag |= CMD_FLAG_REPEAT;
#ifdef CONFIG_BOOT_RETRY_TIME
		else if (len == -2) {
			/* -2 means timed out, retry autoboot
			 */
			printf("\nTimed out waiting for command\n");
# ifdef CONFIG_RESET_TO_RETRY
			/* Reinit board to run initialization code again */
			do_reset (NULL, 0, 0, NULL);
# else
			return;		/* retry autoboot */
# endif
		}
#endif

		if (len == -1)
			printf ("<INTERRUPT>\n");
		else
			rc = run_command (lastcommand, flag);

		if (rc <= 0) {
			/* invalid command or not repeatable, forget it */
			lastcommand[0] = 0;
		}
	}
#endif /*CFG_HUSH_PARSER*/
}

/***************************************************************************
 * reset command line timeout to retry_time seconds
 */
#ifdef CONFIG_BOOT_RETRY_TIME
void reset_cmd_timeout(void)
{
	endtime = endtick(retry_time);
}
#endif

/****************************************************************************/

/*
 * Prompt for input and read a line.
 * If  CONFIG_BOOT_RETRY_TIME is defined and retry_time >= 0,
 * time out when time goes past endtime (timebase time in ticks).
 * Return:	number of read characters
 *		-1 if break
 *		-2 if timed out
 */
int readline (const char *const prompt)
{
	char   *p = console_buffer;
	int	n = 0;				/* buffer index		*/
	int	plen = 0;			/* prompt length	*/
	int	col;				/* output column cnt	*/
	char	c;

	/* print prompt */
	if (prompt) {
		plen = strlen (prompt);
		puts (prompt);
	}
	col = plen;

	for (;;) {
#ifdef CONFIG_BOOT_RETRY_TIME
		while (!tstc()) {	/* while no incoming data */
			if (retry_time >= 0 && get_ticks() > endtime)
				return (-2);	/* timed out */
		}
#endif
		WATCHDOG_RESET();		/* Trigger watchdog, if needed */

#ifdef CONFIG_SHOW_ACTIVITY
		while (!tstc()) {
			extern void show_activity(int arg);
			show_activity(0);
		}
#endif		
		c = getc();

		/*
		 * Special character handling
		 */
		switch (c) {
		case '\r':				/* Enter		*/
		case '\n':
			*p = '\0';
			puts ("\r\n");
			return (p - console_buffer);

		case 0x03:				/* ^C - break		*/
			console_buffer[0] = '\0';	/* discard input */
			return (-1);

		case 0x15:				/* ^U - erase line	*/
			while (col > plen) {
				puts (erase_seq);
				--col;
			}
			p = console_buffer;
			n = 0;
			continue;

		case 0x17:				/* ^W - erase word 	*/
			p=delete_char(console_buffer, p, &col, &n, plen);
			while ((n > 0) && (*p != ' ')) {
				p=delete_char(console_buffer, p, &col, &n, plen);
			}
			continue;

		case 0x08:				/* ^H  - backspace	*/
		case 0x7F:				/* DEL - backspace	*/
			p=delete_char(console_buffer, p, &col, &n, plen);
			continue;

		default:
			/*
			 * Must be a normal character then
			 */
			if (n < CFG_CBSIZE-2) {
				if (c == '\t') {	/* expand TABs		*/
					puts (tab_seq+(col&07));
					col += 8 - (col&07);
				} else {
					++col;		/* echo input		*/
					putc (c);
				}
				*p++ = c;
				++n;
			} else {			/* Buffer full		*/
				putc ('\a');
			}
		}
	}
}

/****************************************************************************/

static char * delete_char (char *buffer, char *p, int *colp, int *np, int plen)
{
	char *s;

	if (*np == 0) {
		return (p);
	}

	if (*(--p) == '\t') {			/* will retype the whole line	*/
		while (*colp > plen) {
			puts (erase_seq);
			(*colp)--;
		}
		for (s=buffer; s<p; ++s) {
			if (*s == '\t') {
				puts (tab_seq+((*colp) & 07));
				*colp += 8 - ((*colp) & 07);
			} else {
				++(*colp);
				putc (*s);
			}
		}
	} else {
		puts (erase_seq);
		(*colp)--;
	}
	(*np)--;
	return (p);
}

/****************************************************************************/

int parse_line (char *line, char *argv[])
{
	int nargs = 0;

#ifdef DEBUG_PARSER
	printf ("parse_line: \"%s\"\n", line);
#endif
	while (nargs < CFG_MAXARGS) {

		/* skip any white space */
		while ((*line == ' ') || (*line == '\t')) {
			++line;
		}

		if (*line == '\0') {	/* end of line, no more args	*/
			argv[nargs] = NULL;
#ifdef DEBUG_PARSER
		printf ("parse_line: nargs=%d\n", nargs);
#endif
			return (nargs);
		}

		argv[nargs++] = line;	/* begin of argument string	*/

		/* find end of string */
		while (*line && (*line != ' ') && (*line != '\t')) {
			++line;
		}

		if (*line == '\0') {	/* end of line, no more args	*/
			argv[nargs] = NULL;
#ifdef DEBUG_PARSER
		printf ("parse_line: nargs=%d\n", nargs);
#endif
			return (nargs);
		}

		*line++ = '\0';		/* terminate current arg	 */
	}

	printf ("** Too many args (max. %d) **\n", CFG_MAXARGS);

#ifdef DEBUG_PARSER
	printf ("parse_line: nargs=%d\n", nargs);
#endif
	return (nargs);
}

/****************************************************************************/

static void process_macros (const char *input, char *output)
{
	char c, prev;
	const char *varname_start = NULL;
	int inputcnt  = strlen (input);
	int outputcnt = CFG_CBSIZE;
	int state = 0;	/* 0 = waiting for '$'	*/
			/* 1 = waiting for '('	*/
			/* 2 = waiting for ')'	*/

#ifdef DEBUG_PARSER
	char *output_start = output;

	printf ("[PROCESS_MACROS] INPUT len %d: \"%s\"\n", strlen(input), input);
#endif

	prev = '\0';			/* previous character	*/

	while (inputcnt && outputcnt) {
	    c = *input++;
	    inputcnt--;

	    /* remove one level of escape characters */
	    if ((c == '\\') && (prev != '\\')) {
		if (inputcnt-- == 0)
			break;
		prev = c;
	    	c = *input++;
	    }

	    switch (state) {
	    case 0:			/* Waiting for (unescaped) $	*/
		if ((c == '$') && (prev != '\\')) {
			state++;
		} else {
			*(output++) = c;
			outputcnt--;
		}
		break;
	    case 1:			/* Waiting for (	*/
		if (c == '(') {
			state++;
			varname_start = input;
		} else {
			state = 0;
			*(output++) = '$';
			outputcnt--;

			if (outputcnt) {
				*(output++) = c;
				outputcnt--;
			}
		}
		break;
	    case 2:			/* Waiting for )	*/
		if (c == ')') {
			int i;
			char envname[CFG_CBSIZE], *envval;
			int envcnt = input-varname_start-1; /* Varname # of chars */

			/* Get the varname */
			for (i = 0; i < envcnt; i++) {
				envname[i] = varname_start[i];
			}
			envname[i] = 0;

			/* Get its value */
			envval = getenv (envname);

			/* Copy into the line if it exists */
			if (envval != NULL)
				while ((*envval) && outputcnt) {
					*(output++) = *(envval++);
					outputcnt--;
				}
			/* Look for another '$' */
			state = 0;
		}
		break;
	    }

	    prev = c;
	}

	if (outputcnt)
		*output = 0;

#ifdef DEBUG_PARSER
	printf ("[PROCESS_MACROS] OUTPUT len %d: \"%s\"\n",
		strlen(output_start), output_start);
#endif
}

/****************************************************************************
 * returns:
 *	1  - command executed, repeatable
 *	0  - command executed but not repeatable, interrupted commands are
 *	     always considered not repeatable
 *	-1 - not executed (unrecognized, bootd recursion or too many args)
 *           (If cmd is NULL or "" or longer than CFG_CBSIZE-1 it is
 *           considered unrecognized)
 *
 * WARNING:
 *
 * We must create a temporary copy of the command since the command we get
 * may be the result from getenv(), which returns a pointer directly to
 * the environment data, which may change magicly when the command we run
 * creates or modifies environment variables (like "bootp" does).
 */

int run_command (const char *cmd, int flag)
{
	cmd_tbl_t *cmdtp;
	char cmdbuf[CFG_CBSIZE];	/* working copy of cmd		*/
	char *token;			/* start of token in cmdbuf	*/
	char *sep;			/* end of token (separator) in cmdbuf */
	char finaltoken[CFG_CBSIZE];
	char *str = cmdbuf;
	char *argv[CFG_MAXARGS + 1];	/* NULL terminated	*/
	int argc;
	int repeatable = 1;

#ifdef DEBUG_PARSER
	printf ("[RUN_COMMAND] cmd[%p]=\"", cmd);
	puts (cmd ? cmd : "NULL");	/* use puts - string may be loooong */
	puts ("\"\n");
#endif

	clear_ctrlc();		/* forget any previous Control C */

	if (!cmd || !*cmd) {
		return -1;	/* empty command */
	}

	if (strlen(cmd) >= CFG_CBSIZE) {
		puts ("## Command too long!\n");
		return -1;
	}

	strcpy (cmdbuf, cmd);

	/* Process separators and check for invalid
	 * repeatable commands
	 */

#ifdef DEBUG_PARSER
	printf ("[PROCESS_SEPARATORS] %s\n", cmd);
#endif
	while (*str) {

		/*
		 * Find separator, or string end
		 * Allow simple escape of ';' by writing "\;"
		 */
		for (sep = str; *sep; sep++) {
			if ((*sep == ';') &&	/* separator		*/
			    ( sep != str) &&	/* past string start	*/
			    (*(sep-1) != '\\'))	/* and NOT escaped	*/
				break;
		}

		/*
		 * Limit the token to data between separators
		 */
		token = str;
		if (*sep) {
			str = sep + 1;	/* start of command for next pass */
			*sep = '\0';
		}
		else
			str = sep;	/* no more commands for next pass */
#ifdef DEBUG_PARSER
		printf ("token: \"%s\"\n", token);
#endif

		/* find macros in this token and replace them */
		process_macros (token, finaltoken);

		/* Extract arguments */
		argc = parse_line (finaltoken, argv);

		/* Look up command in command table */
		if ((cmdtp = find_cmd(argv[0])) == NULL) {
			printf ("Unknown command '%s' - try 'help'\n", argv[0]);
			return -1;	/* give up after bad command */
		}

		/* found - check max args */
		if (argc > cmdtp->maxargs) {
			printf ("Usage:\n%s\n", cmdtp->usage);
			return -1;
		}

#if (CONFIG_COMMANDS & CFG_CMD_BOOTD)
		/* avoid "bootd" recursion */
		if (cmdtp->cmd == do_bootd) {
#ifdef DEBUG_PARSER
			printf ("[%s]\n", finaltoken);
#endif
			if (flag & CMD_FLAG_BOOTD) {
				printf ("'bootd' recursion detected\n");
				return -1;
			}
			else
				flag |= CMD_FLAG_BOOTD;
		}
#endif	/* CFG_CMD_BOOTD */

		/* OK - call function to do the command */
		if ((cmdtp->cmd) (cmdtp, flag, argc, argv) != 0) {
			return (-1);
		}

		repeatable &= cmdtp->repeatable;

		/* Did the user stop this? */
		if (had_ctrlc ())
			return 0;	/* if stopped then not repeatable */
	}

	return repeatable;
}

/****************************************************************************/

#if (CONFIG_COMMANDS & CFG_CMD_RUN)
int do_run (cmd_tbl_t * cmdtp, int flag, int argc, char *argv[])
{
	int i;
	int rcode = 1;

	if (argc < 2) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}

	for (i=1; i<argc; ++i) {
#ifndef CFG_HUSH_PARSER
	    if (run_command (getenv (argv[i]), flag) != -1) ++rcode;
#else
   	    if (parse_string_outer(getenv (argv[i]),
		    FLAG_PARSE_SEMICOLON | FLAG_EXIT_FROM_LOOP) == 0) ++rcode;
#endif
	}
	return ((rcode == i) ? 0 : 1);
}
#endif

// Modify by George for POST at 06/21/2005
void run_nh230_post()
{
	int i, j, k, nxt;
	char val[10];

	k = -1;  i = 0;

	memset(val, 0, sizeof(val));
	for (j=0; env_get_char(j) != '\0'; j=nxt+1) {

		for (nxt=j; env_get_char(nxt) != '\0'; ++nxt)
				;
		k = postcnt("post", j);
		if (k < 0) {
			continue;
		}
		while (k < nxt)
			val[i++] = env_get_char(k++);
		break;
	}

	if (k < 0) {
		return;
	}
	
	if (val[0] == 'y' || val[0] == 'Y') {
		do_run_nh230_post(240);
	}
	else if (val[0] < 0x30 || val[0] > 0x39) {
		do_run_nh230_post(0);
	}
	else {
		do_run_nh230_post(StringToInt(val));
	}
}

static int
postcnt (uchar *s1, int i2)
{

	while (*s1 == env_get_char(i2++))
		if (*s1++ == '=')
			return(i2);
	if (*s1 == '\0' && env_get_char(i2-1) == '=')
		return(i2);
	return(-1);
}

void do_run_nh230_post(int cnt)
{
	int i;
	unsigned int offset=0;
	unsigned char c_byte=0;

	if (cnt == 0)
	    return;

	for(i=0; i<cnt; i++) {
		if (flash_test(cnt) != 0) {
			WritePOSTCntToFLASH(cnt, cnt);
			DeadLoop(1);
		}
		if (memory_test(cnt) != 0) {
			WritePOSTCntToFLASH(cnt, cnt);
			DeadLoop(1);
		}
		if (pci_device_scan(cnt) != 0) {
			WritePOSTCntToFLASH(cnt, cnt);
			DeadLoop(1);
		}
		WritePOSTCntToFLASH(i, cnt);
		printf("POST loop %d done.\n\n", i+1);
	}
	WritePOSTCntToFLASH(i, cnt);
	DeadLoop(0);
}

int  flash_test(int cnt)
{
    unsigned char c_byte = 0;
    unsigned int offset = 0;      // start from 0
    uchar pattern[] = {0x55, 0xAA};
//    uchar pattern[16];
    int i, rc;
/* 
    offset = 0x00300000;
    memset(pattern, 0, sizeof(pattern));
    for (i=0; i<12; i++, offset++) {
        c_byte =  ( *((volatile unsigned char *)(FLASH_BASE0_PRELIM + offset) ) );
        pattern[i] = c_byte;
    }
    if (strstr(pattern, "VPPCBoot") == NULL) {
        sprintf(ErrMsg, "flash test error!\n");
	return 1;
    }
    return 0;
*/

    offset = 0x0000A000;    //Reserver Partition 0x0000A0000~0x00010000 24KB
    if (flash_sect_protect (0, 0xFFC0A000, 0xFFC0FFFF)) {
    	puts("flash_sect_protect error!\n");
        sprintf(ErrMsg, "Loop%d: flash_sect_protect error!\n", cnt);
        return 1;
    }

    if (flash_sect_erase (0xFFC0A000, 0xFFC0FFFF)) {
    	puts("flash_sect_erase error!\n");
        sprintf(ErrMsg, "Loop%d: flash_sect_erase error!\n", cnt);
        goto Err;
    }

    // Continue to write flash must interval 1 offset, or else will write fial
    // ex. write offset: 00, 02, 04, 06....
    for (i=0; i<6*1024; i++, offset+=4) {
	udelay(10000);
	rc = flash_write(pattern, FLASH_BASE0_PRELIM + offset, 2);
        if (rc != 0) {
            printf("flash write error! offset: %08lX\n", (ulong)offset);
            sprintf(ErrMsg, "Loop%d: flash write error! offset: %08lX\n", cnt, (ulong)offset);
            break;
        }
        
        c_byte =  ( *((volatile unsigned char *)(FLASH_BASE0_PRELIM + offset) ) );
	if (c_byte != 0x55) {
	    printf("read flash error! offset: %08lX\n", (ulong)offset);
            sprintf(ErrMsg, "Loop%d: read flash error! offset: %08lX\n", cnt, (ulong)offset);
            break;
        }
        
        c_byte =  ( *((volatile unsigned char *)(FLASH_BASE0_PRELIM + offset+1) ) );
	if (c_byte != 0xAA) {
	    printf("read flash error! offset: %08lX\n", (ulong)offset);
            sprintf(ErrMsg, "Loop%d: read flash error! offset: %08lX\n", cnt, (ulong)offset+1);
            break;
        }
    }

    if (flash_sect_erase (0xFFC0A000, 0xFFC0FFFF)) {
    	puts("flash_sect_erase error!\n");
        sprintf(ErrMsg, "Loop%d: flash_sect_erase error!\n", cnt);
    }
    
Err:
    if (flash_sect_protect (1, 0xFFC0A000, 0xFFC0FFFF)) {
    	puts("flash_sect_protect error!\n");
        sprintf(ErrMsg, "Loop%d: flash_sect_protect error!\n", cnt);
        return 1;
    }
    return 0;
}

int memory_test(int cnt)
{
     unsigned char c_byte = 0;
     ulong offset = 0;    // start from 0 
     ulong i;
 
     offset = CFG_LOAD_ADDR;
     for (i=0; i<30*1024*0124; i++, offset+=2) {
        /* memory write to 0x55*/
         ( *((volatile unsigned char *)(offset) ) )  = 0x55;

        /* memory read */
         c_byte =  ( *((volatile unsigned char *)(offset) ) );
	 if (c_byte != 0x55) {
	     printf("DRAM test Error! offset: %08lX\n", offset);
	     sprintf(ErrMsg, "Loop%d: DRAM test Error! offset: %08lX\n", cnt, offset);
	     return 1; 
	 }

        /* memory write to 0xAA*/
         ( *((volatile unsigned char *)(offset+1) ) )  = 0xAA;

        /* memory read */
         c_byte =  ( *((volatile unsigned char *)(offset+1) ) );
	 if (c_byte != 0xAA) {
	     printf("DRAM test Error! offset: %08lX\n", offset+1);
	     sprintf(ErrMsg, "Loop%d: DRAM test Error! offset: %08lX\n", cnt,  offset+1);
	     return 1;
	 }
    }
    return 0;
}

int pci_device_scan(int cnt)
{
	pci_dev_t devno;

	/* Find RTL8169 */
	if ((devno = pci_find_devices(RTL8169S, 0)) < 0) {
		puts("PCI device scan error: not found RTL169\n");
		sprintf(ErrMsg, "Loop%d: PCI device scan error: not found RTL169\n", cnt);
		return 1;
	}

	/* Find IT8212 & IT8211 */
	if ((devno = pci_find_devices(IT8212, 0)) < 0) {
		if ((devno = pci_find_devices(IT8211, 0)) < 0) {
			puts("PCI device scan error: not found IT8212 or IT8211\n");
			sprintf(ErrMsg, "Loop%d: Loop%d: PCI device scan error: not found IT8212 or IT8211\n", cnt);
			return 1;
		}
	}

	/* Find GL880S */
	if ((devno = pci_find_devices(EHCI, 0)) < 0) {
		puts("PCI device scan error: not found EHCI controller\n");
		sprintf(ErrMsg, "Loop%d: PCI device scan error: not found EHCI controller\n", cnt);
		return 1;
	}
	if ((devno = pci_find_devices(UHCI, 0)) < 0) {
		puts("PCI device scan error: not found UHCI controller\n");
		sprintf(ErrMsg, "Loop%d: PCI device scan error: not found UHCI controller\n", cnt);
		return 1;
	}
	
	return 0;
}

int StringToInt(char *lpsz_str)
{
    int i, j, value = 0;
    char c_byte = 0;

    if ( lpsz_str == NULL )
      return (0);    
    
    i = strlen(lpsz_str)-1;
    for ( j=0; i>=0; i--, j++) {
      c_byte = if_ascii_2_hex(lpsz_str[i]);
      value = (int)(c_byte * exp10(j));
    }
    
    if (value > 240)
      value = 240;

    return (value);
}

int exp10(int x)
{
    int i, res=1;
	
    for (i=0; i<x; i++) {
      res *= 10;
    }

    return (res);
}

void WritePOSTCntToFLASH(int LoopCnt, int Cnt)
{
    int value;
    char varvalue[10];

    value = Cnt - LoopCnt;
    sprintf(varvalue, "%d", value);
    setenv ("post", varvalue);
    saveenv();
}
void DeadLoop(int flag)
{
	if (flag == 1)
		WritePOSTResultToFLASH(1);
	else
		WritePOSTResultToFLASH(0);
/*
	while(1) {
//		if (flag == 0) {
			;
//		}
//		else {
			;
//		}
	}
*/
}
void WritePOSTResultToFLASH(int flag)
{
     
     unsigned int offset = 0;    // start from 0 
     int rc;

    offset = 0x00002000;
    if (flash_sect_protect (0, 0xFFC02000, 0xFFC03FFF))
        printf("flash_sect_protect error!\n");

    if (flash_sect_erase (0xFFC02000, 0xFFC03FFF))
        printf("flash_sect_erase error!\n");

    strcpy(PassMsg, "POST all items tested pass!");
    if (flag == 1)
        rc = flash_write(ErrMsg, FLASH_BASE0_PRELIM + offset, sizeof(ErrMsg));
    else
        rc = flash_write(PassMsg, FLASH_BASE0_PRELIM + offset, sizeof(PassMsg));
//        rc = flash_write("POST all items tested pass!", FLASH_BASE0_PRELIM + offset, 27);

    if (rc != 0) {
         printf("flash write error!\n");
    }
 
    if (flash_sect_protect (1, 0xFFC02000, 0xFFC03FFF))
        printf("flash_sect_protect error!\n");
}
