/*
 * (C) Copyright 2001, 2002
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/* ------------------------------------------------------------------------- */

/*
 *
 * History : 
 *   V1.00  01/14/2005  Hank Chu
 *     1) Modify to support the NH-230 NAS.
 *        1-1) The H/W architecture :
 *             CPU : Motorola MPC8241, 266MHz
 *             OS: Linux
 *             Flash : 1 pcs 4MB flash chip (bottom-boot)
 *             Memory : 4 pcs SDRAM chips (total 64MB)       
 * 
 *
 * Notes :
 *   1) The flash partitions for NH-230 NAS.
 *
 *        Flash Partition   Address Map B   Offset of Flash   Size
 *                          Start           Start
 *        ---------------   -------------   ---------------   ----------
 *        Boot CFG          0xFFC0_0000     0x0000_0000       16KB
 *         (Environment)
 *
 *        POST Result       0xFFC0_4000     0x0000_4000       8KB
 *
 *
 *        Reserved          0xFFC0_6000     0x0000_6000       8KB
 *
 *
 *        Linux CFG         0xFFC0_8000     0x0000_8000       32KB
 *
 *
 *        Linux Image       0xFFC1_0000     0x0001_0000       1536KB
 *
 *
 *        RAM disk #2       0xFFD9_0000     0x0019_0000       1472KB
 *         (Other App.)
 *
 *        Boot-loader       0xFFF0_0000     0x0030_0000       128KB
 *         + POST
 *
 *        RAM disk #1       0xFFF2_0000     0x0032_0000       896KB
 *         (root)
 *
 *   2) PCI Bus Assignment for NH-230 :
 *        The PCI Device Number is used to select which IDESL to assert.
 *
 *        Device Number  Device                     IRQ#  IDESL
 *        -------------  ------------------------   ----  ---------
 *            14          PCI-to-USB                3     PCI_AD14
 *            15          PCI-to-Gigabit Ethernet   1     PCI_AD15
 *            16          PCI-to-IDE                2     PCI_AD16
 *
 *
 *
 *
 *
 */


/*
 * board/config.h - configuration options, board specific
 */

#ifndef __CONFIG_H
#define __CONFIG_H

/*
 * High Level Configuration Options
 * (easy to change)
 */
/* //#define DEBUG 1  */
#define CONFIG_MPC824X		1
#define CONFIG_MPC8245		1
#define CONFIG_SANDPOINT	1

#if 0
#define USE_DINK32		1
#else
#undef USE_DINK32
#endif

/* General Setting */

#ifdef M_NH230
   /* NH-230, H/W use 1 chip-select for 1 physical bank only. */
  #define DRAM_BANKS         1

  /* PCI Bus : Device Number (same as IDESL) */
  #define PCI_DEVICE_NUMBER_USB       14  /* PCI Device Number for PCI-to-USB */
  #define PCI_DEVICE_NUMBER_ETHERNET  15  /* PCI Device Number for PCI-to-Ethernet */
  #define PCI_DEVICE_NUMBER_IDE       16  /* PCI Device Number for PCI-to-IDE */
  
  #define CONFIG_ENV_OVERWRITE  /* deinfe to allow to overwrite environment variables*/
  
  #define STR_ETHADDR   "ethaddr"         /* Name of environment variable for MAC addresses */
  
  /* NH-230: H/W design supports up to 128MB. But we just use 64MB now. */
  #define DRAM_SIZE_MB           128      /* The maximum size of memory supported */
  
  #define S_BANK_FLASH_SIZE_MB    4       /* The size of single FLASH chip in MB */
  #define CFG_MAX_FLASH_BANKS     1       /* Only 1 bank for 1 flash chip */
  #define CFG_MAX_FLASH_SECT	 71       /* Maximum NO# of sectors in 1 flash chip (up to 4MB) */
  
  /*  Starting address of Extended ROM space selected by RCS3 for GPIO pins access */
  #define CFG_ERCR4_RCS3_START_ADDRESS    0x70000000  /* 4KB : 0x7000_0000 ~ 0x7000_0FFF */
  
    /*
     * The constants for Boot CFG (Environment) in flash
     */
  #define CFG_ENV_IS_IN_FLASH	  1
  #define CFG_ENV_OFFSET	0x0       /* Offset in flash for Environment variable (Boot CFG) */
    /* Total Size of Environment Sector */
//  #define CFG_ENV_SIZE		0x00004000
  #define CFG_ENV_SIZE		0x00002000
  
    /* The autoboot command ->  bootm kernel_addr ramdisk_addr  (address in Address Map B) */
  #define CONFIG_BOOTCOMMAND	"bootm ffc10000 fff20000"
  
    /* The default values of environment variables   */
  #define CONFIG_IPADDR		192.168.1.1
  #define CONFIG_SERVERIP	192.168.1.254
  #define CONFIG_GATEWAYIP      192.168.1.254
  #define CONFIG_NETMASK        255.255.255.0
  #define CONFIG_RUNPOST        0                    /* 1 to run POST during boot up, 0 not */
  
  #define CONFIG_BAUDRATE	9600      /* Should we set to 57600 ???? */
  
  /* MPC8241 Integrated Processor Hardware Specifications :                            */
  /*   Table 19. PLL Configurations (266MHz Parts) :                                   */
  /*     PLL_CFG[4:0] = 0b10000, Peripheral Logic/Mem Bus Clock Range (MHz) = 75 ~ 132 */
  /* It is the maximum Mem Bus Clock for SDRAM used. For SDRAM (PC133), assume 133MHz. */
  #define CONFIG_DRAM_SPEED	133
  
#else
  #define DRAM_BANKS 2
  #define DRAM_SIZE_MB 32
  #define S_BANK_FLASH_SIZE_MB 4 /*The size of single FLASH*/
  #define CFG_MAX_FLASH_BANKS	2
  #define CFG_MAX_FLASH_SECT	71	/* org=20 , max number of sectors on one chip	*/
  
  #define CFG_ENV_IS_IN_FLASH	1
  #define CFG_ENV_OFFSET		0x3fe000	/* Offset in flash for Environment variable */
  #define CFG_ENV_SIZE		0x00002000	/* Total Size of Environment Sector*/
  #define CONFIG_BOOTCOMMAND	"bootm fff30000 ff400000" /* autoboot command -> bootm kernel_addr ramdisk_addr	*/  
  
  #define CONFIG_BAUDRATE		9600
  #define CONFIG_DRAM_SPEED	100		/* MHz */  
#endif  /* end of #ifdef {M_NH230} */


#define FLASH_DATA_BUS_WIDTH  8		/*the bus width of ROM is 8 bit */

#define CONFIG_CONS_INDEX     3               /* set to '3' for on-chip DUART */

/* General Setting */


#ifdef M_NH230
  /* Define CONFIG_POST with more CFG_POST_xxxx for more POST tests, but CPU dependent.
   * Currently, POST supported on MPC8260 & MPC8xx only.
   * Requirements :
   *   1) The CFG_CMD_DIAG should be included in CONFIG_COMMANDS.
   *   2) The COFNIG_POST needs to be defined.
   *
   */
  /* #define CONFIG_POST 		        (CFG_POST_MEMORY) */  
  
  #define CONFIG_COMMANDS ( CONFIG_CMD_DFL & 	\
  			    ~(CFG_CMD_BDI | CFG_CMD_MEMORY | CFG_CMD_CONSOLE |	 \
  			      CFG_CMD_ECHO | CFG_CMD_LOADB | CFG_CMD_LOADS |	 \
  			      CFG_CMD_RUN | CFG_CMD_MISC | CFG_CMD_AUTOSCRIPT))
#else
  #define CONFIG_COMMANDS		( CONFIG_CMD_DFL | \
  					  CFG_CMD_ELF    | \
					  CFG_CMD_I2C 	 | \
					  CFG_CMD_EEPROM | \
					  CFG_CMD_PCI )
#endif


/* this must be included AFTER the definition of CONFIG_COMMANDS (if any)	*/
#include <cmd_confdefs.h>


/*
 * Miscellaneous configurable options
 */
#ifdef M_NH230
  #undef  CFG_LONGHELP		/* undef to save memory		*/
#else
  #define CFG_LONGHELP		1		/* undef to save memory		*/
#endif

#define CFG_PROMPT		"_MPC824X > "		/* Monitor Command Prompt	*/
#define CFG_CBSIZE		256		/* Console I/O Buffer Size	*/
#define CFG_PBSIZE (CFG_CBSIZE+sizeof(CFG_PROMPT)+16)	/* Print Buffer Size	*/
#define CFG_MAXARGS		16		/* max number of command args	*/
#define CFG_BARGSIZE		CFG_CBSIZE	/* Boot Argument Buffer Size	*/
#define CFG_LOAD_ADDR		0x00100000	/* default load address		*/
#define CFG_HZ			1000		/* decrementer freq: 1 ms ticks */

#ifdef M_NH230
  #define CONFIG_BOOTDELAY	3	/* autoboot after 3 seconds	*/
#else
  #define CONFIG_BOOTDELAY	1	/* autoboot after 3 seconds	*/
#endif

#define CONFIG_AUTOBOOT_KEYED 1
#define CONFIG_AUTOBOOT_PROMPT "Press space to abort autoboot in %d second\n"
#define CONFIG_AUTOBOOT_DELAY_STR "d"
#define CONFIG_AUTOBOOT_STOP_STR " "
/*-----------------------------------------------------------------------
 * PCI stuff
 *-----------------------------------------------------------------------
 */
#define CONFIG_PCI				/* include pci support		*/
#undef CONFIG_PCI_PNP

#define CONFIG_NET_MULTI			/* Multi ethernet cards support */

#ifdef M_NH230
	#define CONFIG_NET_RTL8169S             /* added for Gigabit Ethernet NIC */
#else
	#define CONFIG_EEPRO100
	#define CONFIG_NATSEMI
	#define CONFIG_NS8382X
#endif


#define PCI_ENET0_IOADDR	0x80000000
#define PCI_ENET0_MEMADDR	0x80000000
#define	PCI_ENET1_IOADDR	0x81000000
#define	PCI_ENET1_MEMADDR	0x81000000


/*-----------------------------------------------------------------------
 * Start addresses for the final memory configuration
 * (Set up by the startup code)
 * Please note that CFG_SDRAM_BASE _must_ start at 0
 */
#define CFG_SDRAM_BASE		0x00000000

#define CFG_MAX_RAM_SIZE	( DRAM_SIZE_MB * 1024 * 1024 )  

#define CFG_RESET_ADDRESS	0xFFF00100

#if defined (USE_DINK32)
  #define CFG_MONITOR_LEN		0x00030000
  #define CFG_MONITOR_BASE	0x00090000
  #define CFG_RAMBOOT		1
  #define CFG_INIT_RAM_ADDR	(CFG_MONITOR_BASE + CFG_MONITOR_LEN)
  #define CFG_INIT_RAM_END	0x10000
  #define CFG_GBL_DATA_SIZE	256  /* size in bytes reserved for initial data */
  #define CFG_GBL_DATA_OFFSET	(CFG_INIT_RAM_END - CFG_GBL_DATA_SIZE)
  #define CFG_INIT_SP_OFFSET	CFG_GBL_DATA_OFFSET
#else
  #undef	CFG_RAMBOOT

  #ifdef M_NH230
    #define CFG_MONITOR_LEN  0x00030000  /* 192KB for flash partition of boot-loader */
  #else
    #define CFG_MONITOR_LEN  0x00030000
  #endif

  #define CFG_MONITOR_BASE  TEXT_BASE    /* Address Map B of MPC8241, boot-loader on 0xFFF0-0000 */

  /*#define CFG_GBL_DATA_SIZE    256*/
  #define CFG_GBL_DATA_SIZE	128

  #define CFG_INIT_RAM_ADDR     0x40000000
  #define CFG_INIT_RAM_END      0x1000

  #define CFG_GBL_DATA_OFFSET  (CFG_INIT_RAM_END - CFG_GBL_DATA_SIZE)

#endif

#define S_ROM_SIZE  ( S_BANK_FLASH_SIZE_MB * 1024 * 1024 ) /*The size of single FLASH*/
#if CFG_MAX_FLASH_BANKS == 1
#define CFG_FLASH_BASE		( 0xFFFFFFFF - S_ROM_SIZE ) + 1 
#else
#define CFG_FLASH_BASE		( 0xff800000 - S_ROM_SIZE )   
#endif

#define CFG_FLASH_SIZE		S_ROM_SIZE 	/* Unity has onboard 2MByte flash modify by andyw */

#define CFG_FLASH_RANGE_BASE	CFG_FLASH_BASE /* org=0xFF000000 flash memory address range	//andyw cs0*/
#define CFG_FLASH_RANGE_SIZE	( S_ROM_SIZE * CFG_MAX_FLASH_BANKS ) /* org=0x01000000 ,4MB andyw*/
#define FLASH_BASE0_PRELIM	( 0xFFFFFFFF - S_ROM_SIZE ) + 1 
#define FLASH_BASE1_PRELIM	( 0xff800000 - S_ROM_SIZE )	

#define CFG_ENV_ADDR	(FLASH_BASE0_PRELIM + CFG_ENV_OFFSET)
#define CFG_MALLOC_LEN		(512 << 10)	/* Reserve 512 kB for malloc()	*/

#define CFG_MEMTEST_START	0x00000000	/* memtest works on		*/
#define CFG_MEMTEST_END		CFG_MAX_RAM_SIZE	/* 0 ... 32 MB in DRAM		*/

#define CFG_EUMB_ADDR		0xFC000000 

#define CFG_ISA_MEM		0xFD000000
#define CFG_ISA_IO		0xFE000000

/*
 * select i2c support configuration
 *
 * Supported configurations are {none, software, hardware} drivers.
 * If the software driver is chosen, there are some additional
 * configuration items that the driver uses to drive the port pins.
 */
#ifdef M_NH230
  #undef  CONFIG_HARD_I2C		        /* To enable I2C support	*/
  #undef  CONFIG_SOFT_I2C			/* I2C bit-banged		*/
#else
  #define CONFIG_HARD_I2C	1		/* To enable I2C support	*/
  #undef  CONFIG_SOFT_I2C			/* I2C bit-banged		*/
#endif

#define CFG_I2C_SPEED		400000		/* I2C speed and slave address	*/

#ifdef M_NH230
  #define CFG_I2C_SLAVE		0x51
#else
  #define CFG_I2C_SLAVE		0x7F
#endif

#ifdef CONFIG_SOFT_I2C
#error "Soft I2C is not configured properly.  Please review!"
#define I2C_PORT		3               /* Port A=0, B=1, C=2, D=3 */
#define I2C_ACTIVE		(iop->pdir |=  0x00010000)
#define I2C_TRISTATE		(iop->pdir &= ~0x00010000)
#define I2C_READ		((iop->pdat & 0x00010000) != 0)
#define I2C_SDA(bit)		if(bit) iop->pdat |=  0x00010000; \
				else    iop->pdat &= ~0x00010000
#define I2C_SCL(bit)		if(bit) iop->pdat |=  0x00020000; \
				else    iop->pdat &= ~0x00020000
#define I2C_DELAY		udelay(5)	/* 1/4 I2C clock duration */
#endif /* CONFIG_SOFT_I2C */

#define CFG_I2C_EEPROM_ADDR	0x57		/* EEPROM IS24C02		*/
#define CFG_I2C_EEPROM_ADDR_LEN	1		/* Bytes of address		*/
#define CFG_EEPROM_PAGE_WRITE_BITS	3
#define CFG_EEPROM_PAGE_WRITE_DELAY_MS	10	/* and takes up to 10 msec */

#define CFG_BAUDRATE_TABLE	{ 9600, 19200, 38400, 57600, 115200 }

#if CFG_MAX_FLASH_BANKS == 1  
#define CFG_FLASH_BANKS		{ FLASH_BASE0_PRELIM }
#else
#define CFG_FLASH_BANKS		{ FLASH_BASE0_PRELIM , FLASH_BASE1_PRELIM }
#endif


/*-----------------------------------------------------------------------
 * Definitions for initial stack pointer and data area (in DPRAM)
 */


/*#define CFG_WINBOND_83C553	1*/	/*has a winbond bridge	//andyw none use		*/
#define CFG_USE_WINBOND_IDE	0	/*use winbond 83c553 internal IDE ctrlr */
#define CFG_WINBOND_ISA_CFG_ADDR    0x80005800	/*pci-isa bridge config addr	*/
#define CFG_WINBOND_IDE_CFG_ADDR    0x80005900	/*ide config addr		*/

#define CFG_IDE_MAXBUS		2   /* max. 2 IDE busses	*/
#define CFG_IDE_MAXDEVICE	(CFG_IDE_MAXBUS*2) /* max. 2 drives per IDE bus */

/*
 * NS87308 Configuration
 */
#define CFG_NS87308			/* Nat Semi super-io controller on ISA bus */

#define CFG_NS87308_BADDR_10	1

#define CFG_NS87308_DEVS	( CFG_NS87308_UART1   | \
				  CFG_NS87308_UART2   | \
				  CFG_NS87308_POWRMAN | \
				  CFG_NS87308_RTC_APC )

#undef  CFG_NS87308_PS2MOD

#define CFG_NS87308_CS0_BASE	0x0076
#define CFG_NS87308_CS0_CONF	0x30
#define CFG_NS87308_CS1_BASE	0x0075
#define CFG_NS87308_CS1_CONF	0x30
#define CFG_NS87308_CS2_BASE	0x0074
#define CFG_NS87308_CS2_CONF	0x30

/*
 * NS16550 Configuration
 */
#define CFG_NS16550		
#define CFG_NS16550_SERIAL

#define CFG_NS16550_REG_SIZE	1

#define CFG_NS16550_CLK		( CONFIG_DRAM_SPEED * 1000000 ) /* modify by andyw  100Mhz 9600x16x434 */

#define CFG_NS16550_COM1	(CFG_ISA_IO + CFG_NS87308_UART1_BASE)
#define CFG_NS16550_COM2	(CFG_ISA_IO + CFG_NS87308_UART2_BASE)
#define CFG_NS16550_COM3	(CFG_EUMB_ADDR + 0x4500)
#define CFG_NS16550_COM4	(CFG_EUMB_ADDR + 0x4600)

/*
 * Low Level Configuration Settings
 * (address mappings, register initial values, etc.)
 * You should know what you are doing if you make changes here.
 */

#define CONFIG_SYS_CLK_FREQ  33333333	/* external frequency to pll (Oscillator clock is 33MHz)*/

#define CFG_ROMNAL		7	/*rom/flash next access time		 */
#define CFG_ROMFAL		11	/*rom/flash access time			 */

#ifdef M_NH230
  #define CFG_REFINT            480
#else
  #define CFG_REFINT		430	/* no of clock cycles between CBR refresh cycles */
#endif

/* the following are for SDRAM only
 * The following parameters are SDRAM dependent. Please refer to its datasheet for details.
 */
#ifdef M_NH230
  #define CFG_BSTOPRE        121/* Burst To Precharge, sets open page interval */
  #define CFG_REFREC          9	/* Refresh to activate interval (t_RC / CLK Cycle Time) */
  #define CFG_RDLAT           3	/* data latency from read command (W981216BH is 3) */
  #define CFG_PRETOACT        3	/* Precharge to activate interval (t_RP / CLK Cycle Time) */
  #define CFG_ACTTOPRE        6	/* Activate to Precharge interval (t_RAS / CLK Cycle Time) */
  #define CFG_ACTORW          3	/* Activate to R/W (t_RCD / CLK Cycle Time) */
  #define CFG_SDMODE_CAS_LAT  3	/* SDMODE CAS latency (CL) */
  #define CFG_SDMODE_WRAP     0	/* SDMODE wrap type */
#else
  #define CFG_BSTOPRE        121	/* Burst To Precharge, sets open page interval */
  #define CFG_REFREC          8	/* Refresh to activate interval (t_RC / CLK Cycle Time) */
  #define CFG_RDLAT           4	/* data latency from read command (W981216BH is 3) */
  #define CFG_PRETOACT        3	/* Precharge to activate interval (t_RP / CLK Cycle Time) */
  #define CFG_ACTTOPRE        5	/* Activate to Precharge interval (t_RAS / CLK Cycle Time) */
  #define CFG_ACTORW          3	/* Activate to R/W (t_RCD / CLK Cycle Time) */
  #define CFG_SDMODE_CAS_LAT  3	/* SDMODE CAS latency (CL) */
  #define CFG_SDMODE_WRAP	    0	/* SDMODE wrap type */
#endif

#if 0
#define CFG_SDMODE_BURSTLEN 2	/* OBSOLETE!  SDMODE Burst length 2=4, 3=8		*/
#endif

#define CFG_REGISTERD_TYPE_BUFFER   1  /* MCCR4[20] & MCCR4[22] is BUF_TYPE[1:0] */
#define CFG_EXTROM 1   /* MCCR[21] : 1 to enable Extended ROM/Flash (256MB) in Address Map B */
#define CFG_REGDIMM 0  /* MCCR[15] : 0  to Normal DIMMs without delay 1 cycle. */


/* memory bank settings*/
/*
 * only bits 20-29 are actually used from these vales to set the
 * start/end address the upper two bits will be 0, and the lower 20
 * bits will be set to 0x00000 for a start address, or 0xfffff for an
 * end address
 */

#define CFG_BANK0_START		0x00000000
#define CFG_BANK0_END		( ( CFG_MAX_RAM_SIZE/DRAM_BANKS ) - 1 )
#define CFG_BANK0_ENABLE	1

#if ( DRAM_BANKS >= 2 )
#define CFG_BANK1_START		( CFG_MAX_RAM_SIZE/DRAM_BANKS )
#define CFG_BANK1_END		( ( ( CFG_MAX_RAM_SIZE/DRAM_BANKS )*2 ) - 1)
#define CFG_BANK1_ENABLE	1
#else
#define CFG_BANK1_START		0x00000000
#define CFG_BANK1_END		0x00000000
#define CFG_BANK1_ENABLE	0
#endif 

#if ( DRAM_BANKS >= 3 )
#define CFG_BANK2_START		( ( CFG_MAX_RAM_SIZE/DRAM_BANKS )*2 )
#define CFG_BANK2_END		( ( ( CFG_MAX_RAM_SIZE/DRAM_BANKS )*3 ) - 1)
#define CFG_BANK2_ENABLE	1
#else
#define CFG_BANK2_START		0x00000000
#define CFG_BANK2_END		0x00000000
#define CFG_BANK2_ENABLE	0
#endif

#if ( DRAM_BANKS >= 4 )
#define CFG_BANK3_START		( ( CFG_MAX_RAM_SIZE/DRAM_BANKS )*3 )
#define CFG_BANK3_END		( ( ( CFG_MAX_RAM_SIZE/DRAM_BANKS )*4 ) - 1)
#define CFG_BANK3_ENABLE	1
#else
#define CFG_BANK3_START		0x00000000
#define CFG_BANK3_END		0x00000000
#define CFG_BANK3_ENABLE	0
#endif

#if ( DRAM_BANKS >= 5 )
#define CFG_BANK4_START		( ( CFG_MAX_RAM_SIZE/DRAM_BANKS )*4 )
#define CFG_BANK4_END		( ( ( CFG_MAX_RAM_SIZE/DRAM_BANKS )*5 ) - 1)
#define CFG_BANK4_ENABLE	1
#else
#define CFG_BANK4_START		0x00000000
#define CFG_BANK4_END		0x00000000
#define CFG_BANK4_ENABLE	0
#endif


#if ( DRAM_BANKS >= 6 )
#define CFG_BANK5_START		( ( CFG_MAX_RAM_SIZE/DRAM_BANKS )*5 )
#define CFG_BANK5_END		( ( ( CFG_MAX_RAM_SIZE/DRAM_BANKS )*6 ) - 1)
#define CFG_BANK5_ENABLE	1
#else
#define CFG_BANK5_START		0x00000000
#define CFG_BANK5_END		0x00000000
#define CFG_BANK5_ENABLE	0
#endif

#if ( DRAM_BANKS >= 7 )
#define CFG_BANK6_START		( ( CFG_MAX_RAM_SIZE/DRAM_BANKS )*6 )
#define CFG_BANK6_END		( ( ( CFG_MAX_RAM_SIZE/DRAM_BANKS )*7 ) - 1)
#define CFG_BANK6_ENABLE	1
#else
#define CFG_BANK6_START		0x00000000
#define CFG_BANK6_END		0x00000000
#define CFG_BANK6_ENABLE	0
#endif

#if ( DRAM_BANKS >= 8 )
#define CFG_BANK7_START		( ( CFG_MAX_RAM_SIZE/DRAM_BANKS )*7 )
#define CFG_BANK7_END		( ( ( CFG_MAX_RAM_SIZE/DRAM_BANKS )*8 ) - 1)
#define CFG_BANK7_ENABLE	1
#else
#define CFG_BANK7_START		0x00000000
#define CFG_BANK7_END		0x00000000
#define CFG_BANK7_ENABLE	0
#endif

/*
 * Memory bank enable bitmask, specifying which of the banks defined above
 are actually present. MSB is for bank #7, LSB is for bank #0.
 */
#define CFG_BANK_ENABLE		0x01

#ifdef M_NH230
  /* PMCR1 : Table 4-17 of User Manual */
  #define CFG_PMCR1_CKO_MODE_MASK  0x0006   /* bit[2:1] CKO_MODE */
  
  /* PMCR2 : Table 4-18 of User Manual */
  #define CFG_PMCR2_DLL_EXTEND     0x80     /* bit[7] DLL_EXTEND */
  #define CFG_PMCR2_PCI_HOLD_DEL_MASK   0x30  /* bit[5:4] PCI_HOLD_DEL bits */
  #define CFG_PMCR2_PCI_HOLD_DEL_33MHZ  0x20  /* bit[5:4] PCI_HOLD_DEL is 33MHz PCI */

  /* DRV_PCI (bit[7]=0) : 20 ohm, DRV_MEM_CTRL(bit[5:4]=01) L 40 ohm     */
  /* DRV_PCI_CLK(bit[3:2]=10) : 20 ohm, DRV_MEM_CLK(bit1:0]=01) : 40 ohm */
  #define CFG_ODCR              0x19    /* bit[6] must be preserved */
  #define CFG_ODCR_PRESERVED    0x40  
  
  /* CLKDCR : Table 4-20 of User Manual */
  #define CFG_CLKDCR_PCI_CLK_MASK   0x7c00  /* bit[14:10] PCI_CLKx_DIS bits */
  #define CFG_CLKDCR_PCI_CLK4_DIS   0x0400  /* bit[10] PCI_CLK4_DIS */
  #define CFG_CLKDCR_SDRAM_CLK_MASK 0x0078  /* bit[6:3] SDRAM_CLKx_DIS bits */
  #define CFG_CLKDCR_SDRAM_CLK3_DIS 0x0008  /* bit[3] SDRAM_CLK3_DIS */
  
  /* MIOCR1 : Table 4-21 of User Manual */
  #define CFG_MIOCR1_DLL_MAX_DELAY   0x04  /* bit[2] of MIOCR1 : DLL_MAX_DELAY */
  #define CFG_MIOCR1_CLK_FLIP        0x02  /* bit[1] of MIOCR1 : CLK_FLIP */
#else
  #define CFG_ODCR		0xff	/* configures line driver impedances,	*/
					/* see 8240 book for bit definitions	*/
#endif

#define CFG_PGMAX		0x32	/* how long the 8240 retains the	*/
					/* currently accessed page in memory	*/
					/* see 8240 book for details		*/

/* SDRAM 0 - 256MB */
#define CFG_IBAT0L	(CFG_SDRAM_BASE | BATL_PP_10 | BATL_MEMCOHERENCE)
#define CFG_IBAT0U	(CFG_SDRAM_BASE | BATU_BL_256M | BATU_VS | BATU_VP)

/* stack in DCACHE @ 1GB (no backing mem) */
#if defined(USE_DINK32)
#define CFG_IBAT1L	(0x40000000 | BATL_PP_00 )
#define CFG_IBAT1U	(0x40000000 | BATU_BL_128K )
#else
#define CFG_IBAT1L	(CFG_INIT_RAM_ADDR | BATL_PP_10 | BATL_MEMCOHERENCE)
#define CFG_IBAT1U	(CFG_INIT_RAM_ADDR | BATU_BL_128K | BATU_VS | BATU_VP)
#endif

/* PCI memory */
#define CFG_IBAT2L	(0x80000000 | BATL_PP_10 | BATL_CACHEINHIBIT)
#define CFG_IBAT2U	(0x80000000 | BATU_BL_256M | BATU_VS | BATU_VP)

/* Flash, config addrs, etc */
#define CFG_IBAT3L	(0xF0000000 | BATL_PP_10 | BATL_CACHEINHIBIT)
#define CFG_IBAT3U	(0xF0000000 | BATU_BL_256M | BATU_VS | BATU_VP)

#define CFG_DBAT0L	CFG_IBAT0L
#define CFG_DBAT0U	CFG_IBAT0U
#define CFG_DBAT1L	CFG_IBAT1L
#define CFG_DBAT1U	CFG_IBAT1U
#define CFG_DBAT2L	CFG_IBAT2L
#define CFG_DBAT2U	CFG_IBAT2U
#define CFG_DBAT3L	CFG_IBAT3L
#define CFG_DBAT3U	CFG_IBAT3U

/*
 * For booting Linux, the board info and command line data
 * have to be in the first 8 MB of memory, since this is
 * the maximum mapped by the Linux kernel during initialization.
 */
#define CFG_BOOTMAPSZ		(8 << 20)	/* Initial Memory map for Linux */
/*-----------------------------------------------------------------------
 * FLASH organization
 */


#define CFG_FLASH_ERASE_TOUT	120000	/* Timeout for Flash Erase (in ms)	*/
#define CFG_FLASH_WRITE_TOUT	500	/* Timeout for Flash Write (in ms)	*/

/*-----------------------------------------------------------------------
 * Cache Configuration
 */
#define CFG_CACHELINE_SIZE	32	/* For MPC8240 CPU			*/
#if (CONFIG_COMMANDS & CFG_CMD_KGDB)
#  define CFG_CACHELINE_SHIFT	5	/* log base 2 of the above value */
#endif


/*
 * Internal Definitions
 *
 * Boot Flags
 */
#define BOOTFLAG_COLD		0x01	/* Normal Power-On: Boot from FLASH	*/
#define BOOTFLAG_WARM		0x02	/* Software reboot			*/


/* values according to the manual */

#define CONFIG_DRAM_50MHZ	1
#define CONFIG_SDRAM_50MHZ

#undef	NR_8259_INTS
#define NR_8259_INTS		1


#define CONFIG_DISK_SPINUP_TIME 1000000


#endif	/* __CONFIG_H */
