#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/vfs.h>
#include <errno.h>
#include <dirent.h>
#include <fcntl.h>

#define MAX_TORRENT 10
typedef struct{
    char filename[4096];
    char dirname[4096];
    char hours[3];
    char port[6];
    char max_peer[3];
    char min_peer[3];
    char max_bw[8];
    char min_bw[8];
    char time[21];
    char finish[2];
}CONFIGURATION_DATA;

char TCONFIG_FOLDER[17];
char TCONFIG_FILENAME[17];
char TORRENT_USER[3];
char TIME_ZONE[10];
CONFIGURATION_DATA queue_record[MAX_TORRENT];

#define DEBUG
#undef DEBUG
#ifdef DEBUG
#define MSG(string,args...) fprintf(stderr,string,##args)
#else
#define MSG(string,args...)
#endif

int read_p2p_config(void);
void read_etc_config(void);
void stop_btclnt_process(int count);
void init_btclnt_process(int count);

int main(int argc, char *argv[]){
	
	int count = 0;
	
	memset(TCONFIG_FOLDER,0,17);
	memset(TCONFIG_FILENAME,0,17);
	    
	//Read configuration to get all the necessary parameters    
	read_etc_config(); 	
	count = read_p2p_config();   
	MSG("%s: Line%d - count = %d\n",__func__,__LINE__,count);	
	// -del: kill the P2P process, -init: start the P2P process
	if(argc > 1 && !strcmp(argv[1],"-del"))
	    stop_btclnt_process(count);
	if(argc > 1 && !strcmp(argv[1],"-init"))
	    init_btclnt_process(count);
	return 1;
}

//To read /etc/config to get the P2P configuration filename
void read_etc_config(void){
	
	FILE *fp;
	char *ptr, *end;
	char LineBuf[256];
	
	memset(TIME_ZONE,0,10);
	memset(TCONFIG_FOLDER,0,17);
	memset(TCONFIG_FILENAME,0,17);
	memset(TORRENT_USER,0,3);
	
	if((fp = fopen("/etc/config","r")) == NULL){
		printf("Can't open config to read.\n");
		return;
	}else{
	    while(fgets(LineBuf,255,fp) != NULL){
	    	if((ptr = strstr(LineBuf,"TIME_ZONE='")) != NULL){
		        ptr+=11;
		        end = strchr(ptr,'\'');
		    	*end = '\0';
		        strcpy(TIME_ZONE, ptr);
		    }
	    	if((ptr = strstr(LineBuf,"TCONFIG_FOLDER='")) != NULL){
		        ptr+=16;
		        end = strchr(ptr,'\'');
		    	*end = '\0';
		        strcpy(TCONFIG_FOLDER, ptr);
		        MSG("%s: Line%d - TCONFIG_FOLDER = %s\n",__func__,__LINE__,TCONFIG_FOLDER);
		    }
		    if((ptr = strstr(LineBuf,"TCONFIG_FILENAME='")) != NULL){
		        ptr+=18;
		        end = strchr(ptr,'\'');
		    	*end = '\0';
		        strcpy(TCONFIG_FILENAME, ptr);
		        MSG("%s: Line%d - TCONFIG_FILENAME = %s\n",__func__,__LINE__,TCONFIG_FILENAME);
		    }
		    if((ptr = strstr(LineBuf,"TORRENT_USER='")) != NULL){
		        ptr+=14;
		        end = strchr(ptr,'\'');
		    	*end = '\0';
		        strcpy(TORRENT_USER, ptr);
		        MSG("%s: Line%d - TORRENT_USER = %s\n",__func__,__LINE__,TORRENT_USER);
		    }
		    if(strlen(TIME_ZONE) && strlen(TCONFIG_FOLDER) && strlen(TCONFIG_FILENAME) && strlen(TORRENT_USER))
		        break;
		}
		fclose(fp);
	}
	return;
}

//To read P2P configuration file
int read_p2p_config(void){

    FILE *fp;
    //char *ptr, *end;
    char *ptr;
    int no_queue = 0;
    char LineBuf[256], filename[100];
    
	strcpy(filename,"/mnt/");
	strcat(filename,TCONFIG_FOLDER);
	strcat(filename,"/.");
	strcat(filename,TCONFIG_FILENAME);
	MSG("%s: Line%d - filename = %s\n",__func__,__LINE__,filename);
	
	if((fp = fopen(filename,"r")) == NULL){
		printf("Can't open %s to read.\n",filename);
		return 0;
	}else{
	    while(fgets(LineBuf,255,fp) != NULL){
	    	ptr = strstr(LineBuf,"TORRENT_RECORD");
	    	if(ptr == NULL)
	            continue;
	    	ptr = strchr(ptr,'=');
	    	ptr++;
	    	ptr = strtok(ptr,",");
	    	if(ptr == NULL)
	    	    continue;
	    	strcpy(queue_record[no_queue].filename,ptr);
    		MSG("queue_record[%d].filename = %s\n",no_queue,queue_record[no_queue].filename);
    		ptr = strtok(NULL,",");
    		strcpy(queue_record[no_queue].dirname,ptr);
    		MSG("queue_record[%d].dirname = %s\n",no_queue,queue_record[no_queue].dirname);
    		ptr = strtok(NULL,",");
    		strcpy(queue_record[no_queue].hours,ptr);
    		MSG("queue_record[%d].hours = %s\n",no_queue,queue_record[no_queue].hours);
    		ptr = strtok(NULL,",");
    		strcpy(queue_record[no_queue].port,ptr);
    		MSG("queue_record[%d].port = %s\n",no_queue,queue_record[no_queue].port);
    		ptr = strtok(NULL,",");
    		strcpy(queue_record[no_queue].max_peer,ptr);
    		MSG("queue_record[%d].max_peer = %s\n",no_queue,queue_record[no_queue].max_peer);
    		ptr = strtok(NULL,",");
    		strcpy(queue_record[no_queue].min_peer,ptr);
    		MSG("queue_record[%d].min_peer = %s\n",no_queue,queue_record[no_queue].min_peer);
    		ptr = strtok(NULL,",");
    		strcpy(queue_record[no_queue].max_bw,ptr);
    		MSG("queue_record[%d].max_bw = %s\n",no_queue,queue_record[no_queue].max_bw);
    		ptr = strtok(NULL,",");
    		strcpy(queue_record[no_queue].min_bw,ptr);
    		MSG("queue_record[%d].min_bw = %s\n",no_queue,queue_record[no_queue].min_bw);
    		strcpy(queue_record[no_queue].finish,"0");
    		no_queue++;
    		if(no_queue == MAX_TORRENT)
    		    break;
    	}
    	fclose(fp);
    }
    return no_queue;
}
	    	    
void stop_btclnt_process(int count){
	
	FILE *fp, *fo;
	char *ptr, *end;
	char piece_receive[11], piece_file[11];
    int i, finish = 0, queue_num = 0, del = 0;
	char filename[200], command[100], LineBuf[256], DataBuf[1024];
	
	//kill all the P2P daemon
	system("killall bt_clnt");
	//rewrite xxx.torrent.status and check if any download is completed
	for(i=0;i<count;i++){
		strcpy(filename,queue_record[i].filename);
		strcat(filename,".status");
		MSG("%s: Line%d - filename = %s\n",__func__,__LINE__,filename);
		if((fp = fopen(filename,"r")) != NULL){
			fo = fopen("/var/run/change_status","w+");
			while(fgets(LineBuf,255,fp)){
				if((ptr = strstr(LineBuf,"NO# of pieces received :")) != NULL){
					ptr += 25;
					end = strchr(ptr,'\n');
					*end = '\0';
					strcpy(piece_receive,ptr);
					MSG("%s: Line %d - piece_receive = %s\n",__func__,__LINE__,piece_receive);
					fputs(LineBuf,fo);
			    }else if((ptr = strstr(LineBuf,"NO# of pieces of file :")) != NULL){
			    	ptr += 24;
			    	end = strchr(ptr,'\n');
			    	*end = '\0';
			    	strcpy(piece_file,ptr);
			    	MSG("%s: Line %d - piece_file = %s\n",__func__,__LINE__,piece_file);
			    	fputs(LineBuf,fo);
			    	if(strlen(piece_receive) && strlen(piece_file)){
			    		if(!strcmp(piece_file,piece_receive)){
			    			finish = 1;    // download is completed
			    			strcpy(queue_record[i].finish,"1");
			    		}
			    	}
			    }else if((ptr = strstr(LineBuf,"Status     :")) != NULL){
		        	fprintf(fo,"Status     : done\n");
		        }else if((ptr = strstr(LineBuf,"Check existing pieces :")) != NULL){
		        	fprintf(fo,"Status     : done\n");
		        }else{
		            fputs(LineBuf,fo);
		        }
			}
			fclose(fp);
			fclose(fo);
			fo = fopen(filename,"w+");
			fp = fopen("/var/run/change_status","r");
			if((fp != NULL) && (fo != NULL)){
            	while(fgets(LineBuf,255,fp))
            		fputs(LineBuf,fo);
            	fclose(fp);
            	fclose(fo);
            }
            system("/bin/rm -f /var/run/change_status");
		}
	}
	//If any download is finished, rewrite the P2P configuration
	if(finish == 1){
		strcpy(filename,"/mnt/");
		strcat(filename,TCONFIG_FOLDER);
		strcat(filename,"/.");
		strcat(filename,TCONFIG_FILENAME);
		MSG("%s: Line%d - filename = %s\n",__func__,__LINE__,filename);
		if((fp = fopen(filename,"r")) == NULL)
		    return;
		if((fo = fopen("/var/run/finish_status","w+")) == NULL)
			return;
		else{
		    while(fgets(DataBuf,1023,fp)){
		    	if(strstr(DataBuf,"TORRENT_RECORD")){
		    	    ptr = strchr(DataBuf,'=');
		    	    ptr++;
		    	    del = 0;
		    	    for(i=0;i<count;i++){
		    		    if(strstr(DataBuf,queue_record[i].filename) && !strcmp(queue_record[i].finish,"1")){
		    		    	del = 1;
		    		        break;
		    		    }
		    		}
		    		if(del != 1){
		    			queue_num++;
		    			fprintf(fo,"TORRENT_RECORD%d=%s\n",queue_num,ptr);
		    		}
		    	}
		    }
		    fclose(fp);
		    fclose(fo);
		    sprintf(command,"/bin/mv /var/run/finish_status %s",filename);
		    system(command);
		}
	}							
}

void init_btclnt_process(int count){
	
	FILE *fi, *fp;
	char *ptr, *end;
	int i, mark = 0, checked = 0, limit = 0;
	char command[1024], LineBuf[1024], filename[100];
	
	strcpy(filename,"/mnt/");
	strcat(filename,TCONFIG_FOLDER);
	strcat(filename,"/.");
	strcat(filename,TCONFIG_FILENAME);
	MSG("%s: Line%d - filename = %s\n",__func__,__LINE__,filename);
	limit = atoi(TORRENT_USER);
	MSG("%s: Line%d - limit = %d, count = %d\n",__func__,__LINE__,limit,count);
	
	//Excute the command to run P2P
	for(i=0;i<count && limit>0;i++,limit--){
		if(strlen(queue_record[i].filename) && strlen(queue_record[i].dirname)){
			sprintf(command,"/mnt/bt_clnt -s %s -e %s -p %s -M %s -m %s -Y %s -Z %s \"%s\" &",queue_record[i].dirname,queue_record[i].hours,queue_record[i].port,queue_record[i].max_peer,queue_record[i].min_peer,queue_record[i].max_bw,queue_record[i].min_bw,queue_record[i].filename);
			MSG("%s: Line %d - command = %s\n",__func__,__LINE__,command);
	        system(command);
	        sprintf(command,"TZ=%s date \"+%%D %%T\" > /var/run/excute_time",TIME_ZONE);
	        MSG("%s: Line %d - command = %s\n",__func__,__LINE__,command);
	        system(command);
	        if((fi = fopen("/var/run/excute_time","r")) != NULL){
		        while(fgets(command,1023,fi))
		            strncpy(queue_record[i].time,command,20);
		        queue_record[i].time[20] = '\0';
	            fclose(fi);
	        }
	    }
    }
    //Record time in P2P configuration     
    if((fp = fopen(filename,"r")) == NULL)
        return;
    if((fi = fopen("/var/run/time_status","w+")) == NULL){
        return;
    }else{
        while(fgets(LineBuf,1023,fp)){
            checked = 0;
	        if((ptr = strstr(LineBuf,"TORRENT_RECORD")) != NULL){
		        ptr = strchr(LineBuf,'=');
		        ptr++;
		        for(i=0;i<count;i++){
			        if(strstr(ptr,queue_record[i].filename) && strstr(ptr,queue_record[i].dirname) && strlen(queue_record[i].time)){
			        	MSG("%s: Line %d - queue_record[%d].filename = %s, queue_record[%d].dirname = %s\n",__func__,__LINE__,i,queue_record[i].filename,i,queue_record[i].dirname);
			            checked = 1;
			            mark = i;
			            break;
			        }
		        }
		        if(checked == 1){
		        	end = strrchr(LineBuf,',');
		        	*end = '\0';
			        MSG("%s: Line%d - %s,%s\n",__func__,__LINE__,LineBuf,queue_record[mark].time);
			        fprintf(fi,"%s,%s",LineBuf,queue_record[mark].time);
		        }else if(checked == 0){
		            MSG("%s: Line%d - LineBuf = %s\n",__func__,__LINE__,LineBuf);
		            fprintf(fi,"%s",LineBuf);
		        }
	        }
        }
        fclose(fp);
	    fclose(fi);
        sprintf(command,"/bin/mv /var/run/time_status /mnt/%s/.%s",TCONFIG_FOLDER,TCONFIG_FILENAME);
        MSG("%s: Line%d - command = %s\n",__func__,__LINE__,command);
        system(command);
    }
}	
