/*
 * Copyright (c) 1985, 1988, 1990, 1992, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* #ifndef lint */
/* static char copyright[] = */
/* "@(#) Copyright (c) 1985, 1988, 1990, 1992, 1993, 1994\n\ */
/* 	The Regents of the University of California.  All rights reserved.\n"; */
/* #endif */ /* not lint */

/* #ifndef lint */
/* static char sccsid[] = "@(#)ftpd.c	8.5 (Berkeley) 4/28/95"; */
/* #endif */ /* not lint */

/*
 * FTP server.
 */
 
//
// History :
//   V1.00  
//     1) The original source codes of FTP server.
//
//   V1.01  04/16/2004  Hank Chu
//     1) Modify to limit the maximum NO# of FTP daemons (ftpd) forked out to support the limited FTP connections.
//        Use the "M_FTP_MAX=xxx" macro defined in other place to maximum NO# of FTP connections.
//        It will store the current NO# of ftpd daemon(s) forked in the "/var/run/ftpd_num" file.
//
//        Add 1 if ftpd is forked, sub 1 if ftpd is closed.
//        These codes are enabled by the "M_NH200" macro defined in the file.
//
//     2) Reduce the code size and memory used to save the free memory if ftpd is forked out.
//        a) Use the "M_FTPD_DEBUG" to disable the support to debug ftpd.
//        b) Use the "M_FTPD_FULL" to disable the useless codes.
//
//
//
//

//#define M_NH200     // define it to patch for ftpd//
//sp/ for temporarily added to make all function in sync.
#ifndef ALLNET
# define ALLNET
#endif

extern void yyparse();
extern int ftpdconf(void);
extern int Get_IP_Security(char *check_ip);
extern int sec_right(char *directory,char *user, int flag);

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#ifdef __STDC__
#include <stdarg.h>
#define va_strt      va_start
#else
#include <varargs.h>
#define va_strt(p,i) va_start(p)
#endif

#if !defined (__GNUC__) && defined (_AIX)
 #pragma alloca
#endif
#ifndef alloca /* Make alloca work the best possible way.  */
#ifdef __GNUC__
#define alloca __builtin_alloca
#else /* not __GNUC__ */
#if HAVE_ALLOCA_H
#include <alloca.h>
#else /* not __GNUC__ or HAVE_ALLOCA_H */
#ifndef _AIX /* Already did AIX, up at the top.  */
char *alloca ();
#endif /* not _AIX */
#endif /* not HAVE_ALLOCA_H */
#endif /* not __GNUC__ */
#endif /* not alloca */

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/wait.h>

#include <netinet/in.h>
#ifdef HAVE_NETINET_IN_SYSTM_H
#include <netinet/in_systm.h>
#endif
#ifdef HAVE_NETINET_IP_H
#include <netinet/ip.h>
#endif

#define	FTP_NAMES
#include <arpa/ftp.h>
#include <arpa/inet.h>
#include <arpa/telnet.h>

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <getopt.h>
#include <netdb.h>
#include <pwd.h>
#include <grp.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>

#include <unistd.h>

#ifdef M_FTPD_DEBUG
#include <syslog.h>

//#include<../libc/misc/dirent/dirstream.h>

static FILE *M4_fd = NULL;
void M4_printf(const char *fmt, ...);
#define M4_syslog(pri, fmt...)	{M4_printf(fmt);}
#else
	#define M4_syslog(pri, fmt...)
#endif

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#include <crypt.h>
/* Include glob.h last, because it may define "const" which breaks
   system headers on some platforms. */
#include <glob.h>

#include "extern.h"

#ifdef ANDY_ICONV
#include "iconv.h"
//#include "iconv.h"
#endif

/*
#if (defined(LOBOS) || defined(ALLNET))	
	int M_FTP_MAX;	
#else
	#include "../../include/maxuser.h"
#endif
*/

#ifndef LINE_MAX
	#ifdef M_FTPD_FULL
#define LINE_MAX 2048
	#else
		#define LINE_MAX 80        // reduce to maximum 80 characters per line
	#endif
#endif

#ifndef LOG_FTP
#define LOG_FTP LOG_DAEMON	/* Use generic facility.  */
#endif

#ifndef HAVE_FCLOSE_DECL
/* Some systems don't declare fclose in <stdio.h>, so do it ourselves.  */
extern int fclose __P ((FILE *));
#endif

extern	off_t restart_point;
extern	char cbuf[];
extern	int root_flag;

struct	sockaddr_in ctrl_addr;
struct	sockaddr_in data_source;
struct	sockaddr_in data_dest;
struct	sockaddr_in his_addr;
struct	sockaddr_in pasv_addr;

int	data;
jmp_buf	errcatch, urgcatch;
int	logged_in;
struct	passwd *pw;
// andy add for judge if log fail
int log_fail=0;

int root_folder=0;

//#ifdef M_FTPD_DEBUG
int	debug;
//#endif

#if (defined(LOBOS) || defined(ALLNET))	
	int timeout;  // we will parser it from config file latter.
#else	
	int	timeout = 900;    /* timeout after 15 minutes of inactivity */
#endif

int	maxtimeout = 7200;/* don't allow idle time to be set beyond 2 hours */
int	logging;
int	guest;
int	type;
int	form;
int	stru;			/* avoid C keyword */
int	mode;
int	usedefault = 1;		/* for data transfers */
int	pdata = -1;		/* for passive mode */
sig_atomic_t transflag;
off_t	file_size;
off_t	byte_count;
#if !defined(CMASK) || CMASK == 0
#undef CMASK
#define CMASK 027
#endif
int	defumask = CMASK;		/* default umask value */
char	tmpline[7];
char	*hostname = 0;
char	*remotehost = 0;

#define NUM_SIMUL_OFF_TO_STRS 4

#ifdef ANDY_ICONV
char CLIENT_CHARSET[10]="CP";
char UTF8[6]="UTF-8";
#endif

#define _BUFSIZ 1024*32

// Ansel modify the error due to getcwd
#define PATH_LENGTH 1024

/*
#ifdef M_NH200
  // The name of file to store the current NO# of ftpd forked out by indetd
  #define STR_FTPD_NUM_FILE    "/var/run/ftpd_num"  

////  #define  M_FTP_MAX  2   //10
  
//  unsigned int ftpd_get_active_connections(void);
// int ftpd_update_active_connections(unsigned char b_flag);
#endif
*/
int GetFullPathName(char *path);

/* Returns a string with the decimal representation of the off_t OFF, taking
   into account that off_t might be longer than a long.  The return value is
   a pointer to a static buffer, but a return value will only be reused every
   NUM_SIMUL_OFF_TO_STRS calls, to allow multiple off_t's to be conveniently
   printed with a single printf statement.  */
static char *
off_to_str (off)
     off_t off;
{
  static char bufs[NUM_SIMUL_OFF_TO_STRS][80];
  static char (*next_buf)[80] = bufs;

  if (next_buf > (bufs+NUM_SIMUL_OFF_TO_STRS))
    next_buf = bufs;

#if defined(__USE_FILE_OFFSET64) && defined(__USE_FILE_OFFSET64)
  sprintf (*next_buf, "%Ld", off);
#else
  sprintf (*next_buf, "%ld", off);
#endif

  return *next_buf++;
}

/*
 * Timeout intervals for retrying connections
 * to hosts that don't accept PORT cmds.  This
 * is a kludge, but given the problems with TCP...
 */
#define	SWAITMAX	90	/* wait at most 90 seconds */
#define	SWAITINT	5	/* interval between retries */

int	swaitmax = SWAITMAX;
int	swaitint = SWAITINT;

#ifdef M_FTPD_DEBUG
#define LOGCMD(cmd, file) \
	if (logging > 1) { \
	    M4_syslog(LOG_INFO,"%s %s%s", cmd, \
		*(file) == '/' ? "" : curdir(), file);}
#define LOGCMD2(cmd, file1, file2) \
	 {if (logging > 1) { \
	    M4_syslog(LOG_INFO,"%s %s%s %s%s", cmd, \
		*(file1) == '/' ? "" : curdir(), file1, \
		*(file2) == '/' ? "" : curdir(), file2); }}
#define LOGBYTES(cmd, file, cnt) \
	{if (logging > 1) { \
		if (cnt == (off_t)-1) {\
		    M4_syslog(LOG_INFO,"%s %s%s", cmd, \
			*(file) == '/' ? "" : curdir(), file); \
		} else {\
		    M4_syslog(LOG_INFO, "%s %s%s = %s bytes", \
			cmd, (*(file) == '/') ? "" : curdir(), file, \
			   off_to_str (cnt)); } \
	}}
#else
	#define LOGCMD(cmd, file)
	#define LOGCMD2(cmd, file1, file2)
	#define LOGBYTES(cmd, file, cnt)
#endif
	
	
static void	 ack __P((char *));
static void	 myoob __P((int));
static int	 checkuser __P((char *));
static FILE	*dataconn __P((char *, off_t, char *));
static void	 dolog __P((struct sockaddr_in *));
#ifdef M_FTPD_DEBUG
static char	*curdir __P((void));
#endif
static void	 end_login __P((void));
static FILE	*getdatasock __P((char *));
static char	*gunique __P((char *));
static void	 lostconn __P((int));
static int	 receive_data __P((FILE *, FILE *));
static void	 send_data __P((FILE *, FILE *, off_t));
static struct passwd *
		 sgetpwnam __P((char *));
static char	*sgetsave __P((char *));
//from ls.c
static char *modestring __P((int));
static char *timestring __P((long));

#ifdef M_FTPD_DEBUG

static char *
curdir()
{
        static char *path = 0;
	if (path)
	        free (path);
	path = getcwd (0, PATH_LENGTH);
	if (! path)
		return ("");
	if (path[1] != '\0') {	/* special case for root dir. */
	        char *new = realloc (path, strlen (path) + 2); /* '/' + '\0' */
		if (! new)
		        return "";
		strcat(new, "/");
		path = new;
	}
	/* For guest account, skip / since it's chrooted */
	return (guest ? path+1 : path);
}
#endif

#ifdef M_FTPD_FULL
static char main_line[LINE_MAX];
#endif

char UserName[32];
static char localhost[256];
int
main(argc, argv)
	int argc;
	char *argv[];
{
	int addrlen, ch, on = 1;

//#error	#ifdef sp_no_cp_init
	static have_init_iconv=0;
	

		have_init_iconv=1;
		sp_iconv_init();

#if (defined(LOBOS) || defined(ALLNET))	
    FILE *fip;
    char LineBuf[256],maxftp[3],ftptimeout[5];
    char *ptr, *end;
    int i;
#endif
    
#ifdef Samba_security	
 //== Hiram ==		
//	while(access("/mnt/ftpd.conf",R_OK))
//	{
		ftpdconf();
//	}
#endif	

#if 0
	int tos;
#endif	/* 0 */

#ifdef M_FTPD_DEBUG
//	char *cp;
//	FILE *fd;
#endif

#if (defined(LOBOS) || defined(ALLNET))
/*
// parser for max ftp user number
if((fip = fopen("/etc/config","r")) == NULL){
        return 0;
    }else{
        while(fgets(LineBuf,255,fip) != NULL){
            for(i=0;i<255;i++){
                if((LineBuf[i] == ' ') || (LineBuf[i] == '\n') || (LineBuf[i] == '\r')){
                    LineBuf[i] = '\0';
                    i=255;
                }
            }
            if((ptr = strstr(LineBuf,"FTP_MAX='")) != NULL){
             ptr+=9;
             end = strchr(ptr,'\'');
             *end = '\0';
             strcpy(maxftp,ptr);
            }
        }
     }
      
   M_FTP_MAX=atoi(maxftp);
//   HDBG("M_FTP_MAX=%d\n",M_FTP_MAX);   
//end 
*/
// parser for max ftp connect time
if((fip = fopen("/etc/config","r")) == NULL){
        return 0;
    }else{
        while(fgets(LineBuf,255,fip) != NULL){
            for(i=0;i<255;i++){
                if((LineBuf[i] == ' ') || (LineBuf[i] == '\n') || (LineBuf[i] == '\r')){
                    LineBuf[i] = '\0';
                    i=255;
                }
            }
            if((ptr = strstr(LineBuf,"FTP_TIMEOUT='")) != NULL){
             ptr+=13;
             end = strchr(ptr,'\'');
             *end = '\0';
             strcpy(ftptimeout,ptr);
            }
            
		 	if((ptr = strstr(LineBuf,"CODE_SET='")) != NULL){
		 		ptr+=strlen("CODE_SET='");
		 		if( (*ptr) != '\''){
	                end = strchr(ptr,'-');
	                *end = '\0';
	                strcat(CLIENT_CHARSET,ptr);
	                CLIENT_CHARSET[strlen(CLIENT_CHARSET)]='\0';
	                M4_syslog(LOG_DEBUG,"ptr=%s,CLIENT_CHARSET=%s\n",ptr,CLIENT_CHARSET);
	                break;
	            }
			}	            
        }
     }
      
   timeout=atoi(ftptimeout)*60;
 //  HDBG("timeout=%d\n",timeout);   
//end 
#endif    // end of #if (defined(LOBOS) || defined(ALLNET))

/*
#ifdef M_NH200
	// 1) Get the NO# of active FTP connection(s) from file
	unsigned int dw_FTP_connections = ftpd_get_active_connections();
	if (M_FTP_MAX <= dw_FTP_connections) {  // exceeds maximum ?
//		M4_syslog(LOG_WARNING, "ftpd: NO# of FTP connections (%d) exceeds maximum (%d)", 
//					dw_FTP_connections, M_FTP_MAX);
		
		reply(421, "Exceeds maximum number of ftp users");
		exit(1);
	}
#endif
*/	
	
	/*
	 * LOG_NDELAY sets up the logging connection immediately,
	 * necessary for anonymous ftp's that chroot and can't do it later.
	 */
#if 0
	openlog("ftpd", LOG_PID | LOG_NDELAY, LOG_FTP);
#endif	/* 0 */

	addrlen = sizeof(his_addr);
	if (getpeername(0, (struct sockaddr *)&his_addr, &addrlen) < 0) {
		M4_syslog(LOG_ERR, "getpeername (%s): %m",argv[0]);
		exit(1);
	}

	addrlen = sizeof(ctrl_addr);
	if (getsockname(0, (struct sockaddr *)&ctrl_addr, &addrlen) < 0) {
		M4_syslog(LOG_ERR, "getsockname (%s): %m",argv[0]);
		exit(1);
	}

//== Modify by George for check IP & MAC filter ====
//#if (defined(LOBOS) || defined(NOVAC))
	
	char ipaddr[32];
	sprintf(ipaddr, "%d.%d.%d.%d", his_addr.sin_addr.s_addr >> 24, (his_addr.sin_addr.s_addr >> 16) & 0xFF,
		(his_addr.sin_addr.s_addr >> 8) & 0xFF, his_addr.sin_addr.s_addr & 0xFF);
	if (Get_IP_Security(ipaddr) != 0) {
		reply(426, "IP/MAC Address Access Denied.");
		_exit(-1);
	}
//#endif
//==================================

#if defined (IP_TOS) && defined (IPTOS_LOWDELAY) && defined (IPPROTO_IP)
	tos = IPTOS_LOWDELAY;
	if (setsockopt(0, IPPROTO_IP, IP_TOS, (char *)&tos, sizeof(int)) < 0)
		M4_syslog(LOG_WARNING, "setsockopt (IP_TOS): %m");
#endif
	data_source.sin_port = htons(ntohs(ctrl_addr.sin_port) - 1);

	
#ifdef M_FTPD_DEBUG
	debug = 1;
#endif

	while ((ch = getopt(argc, argv, "dlt:T:u:v")) != EOF) {
		switch (ch) {
		case 'd':
#ifdef M_FTPD_DEBUG
			debug = 1;
#endif
			break;

		case 'l':
			logging++;	/* > 1 == extra logging */
			break;

		case 't':
			timeout = atoi(optarg);
			if (maxtimeout < timeout)
				maxtimeout = timeout;
			break;

		case 'T':
			maxtimeout = atoi(optarg);
			if (timeout > maxtimeout)
				timeout = maxtimeout;
			break;

		case 'u':
		    {
			long val = 0;

			val = strtol(optarg, &optarg, 8);
			if (*optarg != '\0' || val < 0)
				fprintf (stderr,
					 "%s: bad value for -u", argv[0]);
			else
				defumask = val;
			break;
		    }

		case 'v':
#ifdef M_FTPD_DEBUG
			debug = 1;
#endif
			break;

		default:
			fprintf (stderr,
				 "%s: unknown flag -%c ignored", argv[0], ch);
			break;
		}
	}
		
	(void) freopen("/dev/null", "w", stderr);
	(void) signal(SIGPIPE, lostconn);
	(void) signal(SIGCHLD, SIG_IGN);
	if ((int)signal(SIGURG, myoob) < 0)
		M4_syslog(LOG_ERR, "signal: %m");

	/* Try to handle urgent data inline */
#ifdef SO_OOBINLINE
	if (setsockopt(0, SOL_SOCKET, SO_OOBINLINE, (char *)&on, sizeof(on)) < 0)
		M4_syslog(LOG_ERR, "setsockopt: %m");
#endif

#ifdef	F_SETOWN
	if (fcntl(fileno(stdin), F_SETOWN, getpid()) == -1)
		M4_syslog(LOG_ERR, "fcntl F_SETOWN: %m");
#endif
	dolog(&his_addr);
	/*
	 * Set up default state
	 */
	data = -1;
	type = TYPE_A;
	form = FORM_N;
	stru = STRU_F;
	mode = MODE_S;
	tmpline[0] = '\0';

#ifdef M_FTPD_FULL
	/* If logins are disabled, print out the message. */
	if ((fd = fopen("/etc/nologin","r")) != NULL) {
		while (fgets(main_line, sizeof(main_line), fd) != NULL) {
			if ((cp = strchr(main_line, '\n')) != NULL)
				*cp = '\0';
			lreply(530, "%s", main_line);
		}
		(void) fflush(stdout);
		(void) fclose(fd);
		reply(530, "System not available.");
		exit(0);
	}
	if ((fd = fopen("/etc/ftpwelcome", "r")) != NULL) {
		while (fgets(main_line, sizeof(main_line), fd) != NULL) {
			if ((cp = strchr(main_line, '\n')) != NULL)
				*cp = '\0';
			lreply(220, "%s", main_line);
		}
		(void) fflush(stdout);
		(void) fclose(fd);
		/* reply(220,) must follow */
	}
#endif
/*	
#ifdef M_NH200
	// Increase the NO# of active FTP connection(s) into file
	ftpd_update_active_connections(1);
#endif
*/	
	
	if (gethostname(localhost, sizeof(localhost)) < 0) {
	  strcpy(localhost, "NH200");
	}
	hostname = localhost;
/* 	if (! hostname) */
/* 		perror_reply (550, "Local resource failure: malloc"); */

	reply(220, "%s FTP server ready.", hostname);
	(void) setjmp(errcatch);
	for (;;)
		(void) yyparse();
	/* NOTREACHED */
}

static void
lostconn(signo)
	int signo;
{
#ifdef M_FTPD_DEBUG
	if (debug)
		M4_syslog(LOG_DEBUG, "lost connection");
#endif
	dologout(-1);
}

static char ttyline[20];

/*
 * Helper function for sgetpwnam().
 */
static char *
sgetsave(s)
	char *s;
{
	char *new = malloc((unsigned) strlen(s) + 1);

	if (new == NULL) {
		perror_reply(421, "Local resource failure: malloc");
		dologout(1);
		/* NOTREACHED */
	}
	(void) strcpy(new, s);
	return (new);
}

/*
 * Save the result of a getpwnam.  Used for USER command, since
 * the data returned must not be clobbered by any other command
 * (e.g., globbing).
 */
static struct passwd *
sgetpwnam(name)
	char *name;
{
	static struct passwd save;
	struct passwd *p;

	if ((p = getpwnam(name)) == NULL)
		return (p);
	if (save.pw_name) {
		free(save.pw_name);
		free(save.pw_passwd);
		free(save.pw_gecos);
		free(save.pw_dir);
		free(save.pw_shell);
	}
	save = *p;
	save.pw_name = sgetsave(p->pw_name);
	save.pw_passwd = sgetsave(p->pw_passwd);
	save.pw_gecos = sgetsave(p->pw_gecos);
	save.pw_dir = sgetsave(p->pw_dir);
	save.pw_shell = sgetsave(p->pw_shell);
	return (&save);
}

static int login_attempts;	/* number of failed login attempts */
static int askpasswd;		/* had user command, ask for passwd */
static char curname[10];	/* current USER name */

//andy add for anonymous
#if (defined(LOBOS) || defined(ALLNET))	
static int anonymous=0;
static int anonymous_enable=0;
#endif

static char motd_line[LINE_MAX];
static void
complete_login (passwd)
     char *passwd;
{
	FILE *fd;

	if (!guest && (!pw || *pw->pw_passwd)) 
	{
// == Hiram==
//#ifndef __USE_TEXT_PASSWORDS__
		char *xpasswd = NULL;

		if (pw) {
			char *salt = pw->pw_passwd;
			xpasswd = crypt (passwd, salt);
		}
//@@@@@@@@@@@@@@@
//	Change it to read the real text password !!!
//@@@@@@@@@@@@@@@
		if (!pw || !xpasswd || strcmp (xpasswd, pw->pw_passwd) != 0) {
			reply(530, "Login incorrect.");
			log_fail=1;
		  // andy 20040517 // if login fail , we should decrease ftp user number
//			ftpd_update_active_connections(0);
			
			if (logging)
				M4_syslog(LOG_NOTICE,
				    "FTP LOGIN FAILED FROM %s, %s",
				    remotehost, curname);
			pw = NULL;
			if (login_attempts++ >= 5) {
				M4_syslog(LOG_NOTICE,
				    "repeated login failures from %s",
				    remotehost);
				exit(0);
			}
			return;
		}
		else
			log_fail=0;
	}

	login_attempts = 0;		/* this time successful */

	if (setegid((gid_t)pw->pw_gid) < 0) 
	{
		reply(550, "Can't set gid.");
		return;
	}
 	(void) initgroups(pw->pw_name, pw->pw_gid);

	/* open wtmp before chroot */
	(void)sprintf(ttyline, "ftp%d", getpid());

	logged_in = 1;
	if (guest)   // anonymous is guest
	{
		/*
		 * We MUST do a chdir() after the chroot. Otherwise
		 * the old current directory will be accessible as "."
		 * outside the new root!
		 */
		if (chroot("/mnt") < 0 || chdir("/") < 0) 
		{
			reply(550, "Can't set guest privileges.");
			goto bad;
		}
	}	
	else if (chdir(pw->pw_dir) < 0) 
	{
		if (chdir("/") < 0) {
			reply(530, "User %s: can't change directory to %s.",
			    pw->pw_name, pw->pw_dir);
			goto bad;
		} else
			lreply(230, "No directory! Logging in with home=/");
	}
	
	
	
// Ansel test chroot, not only did the guest account need this
//	reply(230, "User dir is %s", pw->pw_dir);
/////////////

	if(!anonymous)
  	{
		if (chroot(pw->pw_dir) < 0 || chdir("/") < 0) 
		{
			reply(550, "Can't set to ftp folder,the folder isn't exit.");
			goto bad;
		}
 	}    // end of   if(!anonymous)

/////////////////

	if (seteuid((uid_t)pw->pw_uid) < 0) 
	{
		reply(550, "Can't set uid.");
		goto bad;
	}

 
	/*
	 * Display a login message, if it exists.
	 * N.B. reply(230,) must follow the message.
	 */
		  	 
	if ((fd = fopen("/etc/motd", "r")) != NULL) 
	{
		char *cp;

		while (fgets(motd_line, sizeof(motd_line), fd) != NULL) {
			if ((cp = strchr(motd_line, '\n')) != NULL)
				*cp = '\0';
			lreply(230, "%s", motd_line);
		}
		(void) fflush(stdout);
		(void) fclose(fd);
	}
		
	if (guest) 
	{
		reply(230, "Anonymous login ok, access restrictions apply.");
		if (logging)
			M4_syslog(LOG_INFO, "ANONYMOUS FTP LOGIN FROM %s, %s",
			    remotehost, passwd);
	} 
	else 
	{
		reply(230, "User %s logged in.", pw->pw_name);
		if (logging)
			M4_syslog(LOG_INFO, "FTP LOGIN FROM %s as %s",
			    remotehost, pw->pw_name);
	}
	
	strcpy(UserName, pw->pw_name);
	(void) umask(defumask);
	return;
bad:
	/* Forget all about it... */
	end_login();
}

/*
 * USER command.
 * Sets global passwd pointer pw if named account exists and is acceptable;
 * sets askpasswd if a PASS command is expected.  If logged in previously,
 * need to reset state.  If name is "ftp" or "anonymous", the name is not in
 * PATH_FTPUSERS, and ftp account exists, set guest and pw, then just return.
 * If account doesn't exist, ask for passwd anyway.  Otherwise, check user
 * requesting login privileges.  Disallow anyone who does not have a standard
 * shell as returned by getusershell().  Disallow anyone mentioned in the file
 * PATH_FTPUSERS to allow people such as root and uucp to be avoided.
 */
void
user(name)
	char *name;
{
	char *cp, *shell;
	
#if (defined(LOBOS) || defined(ALLNET))	
    FILE *fip;
    char LineBuf[256],anony[2];
    char *ptr, *end;
    int i;
#endif
    
	if (logged_in) {
		if (guest) {
			reply(530, "Can't change user from guest login.");
			return;
		}
		end_login();
	}

	guest = 0;

#if (defined(LOBOS) || defined(ALLNET))		
//andy add for parser config file
    if((fip = fopen("/etc/config","r")) == NULL)
        return ;
    else
    {
        while(fgets(LineBuf,255,fip) != NULL)
        {
            for(i=0;i<255;i++)
            {
                if((LineBuf[i] == ' ') || (LineBuf[i] == '\n') || (LineBuf[i] == '\r'))
                {
                    LineBuf[i] = '\0';
                    i=255;
                }
            }
            if((ptr = strstr(LineBuf,"FTP_ANONYMITY='")) != NULL){
             ptr+=15;
             end = strchr(ptr,'\'');
             *end = '\0';
             strcpy(anony,ptr);
            } 	
	      }
	}
	if(!strcmp(anony,"1"))
	{
	   anonymous_enable=1;
	//   HDBG("anonymous_enable=1\n");
	}
	else
	{
	   anonymous_enable=0;
	//   HDBG("anonymous_enable=0\n");
	}		  
//end	
#endif  // end of #if (defined(LOBOS) || defined(ALLNET))	

	if ( strcmp(name, "anonymous") == 0  && anonymous_enable == 1 )
	{
	  	guest = 1;
		askpasswd = 1;
		anonymous =1;
		pw = sgetpwnam("root");
		reply(331,"Anonymous user login......");

		return;	  	
	}

	
	if ((pw = sgetpwnam(name)) != NULL) 
	{
		if ((shell = pw->pw_shell) == NULL || *shell == 0)
			shell = "/bin/sh";
		while ((cp = getusershell()) != NULL)
			if (strcmp(cp, shell) == 0)
				break;
		endusershell();

		if (cp == NULL || checkuser(name)) {
			reply(530, "User %s access denied.", name);
			if (logging)
				M4_syslog(LOG_NOTICE,
				    "FTP LOGIN REFUSED FROM %s, %s",
				    remotehost, name);
			pw = (struct passwd *) NULL;
			return;
		}
	}
	if (logging)
		strncpy(curname, name, sizeof(curname)-1);

	if (!pw || *pw->pw_passwd) 
	{
		reply(331, "Password required for %s.", name);
		askpasswd = 1;
	} 
	else
	        complete_login (0);

	/*
	 * Delay before reading passwd after first failed
	 * attempt to slow down passwd-guessing programs.
	 */
	if (login_attempts)
		sleep((unsigned) login_attempts);
}

#ifdef M_FTPD_FULL
/*
 * Check if a user is in the file PATH_FTPUSERS
 */
static char cu_line[_BUFSIZ];
#endif

static int
checkuser(name)
	char *name;
{
#ifdef M_FTPD_FULL
	FILE *fd;
	int found = 0;
	char *p;

	if ((fd = fopen("/etc/ftpusers", "r")) != NULL) {
		while (fgets(cu_line, sizeof(cu_line), fd) != NULL)
			if ((p = strchr(cu_line, '\n')) != NULL) {
				*p = '\0';
				if (cu_line[0] == '#')
					continue;
				if (strcmp(cu_line, name) == 0) {
					found = 1;
					break;
				}
			}
		(void) fclose(fd);
	}
	return (found);
#else
  return (0);        // always return 0 as ftpd does not use the file now
#endif
}

/*
 * Terminate login as previous user, if any, resetting state;
 * used when USER command is given or login fails.
 */
static void
end_login()
{

	(void) seteuid((uid_t)0);
/* 	if (logged_in) */
/* 		logwtmp_keep_open (ttyline, "", ""); */
	pw = NULL;
	logged_in = 0;
	guest = 0;
	dologout(0);	
}

void
pass(passwd)
	char *passwd;
{
	if (logged_in || askpasswd == 0) {
		reply(503, "Login with USER first.");
		return;
	}
	askpasswd = 0;

#if (defined(LOBOS) || defined(ALLNET))	
  if(anonymous)  
  {   
     /*	
  	  if(strchr(passwd,'@'))
  	     complete_login (passwd);
      else {
         reply(530,"You should type the mail address for password!");
         return;                  
                           };      	     
     */
     //andy 20040701 anonymous user will not need to type passwd
       complete_login (passwd);
        
  }
  else
  		complete_login (passwd);

#else
    complete_login (passwd);
#endif	
	
}


static char r_line[_BUFSIZ];

void
retrieve(cmd, name)
	char *cmd, *name;
{
	FILE *fin, *dout;
	struct stat st;
	int (*closefunc) __P((FILE *));
	size_t buffer_size=0;
	int num;

#ifdef ANDY_ICONV
	char newname[1024];
	strcpy(newname,name);
	my_iconv("UTF-8",CLIENT_CHARSET,newname);
	name=newname;
#endif

//== Modify by George for setup ftp security ====
	num=GetFullPathName(name);
	if (num==0x1001 || root_flag || num==0) {
		reply(550, "Permission denied.");
		return;
	}
//================================

	if (cmd == 0) {
		fin = fopen(name, "r"), closefunc = fclose;
		st.st_size = 0;
	} else {
		snprintf (r_line, sizeof r_line, cmd, name);
		name = r_line;

//ansel test
//      reply(550, " r_line = %s****", r_line);
//
		fin = ftpd_popen(r_line, "r"), closefunc = ftpd_pclose;
		st.st_size = -1;
		buffer_size = _BUFSIZ;
	}
	if (fin == NULL) {
		if (errno != 0) {
			perror_reply(550, name);
			if (cmd == 0) {
				LOGCMD("get", name);
			}
		}
		return;
	}
	byte_count = -1;
	if (cmd == 0 && (fstat(fileno(fin), &st) < 0 || !S_ISREG(st.st_mode)
			 || !(buffer_size = st.st_blksize))) {
		reply(550, "%s: not a plain file.", name);
		goto done;
	}
	if (restart_point) {
		if (type == TYPE_A) {
			off_t i, n;
			int c;

			n = restart_point;
			i = 0;
			while (i++ < n) {
				if ((c=getc(fin)) == EOF) {
					perror_reply(550, name);
					goto done;
				}
				if (c == '\n')
					i++;
			}
#if defined(__USE_FILE_OFFSET64) && defined(__USE_FILE_OFFSET64)
		} else if (lseek64(fileno(fin), restart_point, SEEK_SET) < 0) { // 64-bit file I/O 
#else
		} else if (lseek(fileno(fin), restart_point, SEEK_SET) < 0) {
#endif
			perror_reply(550, name);
			goto done;
		}
	}
	dout = dataconn(name, st.st_size, "w");
	if (dout == NULL)
		goto done;
	send_data(fin, dout, buffer_size);
	(void) fclose(dout);
	data = -1;
	pdata = -1;
done:
	if (cmd == 0) {
#ifdef M_FTPD_DEBUG
		LOGBYTES("get", name, byte_count);
#else
		;
#endif
	}
	(*closefunc)(fin);
}

void
store(name, s_mode, unique)
	char *name, *s_mode;
	int unique;
{
	FILE *fout, *din;
	struct stat st;
	int (*closefunc) __P((FILE *));
	int num;

#ifdef ANDY_ICONV
	char newname[1024];
	strcpy(newname,name);
	M4_syslog(LOG_DEBUG, "store: before my_iconv(),newname=%s",newname);
	my_iconv(UTF8,CLIENT_CHARSET,newname);
	M4_syslog(LOG_DEBUG, "store: after my_iconv(),newname=%s",newname);
	name=newname;
#endif

//== Modify by George for setup ftp security ====
	num=GetFullPathName(name);
	if (num!=0x1010 || root_flag) {
		reply(550, "Permission denied.");
		return;
	}
//================================

	if (unique && stat(name, &st) == 0 &&
	    (name = gunique(name)) == NULL) {
		LOGCMD(*s_mode == 'w' ? "put" : "append", name);
		return;
	}

	if (restart_point)
		s_mode = "r+";
	fout = fopen(name, s_mode);
	closefunc = fclose;
	if (fout == NULL) {
		perror_reply(553, name);
		LOGCMD(*s_mode == 'w' ? "put" : "append", name);
		return;
	}
	byte_count = -1;
	if (restart_point) {
		if (type == TYPE_A) {
			off_t i, n;
			int c;

			n = restart_point;
			i = 0;
			while (i++ < n) {
				if ((c=getc(fout)) == EOF) {
					perror_reply(550, name);
					goto done;
				}
				if (c == '\n')
					i++;
			}
			/*
			 * We must do this seek to "current" position
			 * because we are changing from reading to
			 * writing.
			 */
#if defined(__USE_FILE_OFFSET64) && defined(__USE_FILE_OFFSET64)
			if (fseeko64(fout, 0, SEEK_CUR) < 0) {
#else
			if (fseek(fout, 0L, SEEK_CUR) < 0) {
#endif
				perror_reply(550, name);
				goto done;
			}
#if defined(__USE_FILE_OFFSET64) && defined(__USE_FILE_OFFSET64)
		} else if (lseek64(fileno(fout), restart_point, SEEK_SET) < 0) {
#else
		} else if (lseek(fileno(fout), restart_point, SEEK_SET) < 0) {
#endif
			perror_reply(550, name);
			goto done;
		}
	}
	din = dataconn(name, (off_t)-1, "r");
	if (din == NULL)
		goto done;
	if (receive_data(din, fout) == 0) {
		if (unique)
			reply(226, "Transfer complete (unique file name:%s).",
			    name);
		else
			reply(226, "Transfer complete.");
	}
	(void) fclose(din);
	data = -1;
	pdata = -1;
done:
	LOGBYTES(*s_mode == 'w' ? "put" : "append", name, byte_count);
	(*closefunc)(fout);
}

static FILE *
getdatasock(g_mode)
	char *g_mode;
{
	int on = 1, s, t, tries;

	if (data >= 0)
		return (fdopen(data, g_mode));
	(void) seteuid((uid_t)0);
	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s < 0)
		goto bad;
	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR,
	    (char *) &on, sizeof(on)) < 0)
		goto bad;
	/* anchor socket to avoid multi-homing problems */
	data_source.sin_family = AF_INET;
	data_source.sin_addr = ctrl_addr.sin_addr;
	for (tries = 1; ; tries++) {
		if (bind(s, (struct sockaddr *)&data_source,
		    sizeof(data_source)) >= 0)
			break;
		if (errno != EADDRINUSE || tries > 10)
			goto bad;
		sleep(tries);
	}
	(void) seteuid((uid_t)pw->pw_uid);
#if defined (IP_TOS) && defined (IPTOS_THROUGHPUT) && defined (IPPROTO_IP)
	on = IPTOS_THROUGHPUT;
	if (setsockopt(s, IPPROTO_IP, IP_TOS, (char *)&on, sizeof(int)) < 0)
		M4_syslog(LOG_WARNING, "setsockopt (IP_TOS): %m");
#endif
	return (fdopen(s, g_mode));
bad:
	/* Return the real value of errno (close may change it) */
	t = errno;
	(void) seteuid((uid_t)pw->pw_uid);
	(void) close(s);
	errno = t;
	return (NULL);
}

static FILE *dataconn(name, size, d_mode)
	char *name;
	off_t size;
	char *d_mode;
{
	char sizebuf[32];
	FILE *file;
	int retry = 0;
#if 0
	int tos;
#endif	/* 0 */

	file_size = size;
	byte_count = 0;
	if (size != (off_t) -1)
		(void) sprintf(sizebuf, " (%s bytes)", off_to_str (size));
	else
		(void) strcpy(sizebuf, "");
	if (pdata >= 0) {
		struct sockaddr_in from;
		int s, fromlen = sizeof(from);

		s = accept(pdata, (struct sockaddr *)&from, &fromlen);
		if (s < 0) {
			reply(425, "Can't open data connection.");
			(void) close(pdata);
			pdata = -1;
			return (NULL);
		}
		(void) close(pdata);
		pdata = s;
#if defined (IP_TOS) && defined (IPTOS_LOWDELAY) && defined (IPPROTO_IP)
		tos = IPTOS_LOWDELAY;
		(void) setsockopt(s, IPPROTO_IP, IP_TOS, (char *)&tos,
		    sizeof(int));
#endif
		reply(150, "Opening %s mode data connection for '%s'%s.",
		     type == TYPE_A ? "ASCII" : "BINARY", name, sizebuf);
		return (fdopen(pdata, d_mode));
	}
	if (data >= 0) {
		reply(125, "Using existing data connection for '%s'%s.",
		    name, sizebuf);
		usedefault = 1;
		return (fdopen(data, d_mode));
	}
	if (usedefault)
		data_dest = his_addr;
	usedefault = 1;
	file = getdatasock(d_mode);
	if (file == NULL) {
		reply(425, "Can't create data socket (%s,%d): %s.",
		    inet_ntoa(data_source.sin_addr),
		    ntohs(data_source.sin_port), strerror(errno));
		return (NULL);
	}
	data = fileno(file);
	while (connect(data, (struct sockaddr *)&data_dest,
	    sizeof(data_dest)) < 0) {
		if (errno == EADDRINUSE && retry < swaitmax) {
			sleep((unsigned) swaitint);
			retry += swaitint;
			continue;
		}
		perror_reply(425, "Can't build data connection");
		(void) fclose(file);
		data = -1;
		return (NULL);
	}
	reply(150, "Opening %s mode data connection for '%s'%s.",
	     type == TYPE_A ? "ASCII" : "BINARY", name, sizebuf);
	return (file);
}

/*
 * Tranfer the contents of "instr" to "outstr" peer using the appropriate
 * encapsulation of the data subject * to Mode, Structure, and Type.
 *
 * NB: Form isn't handled.
 */
static void
send_data(instr, outstr, blksize)
	FILE *instr, *outstr;
	off_t blksize;
{
	int c, cnt, filefd, netfd;
	char *buf;

	transflag++;
	if (setjmp(urgcatch)) {
		transflag = 0;
		return;
	}
	switch (type) {

	case TYPE_A:
		while ((c = getc(instr)) != EOF) {
			byte_count++;
			if (c == '\n') {
				if (ferror(outstr))
					goto data_err;
				(void) putc('\r', outstr);
			}
			(void) putc(c, outstr);
		}
		fflush(outstr);
		transflag = 0;
		if (ferror(instr))
			goto file_err;
		if (ferror(outstr))
			goto data_err;
		reply(226, "Transfer complete.");
		return;

	case TYPE_I:
	case TYPE_L:
		if ((buf = malloc((u_int)blksize)) == NULL) {
			transflag = 0;
			perror_reply(451, "Local resource failure: malloc");
			return;
		}
		netfd = fileno(outstr);
		filefd = fileno(instr);
		while ((cnt = read(filefd, buf, (u_int)blksize)) > 0 &&
		    write(netfd, buf, cnt) == cnt)
			byte_count += cnt;
		transflag = 0;
		(void)free(buf);
		if (cnt != 0) {
			if (cnt < 0)
				goto file_err;
			goto data_err;
		}
		reply(226, "Transfer complete.");
		return;
	default:
		transflag = 0;
		reply(550, "Unimplemented TYPE %d in send_data", type);
		return;
	}

data_err:
	transflag = 0;
	perror_reply(426, "Data connection");
	return;

file_err:
	transflag = 0;
	perror_reply(551, "Error on input file");
}

/*
 * Transfer data from peer to "outstr" using the appropriate encapulation of
 * the data subject to Mode, Structure, and Type.
 *
 * N.B.: Form isn't handled.
 */

static char rd_buf[_BUFSIZ];

static int
receive_data(instr, outstr)
	FILE *instr, *outstr;
{
	int c;
	int cnt, bare_lfs = 0;
	size_t dw_blksize = sizeof(rd_buf);

	transflag++;
	if (setjmp(urgcatch)) {
		transflag = 0;
		return (-1);
	}
	switch (type) {
	case TYPE_I:
	case TYPE_L:
//printf("sizeof(rd_buf) %d \n",sizeof(rd_buf));
		(void) fflush(stdout);
		while ((cnt = read(fileno(instr), rd_buf, dw_blksize)) != -1) {
			if (cnt == 0) {		/* end of file */
				break;
			}
			if (write(fileno(outstr), rd_buf, cnt) != cnt)
				goto file_err;
			byte_count += cnt;
		}
		if (cnt == -1)
			goto data_err;
		transflag = 0;
		return (0);

	case TYPE_E:
		reply(553, "TYPE E not implemented.");
		transflag = 0;
		return (-1);

	case TYPE_A:
		while ((c = getc(instr)) != EOF) {
			byte_count++;
			if (c == '\n')
				bare_lfs++;
			while (c == '\r') {
				if (ferror(outstr))
					goto data_err;
				if ((c = getc(instr)) != '\n') {
					(void) putc ('\r', outstr);
					if (c == '\0' || c == EOF)
						goto contin2;
				}
			}
			(void) putc(c, outstr);
	contin2:	;
		}
		fflush(outstr);
		if (ferror(instr))
			goto data_err;
		if (ferror(outstr))
			goto file_err;
		transflag = 0;
		if (bare_lfs) {
			lreply(226,
		"WARNING! %d bare linefeeds received in ASCII mode",
			    bare_lfs);
		(void)printf("   File may not have transferred correctly.\r\n");
		}
		return (0);
	default:
		reply(550, "Unimplemented TYPE %d in receive_data", type);
		transflag = 0;
		return (-1);
	}

data_err:
	transflag = 0;
	perror_reply(426, "Data Connection");
	return (-1);

file_err:
	transflag = 0;
	perror_reply(452, "Error writing file");
	return (-1);
}


#ifdef M_FTPD_FULL
static char sf_line[LINE_MAX];
#else
	static char sf_line[2048];
#endif
void
statfilecmd(filename)
	char *filename;
{
	FILE *fin;
	int c;

/* 	(void)snprintf(sf_line, sizeof(sf_line), "/bin/ls -lgA %s", filename); */
	(void)snprintf(sf_line, sizeof(sf_line), "/bin/ls -l %s", filename);
	fin = ftpd_popen(sf_line, "r");
	lreply(211, "status of %s:", filename);
	while ((c = getc(fin)) != EOF) {
		if (c == '\n') {
			if (ferror(stdout)){
				perror_reply(421, "control connection");
				(void) ftpd_pclose(fin);
				dologout(1);
				/* NOTREACHED */
			}
			if (ferror(fin)) {
				perror_reply(551, filename);
				(void) ftpd_pclose(fin);
				return;
			}
			(void) putc('\r', stdout);
		}
		(void) putc(c, stdout);
	}
	(void) ftpd_pclose(fin);
	reply(211, "End of Status");
}

void
statcmd()
{
	struct sockaddr_in *s_sin;
	u_char *a, *p;

	lreply(211, "%s FTP server status:", hostname);
	printf("     ftpd\r\n");
	printf("     Connected to %s", remotehost);
	if (!isdigit(remotehost[0]))
		printf(" (%s)", inet_ntoa(his_addr.sin_addr));
	printf("\r\n");
	if (logged_in) {
		if (guest)
			printf("     Logged in anonymously\r\n");
		else
			printf("     Logged in as %s\r\n", pw->pw_name);
	} else if (askpasswd)
		printf("     Waiting for password\r\n");
	else
		printf("     Waiting for user name\r\n");
	printf("     TYPE: %s", typenames[type]);
	if (type == TYPE_A || type == TYPE_E)
		printf(", FORM: %s", formnames[form]);
	if (type == TYPE_L)
#ifdef CHAR_BIT
		printf(" %d", CHAR_BIT);
#else
#if NBBY == 8
		printf(" %d", NBBY);
#else
		printf(" %d", bytesize);	/* need definition! */
#endif
#endif
	printf("; STRUcture: %s; transfer MODE: %s\r\n",
	    strunames[stru], modenames[mode]);
	if (data != -1)
		printf("     Data connection open\r\n");
	else if (pdata != -1) {
		printf("     in Passive mode");
		s_sin = &pasv_addr;
		goto printaddr;
	} else if (usedefault == 0) {
		printf("     PORT");
		s_sin = &data_dest;
printaddr:
		a = (u_char *) &s_sin->sin_addr;
		p = (u_char *) &s_sin->sin_port;
#define UC(b) (((int) b) & 0xff)
		printf(" (%d,%d,%d,%d,%d,%d)\r\n", UC(a[0]),
			UC(a[1]), UC(a[2]), UC(a[3]), UC(p[0]), UC(p[1]));
#undef UC
	} else
		printf("     No data connection\r\n");
	reply(211, "End of status");
}

void
fatal(s)
	char *s;
{

	reply(451, "Error in server: %s\n", s);
	reply(221, "Closing connection due to server error.");
	dologout(0);
	/* NOTREACHED */
}

void reply(int n, const char *fmt, ...)
{
	va_list ap;

	va_strt(ap, fmt);
	(void)printf("%d ", n);
	(void)vprintf(fmt, ap);
	(void)printf("\r\n");
	(void)fflush(stdout);
#ifdef M_FTPD_DEBUG
	if (debug) {
		M4_syslog(LOG_DEBUG, "<--- %d ", n);
		(void)vfprintf(M4_fd, fmt, ap);
		(void)fprintf(M4_fd, "\n");
	}
#endif
}

void lreply(int n, const char *fmt, ...)
{
	va_list ap;

	va_strt(ap, fmt);
	(void)printf("%d- ", n);
	(void)vprintf(fmt, ap);
	(void)printf("\r\n");
	(void)fflush(stdout);
#ifdef M_FTPD_DEBUG
	if (debug) {
		M4_syslog(LOG_DEBUG, "<--- %d- ", n);
		(void)vfprintf(M4_fd, fmt, ap);
		(void)fprintf(M4_fd, "\n");
	}
#endif
}

static void
ack(s)
	char *s;
{

	reply(250, "%s command successful.", s);
}

void
nack(s)
	char *s;
{

	reply(502, "%s command not implemented.", s);
}

/* ARGSUSED */
void
yyerror(s)
	char *s;
{
	char *cp;

	if ((cp = strchr(cbuf,'\n')) != NULL)
		*cp = '\0';
	reply(500, "'%s': command not understood.", cbuf);
}

void
delete(name)
	char *name;
{
	struct stat st;
	int num;

#ifdef ANDY_ICONV
	char newname[1024];
	strcpy(newname,name);
	my_iconv("UTF-8",CLIENT_CHARSET,newname);
	name=newname;
#endif

//== Modify by George for setup ftp security ====
	num=GetFullPathName(name);
	if (num!=0x1010 || root_flag) {
		reply(550, "Permission denied.");
		return;
	}
//================================

	LOGCMD("delete", name);
	if (stat(name, &st) < 0) {
		perror_reply(550, name);
		return;
	}
	if ((st.st_mode&S_IFMT) == S_IFDIR) {
		if (rmdir(name) < 0) {
			perror_reply(550, name);
			return;
		}
		goto done;
	}
	if (unlink(name) < 0) {
		perror_reply(550, name);
		return;
	}
done:
	ack("DELE");
}

void
cwd(path)
	char *path;
{
	int num;

#ifdef ANDY_ICONV
	char newpath[100];
	strcpy(newpath,path);
	M4_syslog(LOG_DEBUG, "cwd: before my_iconv(),newpath=%s",newpath);
	my_iconv(UTF8,CLIENT_CHARSET,newpath);
	M4_syslog(LOG_DEBUG, "cwd: after my_iconv(),newpath=%s",newpath);
	path=newpath;
#endif

//== Modify by George for setup ftp security ====
	num=GetFullPathName(path);
	if (num==0x1001 || num==0) { 
		reply(550, "Permission denied.");
		return;
	}
//================================

	if (chdir(path) < 0)
		perror_reply(550, path);
	else
		ack("CWD");

}

void
makedir(name)
	char *name;
{
	int num;

#ifdef ANDY_ICONV
	char newname[1024];
	strcpy(newname,name);
	M4_syslog(LOG_DEBUG, "makedir: before my_iconv(),name=%s",newname);
	my_iconv(UTF8,CLIENT_CHARSET,newname);
	M4_syslog(LOG_DEBUG, "makedir: after my_iconv(),name=%s",newname);
	name=newname;
#endif

//== Modify by George for setup ftp security ====
	num=GetFullPathName(name);
	if (num!=0x1010 || root_flag) {
		reply(550, "Permission denied.");
		return;
	}
//================================

	LOGCMD("mkdir", name);
	if (mkdir(name, 0777) < 0)
		perror_reply(550, name);
	else if (name[0] == '/')
		reply (257, "\"%s\" new directory created.");
	else {
		/* We have to figure out what our current directory is so that we can
		   give an absolute name in the reply.  */
		char *m_cwd = getcwd (0, PATH_LENGTH);
		if (m_cwd) {
			if (m_cwd[1] == '\0')
				m_cwd[0] = '\0';
			reply (257, "\"%s/%s\" new directory created.", m_cwd, name);
			free (m_cwd);
		} else
			reply (257, "(unknown absolute name) new directory created.");
	}
}

void
removedir(name)
	char *name;
{
	int num;

#ifdef ANDY_ICONV
	char newname[1024];
	strcpy(newname,name);
	my_iconv("UTF-8",CLIENT_CHARSET,newname);
	name=newname;
#endif

//== Modify by George for setup ftp security ====
	num=GetFullPathName(name);
	if (num!=0x1010 || root_flag) {
		reply(550, "Permission denied.");
		return;
	}
//================================

	LOGCMD("rmdir", name);
	if (rmdir(name) < 0)
		perror_reply(550, name);
	else
		ack("RMD");
}

void
pwd()
{
	  unsigned char path[1024];

	//path = getcwd (0, PATH_LENGTH);
	getcwd(path,sizeof(path));

#ifdef ANDY_ICONV 
	M4_syslog(LOG_DEBUG, "pwd: before my_iconv(),path=%s",path);
	my_iconv(CLIENT_CHARSET,UTF8,path);
	M4_syslog(LOG_DEBUG, "pwd: after my_iconv(),path=%s",path);	
#endif
	
	if (path) {
		reply(257, "\"%s\" is current directory.", path);
		//free (path);
	} else
		reply(550, "%s.", strerror (errno));
}


//== Modify by George for setup ftp security ====
int GetFullPathName(path)
	char *path;
{
	char *FullPath, *OldPath;
	char dirt[256], *ptr;
	int flag=0;

	strcpy(dirt, path);
	OldPath = getcwd (0, PATH_LENGTH);
	while (1) {
		if (chdir(dirt) < 0) {
			flag = 1;
			ptr = strrchr(dirt, '/');
			if (ptr != NULL) {
				*ptr = 0x00;
				if (dirt == NULL) {
					dirt[0] = '/';
					FullPath = dirt;
				}
			}
			else {
				FullPath = OldPath;
				break;
			}
		}
		else {
			FullPath = getcwd (0, PATH_LENGTH);
			chdir(OldPath);
			break;
		}
	}

	if(FullPath[0]==0x2F && FullPath[1]=='\0')
		root_folder=1;
	else
		root_folder=0;
	
	//root_folder = (!strcmp('/',FullPath)) ? 1 : 0 ;
	
	return (sec_right(FullPath ,UserName, flag));
}
//== ================================


char *
renamefrom(name)
	char *name;
{
	struct stat st;
	int num;

#ifdef ANDY_ICONV
	char newname[1024];
	strcpy(newname,name);
	my_iconv("UTF-8",CLIENT_CHARSET,newname);
	name=newname;
#endif

//== Modify by George for setup ftp security ====
	num=GetFullPathName(name);
	if (num!=0x1010 || root_flag) {
		reply(550, "Permission denied.");
		return((char *)0);
	}
//================================

	if (stat(name, &st) < 0) {
		perror_reply(550, name);
		return ((char *)0);
	}
	reply(350, "File exists, ready for destination name");
	return (name);
}

void
renamecmd(from, to)
	char *from, *to;
{
	int num;

#ifdef ANDY_ICONV
	char newname[1024];
	strcpy(newname,to);
	my_iconv("UTF-8",CLIENT_CHARSET,newname);
	to=newname;
#endif
	
//== Modify by George for setup ftp security ====
	num=GetFullPathName(to);
	if (num!=0x1010 || root_flag) {
		reply(550, "Permission denied.");
		return;
	}
//================================

	LOGCMD2("rename", from, to);
	if (rename(from, to) < 0)
		perror_reply(550, "rename");
	else
		ack("RNTO");
}

static void
dolog(d_sin)
	struct sockaddr_in *d_sin;
{
	const char *name;
	struct hostent *hp=NULL;
//	struct hostent *hp = gethostbyaddr((char *)&d_sin->sin_addr,
//		sizeof(struct in_addr), AF_INET);
	
	
//	M4_syslog(LOG_DEBUG, "dolog: after gethostbyaddr : %s",hstrerror(h_errno));
	if (hp)
		name = hp->h_name;
	else
		name = inet_ntoa(d_sin->sin_addr);

	if (remotehost)
		free (remotehost);
	remotehost = malloc (strlen (name));
	strcpy (remotehost, name);

#ifdef M_FTPD_DEBUG
	if (logging)
		M4_syslog(LOG_INFO, "connection from %s", remotehost);
#endif
}

/*
 * Record logout in wtmp file
 * and exit with supplied status.
 */
void
dologout(status)
	int status;
{
  /* racing condition with SIGURG: If SIGURG is receive
     here, it will jump back has root in the main loop
     David Greenman:dg@root.com
  */
  transflag = 0; 

	if (logged_in) {
		(void) seteuid((uid_t)0);
/* 		logwtmp_keep_open (ttyline, "", ""); */
	}
/*	
#ifdef M_NH200
	// Decrease the NO# of active FTP connection(s) into file as closed

 mkdir("/~!@#$%^george",0777);   // we already have string "/sh" in memory as a part of
                 // "/bin/sh"
 chroot("/~!@#$%^george");
 chroot("../../../../../../../../../"); // a number of "../" here,
                                        // I used 0x1	
	
  if(!log_fail)
     ftpd_update_active_connections(0);
 
#if (defined(LOBOS) || defined(ALLNET))	
  if(!anonymous)
  {
  	rmdir("/mnt/~!@#$%^george");
  } 
  else
    rmdir("/mnt/~!@#$%^george");
#else
 rmdir("/mnt/~!@#$%^george"); 
#endif
 
#endif   // end ifdef M_NH200
*/	
	/* beware of flushing buffers after a SIGPIPE */
	_exit(status);
}

static void
myoob(signo)
	int signo;
{
	char *cp;

	/* only process if transfer occurring */
	if (!transflag)
		return;
	cp = tmpline;
	if (telnet_fgets(cp, 7, stdin) == NULL) {
		reply(221, "You could at least say goodbye.");
		dologout(0);
	}
	upper(cp);
	if (strcmp(cp, "ABOR\r\n") == 0) {
		tmpline[0] = '\0';
		reply(426, "Transfer aborted. Data connection closed.");
		reply(226, "Abort successful");
		longjmp(urgcatch, 1);
	}
	if (strcmp(cp, "STAT\r\n") == 0) {
		if (file_size != (off_t) -1)
			reply(213, "Status: %s of %s bytes transferred",
			    off_to_str (byte_count), off_to_str (file_size));
		else
			reply(213, "Status: %s bytes transferred",
			      off_to_str (byte_count));
	}
}

/*
 * Note: a response of 425 is not mentioned as a possible response to
 *	the PASV command in RFC959. However, it has been blessed as
 *	a legitimate response by Jon Postel in a telephone conversation
 *	with Rick Adams on 25 Jan 89.
 */
void
passive()
{
	int len;
	char *p, *a;

	pdata = socket(AF_INET, SOCK_STREAM, 0);
	if (pdata < 0) {
		perror_reply(425, "Can't open passive connection");
		return;
	}
	pasv_addr = ctrl_addr;
	pasv_addr.sin_port = 0;
	(void) seteuid((uid_t)0);
	if (bind(pdata, (struct sockaddr *)&pasv_addr, sizeof(pasv_addr)) < 0) {
		(void) seteuid((uid_t)pw->pw_uid);
		goto pasv_error;
	}
	(void) seteuid((uid_t)pw->pw_uid);
	len = sizeof(pasv_addr);
	if (getsockname(pdata, (struct sockaddr *) &pasv_addr, &len) < 0)
		goto pasv_error;
	if (listen(pdata, 1) < 0)
		goto pasv_error;
	a = (char *) &pasv_addr.sin_addr;
	p = (char *) &pasv_addr.sin_port;

#define UC(b) (((int) b) & 0xff)

	reply(227, "Entering Passive Mode (%d,%d,%d,%d,%d,%d)", UC(a[0]),
		UC(a[1]), UC(a[2]), UC(a[3]), UC(p[0]), UC(p[1]));
	return;

pasv_error:
	(void) close(pdata);
	pdata = -1;
	perror_reply(425, "Can't open passive connection");
	return;
}

/*
 * Generate unique name for file with basename "local".
 * The file named "local" is already known to exist.
 * Generates failure reply on error.
 */
static char *
gunique(local)
	char *local;
{
	static char *new = 0;
	struct stat st;
	int count;
	char *cp;

	cp = strrchr(local, '/');
	if (cp)
		*cp = '\0';
	if (stat(cp ? local : ".", &st) < 0) {
		perror_reply(553, cp ? local : ".");
		return ((char *) 0);
	}
	if (cp)
		*cp = '/';

	if (new)
		free (new);

	new = malloc (strlen (local) + 5); /* '.' + DIG + DIG + '\0' */
	if (new) {
		strcpy(new, local);
		cp = new + strlen(new);
		*cp++ = '.';
		for (count = 1; count < 100; count++) {
			(void)sprintf(cp, "%d", count);
			if (stat(new, &st) < 0)
				return (new);
		}
	}

	reply(452, "Unique file name cannot be created.");
	return (NULL);
}

/*
 * Format and send reply containing system error number.
 */
void
perror_reply(code, string)
	int code;
	char *string;
{

	reply(code, "%s: %s.", string, strerror(errno));
}

static char *onefile[] = {
	"",
	0
};

void
send_file_list(whichf)
	char *whichf;
{
	struct stat st;
	DIR *dirp = NULL;
	struct dirent *dir;
	FILE *dout = NULL;
	char **dirlist, *dirname;
	char *path = NULL;
	int simple = 0;
	int freeglob = 0;
	glob_t gl;
	int num;

// andy  from ls.c
    char		*cp;
    char		buf[256];
    char		nbuf[1204];

// owner and group
typedef	int	BOOL;

#define	FALSE	((BOOL) 0)
#define	TRUE	((BOOL) 1)

	struct	passwd	*lspwd;
	struct	group	*grp;
	static	char	username[12];
	static	int	userid;
	static	BOOL	useridknown;
	static	char	groupname[12];
	static	int	groupid;
	static	BOOL	groupidknown;

//== Modify by George for setup ftp security ====
	num = GetFullPathName(whichf);
	if (num==0x1001 || num==0) {
		reply(550, "Permission denied.");
		return;
	}
//================================
  
	if (strpbrk(whichf, "~{[*?") != NULL) {
		int flags = GLOB_NOCHECK;

#ifdef GLOB_BRACE
		flags |= GLOB_BRACE;
#endif
#ifdef GLOB_QUOTE
		flags |= GLOB_QUOTE;
#endif
#ifdef GLOB_TILDE
		flags |= GLOB_TILDE;
#endif

		memset(&gl, 0, sizeof(gl));
		freeglob = 1;

#ifdef ANDY_ICONV
	char newname[1024];
	strcpy(newname,whichf);
	//iconv_pt=iconv_pt_c2u;
	M4_syslog(LOG_DEBUG, "send_file_list: before glob");
	my_iconv(UTF8,CLIENT_CHARSET,newname);
	whichf=newname;
	//strcpy(whichf,bufout);	
#endif
		if (glob(whichf, flags, 0, &gl)) {
			reply(550, "not found");
			goto out;
		} 
		else if (gl.gl_pathc == 0) {
			errno = ENOENT;
			perror_reply(550, whichf);
			goto out;
		}
		dirlist = gl.gl_pathv;
	} 
	else {
		onefile[0] = whichf;
		dirlist = onefile;
		simple = 1;
	}

	if (setjmp(urgcatch)) {
		transflag = 0;
		goto out;
	}

	while ((dirname = *dirlist++)!= NULL) 
	{ 
#if defined(__USE_FILE_OFFSET64) && defined(__USE_FILE_OFFSET64)
		u_int64_t totalsize = 0;
#else
		unsigned int totalsize=0;
#endif
		if (stat(dirname, &st) < 0) 
		{
			/*
			 * If user typed "ls -l", etc, and the client
			 * used NLST, do what the user meant.
			 */
			if (dirname[0] == '-' && *dirlist == NULL && transflag == 0) 
			{
				if ((dirp = opendir(".")) == NULL)
					continue;		    	

				while ((dir = readdir(dirp)) != NULL) 
				{
//					char *nbuf;                      
					cp = buf;
					*cp='\0';
					
				
					if(root_folder)
					{	
						M4_syslog(LOG_INFO, "dir->d_name=%s", dir->d_name);
						if(strstr(dir->d_name,"codepage.") || strstr(dir->d_name,"unicode_map.") || strstr(dir->d_name,"upnp_soho")
							|| !strcmp(dir->d_name,"ntx_libra.tgz")	|| !strcmp(dir->d_name,"bt_clnt") || !strcmp(dir->d_name,"convert")
							|| !strcmp(dir->d_name,"ntx_chset.h") || !strcmp(dir->d_name,"spool") || !strcmp(dir->d_name,"watch_prog")
							|| !strcmp(dir->d_name,"lprm") || !strcmp(dir->d_name,"lpr") || !strcmp(dir->d_name,"lpq") || !strcmp(dir->d_name,"lpd")
							|| !strcmp(dir->d_name,"lpc") || !strcmp(dir->d_name,"lp") || !strcmp(dir->d_name,".ftpd.conf") || !strcmp(dir->d_name,"time_status"))
							continue;	
					}
					
					
					if ( dirname [1] != 'a' && dirname [2] != 'a')
					{      
						if (dir->d_name[0] == '.')
							continue;
						if (!strcmp(dir->d_name,"lost+found"))			  	 
							continue;			  
//						if (dir->d_name[0] == '.' && dir->d_name[1] == '\0')			  	 
//							continue;
//						if (dir->d_name[0] == '.' && dir->d_name[1] == '.' && dir->d_name[2] == '\0')			  
//							continue;
					}

//					nbuf = (char *) alloca (1 + 1 + strlen (dir->d_name) + 1);
					memset(nbuf, 0, sizeof(nbuf));
					sprintf(nbuf, "%s/%s", ".", dir->d_name);

					/*
					  * We have to do a stat to insure it's
					  * not a directory or special file.
					  */
					if (simple || (stat(nbuf, &st) == 0 && S_ISREG(st.st_mode))) 
					{
						if (dout == NULL) 
						{
							dout = dataconn("file list", (off_t)-1,"w");
							if (dout == NULL)
								goto out;
							transflag++;
						}
								
						///  from lsfile
						if ( stat(nbuf, &st) == 0   )
						{
							if (   dirname[1] == 'l' || dirname [2] == 'l' || dirname[1] == 'L' || dirname [2] == 'L' )  
							{
								strcpy(cp, modestring(st.st_mode));
								cp += strlen(cp);
								sprintf(cp, "%3d ", st.st_nlink);
								cp += strlen(cp);

								//  owner and group
								if (!useridknown || (st.st_uid != userid))
								{
									lspwd = getpwuid(st.st_uid);
									if (lspwd)
										strcpy(username, lspwd->pw_name);
									else
										sprintf(username, "%d", st.st_uid);
									userid = st.st_uid;
									useridknown = TRUE;
								}

								sprintf(cp, "%-8s ", username);
								cp += strlen(cp);

								if (!groupidknown || (st.st_gid != groupid))
								{
									grp = getgrgid(st.st_gid);
									if (grp)
										strcpy(groupname, grp->gr_name);
									else
										sprintf(groupname, "%d", st.st_gid);
									groupid = st.st_gid;
									groupidknown = TRUE;
								}

								sprintf(cp, "%-8s ", groupname);
								cp += strlen(cp);
		
								//  end 
								if (S_ISBLK(st.st_mode) || S_ISCHR(st.st_mode))
									sprintf(cp, "%3d, %3d ", (int)(st.st_rdev >> 8),	(int)(st.st_rdev & 0xff));
								else
								{
#if defined(__USE_FILE_OFFSET64) && defined(__USE_FILE_OFFSET64)
									sprintf(cp, "%Lu ", st.st_size);      // 64-bit file size
#else
									sprintf(cp, "%8ld ", st.st_size);
#endif
									totalsize += st.st_size;
								}

								cp += strlen(cp);
								sprintf(cp, " %-12s ", timestring(st.st_mtime));
								
							// Andy 20050420
							#ifdef ANDY_ICONV	
								if (nbuf[0] == '.' && nbuf[1] == '/'){
									M4_syslog(LOG_DEBUG, "nbuf[2]=%s",&nbuf[2]);	
									my_iconv(CLIENT_CHARSET,UTF8,&nbuf[2]);
									M4_syslog(LOG_DEBUG, "new nbuf[2]=%s",&nbuf[2]);	
									fprintf(dout, "%s%s\r\n",buf, &nbuf[2]);
								}
								else{									
									my_iconv(CLIENT_CHARSET,UTF8,nbuf);																												
									fprintf(dout, "%s%s\r\n",buf, nbuf);
								}
		      				#else
								if (nbuf[0] == '.' && nbuf[1] == '/')
									fprintf(dout, "%s%s\r\n",buf, &nbuf[2]);
								else
									fprintf(dout, "%s%s\r\n",buf, nbuf);		      					
		      				#endif
		      				
							}		 // end of 	 if ( dirname[1] == 'l' || dirname [2] == 'l') 
							else    // arg has no '-l' parameter ( mean '-a')
							{
							#ifdef ANDY_ICONV
								if (nbuf[0] == '.' && nbuf[1] == '/'){
									my_iconv(CLIENT_CHARSET,"UTF-8",&nbuf[2]);								
									fprintf(dout, "%s\r\n",&nbuf[2]);
								}									
								else{
									my_iconv(CLIENT_CHARSET,"UTF-8",nbuf);												
									fprintf(dout, "%s\r\n",nbuf);							
								}
							#else	
								if (nbuf[0] == '.' && nbuf[1] == '/')
									fprintf(dout, "%s\r\n",&nbuf[2]);
								else
									fprintf(dout, "%s\r\n",nbuf);
							#endif									
				
							}
						}      // end of if  ( stat(nbuf, &st) == 0    )      		
					}  // end of if (simple || (stat(nbuf, &st) == 0 && S_ISREG(st.st_mode))) 
				}  // end of while ((dir = readdir(dirp)) != NULL) 
				(void) closedir(dirp);
				if (dout == NULL)  //   mean ftp client user type 'ls -l', and directory is empty
				                   //  we must send packet include 'total 0' string to ftp client
				{
					dout = dataconn("file list", (off_t)-1,"w");					
					if (dout == NULL)
						goto out;
					//fprintf(dout, "total 0\r\n");  // totalsize = 0		 				
					transflag++;
					goto out1;		    	
				}
				else
/* 
#if defined(__USE_FILE_OFFSET64) && defined(__USE_FILE_OFFSET64)
					fprintf(dout, "total %Lu\r\n",totalsize);
#else
					fprintf(dout, "total %d\r\n",totalsize);
#endif 
*/
				goto out1;		    	
			}  // end of if (dirname[0] == '-' && *dirlist == NULL && transflag == 0) 
			
			perror_reply(550, whichf);
			if (dout != NULL) {
				(void) fclose(dout);
				transflag = 0;
				data = -1;
				pdata = -1;
			}
			goto out;
		}   // end of if (stat(dirname, &st) < 0) 
		
		if (S_ISREG(st.st_mode)) {
			if (dout == NULL) {
				dout = dataconn("file list", (off_t)-1, "w");
				if (dout == NULL)
					goto out;
				transflag++;
			}
#ifdef ANDY_ICONV
			M4_syslog(LOG_DEBUG, "before fprintf...");
			my_iconv(CLIENT_CHARSET,"UTF-8",dirname);
#endif			
			fprintf(dout, "%s%s\n", dirname,	type == TYPE_A ? "\r" : "");
			byte_count += strlen(dirname) + 1;
			continue;
		} 
		else if (!S_ISDIR(st.st_mode))
			continue;
		if ((dirp = opendir(dirname)) == NULL)
			continue;
		while ((dir = readdir(dirp)) != NULL) 
		{	
			
			if(root_folder)
			{	
				M4_syslog(LOG_INFO, "dir->d_name=%s", dir->d_name);
				if(strstr(dir->d_name,"codepage.") || strstr(dir->d_name,"unicode_map.") || strstr(dir->d_name,"upnp_soho")
					|| !strcmp(dir->d_name,"ntx_libra.tgz")	|| !strcmp(dir->d_name,"bt_clnt") || !strcmp(dir->d_name,"convert")
					|| !strcmp(dir->d_name,"ntx_chset.h") || !strcmp(dir->d_name,"spool") || !strcmp(dir->d_name,"watch_prog")
					|| !strcmp(dir->d_name,"lprm") || !strcmp(dir->d_name,"lpr") || !strcmp(dir->d_name,"lpq") || !strcmp(dir->d_name,"lpd")
					|| !strcmp(dir->d_name,"lpc") || !strcmp(dir->d_name,"lp") || !strcmp(dir->d_name,".ftpd.conf") || !strcmp(dir->d_name,"time_status"))
					continue;	
			}			
			
			if (dir->d_name[0] == '.')
				continue;
			if (!strcmp(dir->d_name,"lost+found"))			  	 
				continue;			  
//			if (dir->d_name[0] == '.' && dir->d_name[1] == '\0')
//				continue;
//			if (dir->d_name[0] == '.' && dir->d_name[1] == '.' && dir->d_name[2] == '\0')
//				continue;

			memset(nbuf, 0, sizeof(nbuf));
			sprintf(nbuf, "%s/%s", dirname, dir->d_name);
       
			/*
			 * We have to do a stat to insure it's
			 * not a directory or special file.
			 */
			if (simple || (stat(nbuf, &st) == 0 && S_ISREG(st.st_mode))) 
			{
				if (dout == NULL) 
				{
					dout = dataconn("file list", (off_t)-1,	"w");
					if (dout == NULL)
						goto out;
					transflag++;
				}
			#ifdef ANDY_ICONV
				if (nbuf[0] == '.' && nbuf[1] == '/'){
					my_iconv(CLIENT_CHARSET,UTF8,&nbuf[2]);
					fprintf(dout, "%s%s\n", &nbuf[2], type == TYPE_A ? "\r" : "");
				}
				else{
					my_iconv(CLIENT_CHARSET,UTF8,nbuf);
					fprintf(dout, "%s%s\n", nbuf, type == TYPE_A ? "\r" : "");
				}
			#else
				if (nbuf[0] == '.' && nbuf[1] == '/')
					fprintf(dout, "%s%s\n", &nbuf[2], type == TYPE_A ? "\r" : "");
				else
					fprintf(dout, "%s%s\n", nbuf, type == TYPE_A ? "\r" : "");			
			#endif					
				byte_count += strlen(nbuf) + 1;
			}
		}
	} // end of while ((dirname = *dirlist++)!= NULL)
out1:
//andy add 20050802 to resolv the bug which ls empty folder will lead to disconnect
#ifdef M_NH230
	if (dout == NULL) 
	{
		dout = dataconn("file list", (off_t)-1,	"w");
		if (dout == NULL)
			goto out;
		transflag++;
	}
#endif	
	
//	if (dout == NULL)
//		reply(550, "No files found.");
//	else if (ferror(dout) != 0)
	if (ferror(dout) != 0)
		perror_reply(550, "Data connection");
	else
		reply(226, "Transfer complete.");

	transflag = 0;
	if (dout != NULL)
		(void) fclose(dout);
	data = -1;
	pdata = -1;
out:
	if (freeglob) {
		freeglob = 0;
		globfree(&gl);
	}
}



#ifdef M_FTPD_DEBUG
void M4_printf(const char *fmt, ...)
{
	va_list ap;

	if (M4_fd == NULL) {
#ifdef M_NH200
	  M4_fd = fopen("/var/run/ftpd.log", "w");        // location of log file
#else
	  M4_fd = fopen("/etc/ftpd.log", "w");
#endif
	  if (M4_fd == NULL) {
	    perror("fopen");
	    exit(1);
	  }
	  setvbuf(M4_fd, NULL, _IONBF, 0);
	}
	va_strt(ap, fmt);
	(void)vfprintf(M4_fd, fmt, ap);
	(void)fprintf(M4_fd, "\n");
	fflush(M4_fd);
	va_end(ap);
}
#endif

/*
#ifdef M_NH200
//
// Subject :
//   Retrieve the NO# of ftpd forked (for each active FTP connection) from file.
// Input :
//   None
// Output :
//   NO# of active FTP connection(s)
//   0 if file not existed
//
unsigned int ftpd_get_active_connections(void)
{
	unsigned int dw_ftp_qty = 0;
	unsigned char sz_line[80];
	FILE *lpst_file = NULL;

	// 1) NULL the local variables
	memset(sz_line, '\0', sizeof(sz_line));
	
	// 2) Open the file to retrieve the NO# of active FTP connection(s)
	lpst_file = fopen(STR_FTPD_NUM_FILE, "rt");
	if (lpst_file) {	// File opened for read OK ?		
		if (fread(sz_line, sizeof(char), sizeof(sz_line) - 1, lpst_file) > 0) {
			dw_ftp_qty = (unsigned int)atoi(sz_line);
		}
#ifdef M_FTPD_DEBUG
		else
			M4_syslog(LOG_WARNING, "ftpd: cannot read from file [%s] with (rt).\n", STR_FTPD_NUM_FILE);
#endif
			
		fclose(lpst_file);
	}
	
	// 3) return the NO# of active FTP connection(s)
	return (dw_ftp_qty);
}


//
// Subject :
//   Retrieve the NO# of ftpd forked (for each active FTP connection) from file.
// Input :
//   b_flag :  1 ==> active FTP connections + 1
//             0 ==> active FTP connections - 1
// Output :
//   0     : update OK
//   not 0 : if something wrong
//
int ftpd_update_active_connections(unsigned char b_flag)
{
	unsigned int dw_FTP_connections = 0;
	unsigned char sz_line[80];
	FILE *lpst_file = NULL;
	
	// 1) Initialize local variables
	memset(sz_line, '\0', sizeof(sz_line));
	dw_FTP_connections = ftpd_get_active_connections();
	
	// 2) Open the file to write the new NO# of active FTP connection(s)	
	lpst_file = fopen(STR_FTPD_NUM_FILE, "w+t");
	if (lpst_file) {	        // File is opened successfully for write ?
		// 2-1) File opened successfully
		// 2-1-2) Add or Sub the NO# of active FTP connection(s)
		if (b_flag) {           // add with 1 ?
			if (dw_FTP_connections < M_FTP_MAX)
				dw_FTP_connections++;
		}
		else {                  // sub with 1 ?
			if (dw_FTP_connections)
				dw_FTP_connections--;
		}
		
		// 2-1-3) Convert NO# to character string
		sprintf(sz_line, "%d", dw_FTP_connections);
		
		// 2-1-4) Write new NO# of FTP connections to file
		if (fwrite(sz_line, sizeof(char), strlen(sz_line), lpst_file) != strlen(sz_line))
#ifdef M_FTPD_DEBUG
			M4_syslog(LOG_WARNING, "ftpd: cannot write to file [%s] with (w+t).\n", STR_FTPD_NUM_FILE);
#else
			;
#endif
			
		// 2-1-5) Close file	
		fclose(lpst_file);
		return (0);             // done
	}
	else {
		// 2-2) File cannot be opened
		M4_syslog(LOG_WARNING, "ftpd: cannot open file [%s] to read and write with (w+t).\n", STR_FTPD_NUM_FILE);
		return (-ENOENT);
	}
}

#endif   // end of #ifdef {M_NH200}
*/

/*
 * Return the standard ls-like mode string from a file mode.
 * This is static and so is overwritten on each call.
 */
static char *modestring(mode)
{
	static	char	buf[12];

	strcpy(buf, "----------");

	/*
	 * Fill in the file type.
	 */
	if (S_ISDIR(mode))
		buf[0] = 'd';
	if (S_ISCHR(mode))
		buf[0] = 'c';
	if (S_ISBLK(mode))
		buf[0] = 'b';
	if (S_ISFIFO(mode))
		buf[0] = 'p';
#ifdef	S_ISLNK
	if (S_ISLNK(mode))
		buf[0] = 'l';
#endif
#ifdef	S_ISSOCK
	if (S_ISSOCK(mode))
		buf[0] = 's';
#endif

	/*
	 * Now fill in the normal file permissions.
	 */
	if (mode & S_IRUSR)
		buf[1] = 'r';
	if (mode & S_IWUSR)
		buf[2] = 'w';
	if (mode & S_IXUSR)
		buf[3] = 'x';
	if (mode & S_IRGRP)
		buf[4] = 'r';
	if (mode & S_IWGRP)
		buf[5] = 'w';
	if (mode & S_IXGRP)
		buf[6] = 'x';
	if (mode & S_IROTH)
		buf[7] = 'r';
	if (mode & S_IWOTH)
		buf[8] = 'w';
	if (mode & S_IXOTH)
		buf[9] = 'x';

	/*
	 * Finally fill in magic stuff like suid and sticky text.
	 */
	if (mode & S_ISUID)
		buf[3] = ((mode & S_IXUSR) ? 's' : 'S');
	if (mode & S_ISGID)
		buf[6] = ((mode & S_IXGRP) ? 's' : 'S');
	if (mode & S_ISVTX)
		buf[9] = ((mode & S_IXOTH) ? 't' : 'T');

	return buf;
}

/*
 * Get the time to be used for a file.
 * This is down to the minute for new files, but only the date for old files.
 * The string is returned from a static buffer, and so is overwritten for
 * each call.
 */
static char *timestring(t)
	long	t;
{
	long		now;
	char		*str;
	static	char	buf[26];

	time(&now);

	str = ctime(&t);

	strcpy(buf, &str[4]);
	buf[12] = '\0';

	if ((t > now) || (t < now - 365*24*60*60L)) {
		strcpy(&buf[7], &str[20]);
		buf[11] = '\0';
	}

	return buf;
}

