//
// Subject :
//   The CTorrent is a P2P client for BitTorrent network.
//
// History :
//   v1.00  08/19/2004
//     1) Define a macro M_NH220 in the Makefile for the patches specific to NAS with P2P client.
//     2) Bug fix :
//        A) Problem :
//             When CTorrent runs with some *.torrent file(s), it will cause the kernel API munmap() error.
//             Eroro message displayed by kernel :
//                ERROR, munmap of non-mmaped memory by process xxxx (ctorrent).
//        B) Reason :
//             When free memory by "delete []b", it works fine under PC with RedHat Linux.
//             But it failed under Brecis MSP2006 with uClinux under some *.torrent file. 
//        C) Solution :
//             In the btcontent.cpp file, InitialFromMI() routine :
//               change the line 
//                    delete []b; 
//               to
//                    delete b;
//
//             And we modify all the codes with this patch.
//
//     3) Bug fix :
//        A) Problem :
//             When run CTorrent a while, the system will be out of memory.
//        B) Reason :
//             CTorrent uses cache I/O (allocate memory for cache by itself).
//        C) Solution :
//             Do not use the cache I/O.
//
//             In the btcontent.cpp file, CacheIO(0 routine :
//             When the routine is entered, we call the following code to direct I/O.
//                 return m_btfiles.IO(buf, off, len, method);
//
//     4) Bug fix :
//        A) Problem :
//             The original source codes supports only up to 2GB (4-byte) file size.
//             So, if the size of file(s) in the *.torrent files are more than 2GB,
//             failed to parse the meta info of *.torrent file(s).
//        B) Reason :
//             The original source codes uses "size_t" (4-byte long integer) to represent
//             the file size. ==> only up to 2GB
//        C) Solution :
//             a) Change to use "u_int64_t" to represent file size.
//             b) Add 2 macroes in the Makefile to enable the 64-bit file APIs.
//                 -D_LARGEFILE64_SOURCE -D_FILE_OFFSET_BITS=64
//
//                Modify to use the 64-bit file APIs to access file to support 64-bit file length.
//                e.g. :
//                  lseek64(), fseeko64, and etc.
//
//             Please refer to the "btfiles.cpp" and "btfiles.h" files for details.
//             The modified codes are enabled by the M_NH220 macro.
//             
//
//     5) Modify the peerlist.cpp and btcontent.cpp files to print the running status of
//        CTorrent to the status file, including the following :
//        a) The status for downloading, including NO# of peers, pieces of files.
//           In the following routine :
//             PeerList::FillFDSET()
//
//        b) The status for checking the existing pieces already downloaded.
//           In the following routine :
//             btContent::CheckExist()
//
//    v1.01  09/02/2004 
//     1) Add 2 command line parameters to limit the maximum bandwidth for 
//        download & upload separatedly.
//
//          -Y  :  maximum bandwidth for download
//          -Z  :  maximum bandwidth for upload 
//
//     2) Reduce the memory used by the P2P client :
//        a) How to calculate the NO# of bytes of memory used by the P2P client ?
//           The NO# of bytes used is caculated as below :
//             (Aroud 1.1MB for program operation) + ( (NO# of peers) * (MAX_BUF_SIZ) )
//
//        b) We changed the following parameters to reduce memory usage :
//               maximum NO# of peers to 10
//               minimum NO# of peers to 1
//               MAX_BUF_SIZ : 132KB (defined in bufio.cpp)
//           
//           So, the maximon bytes of memory used is :
//             (10 peers * 132KB) + (around 1.1MB) = around 2.4MB 
//   
//        c) For NAS with 64MB SDRAM :
//             5 FTP connections + 5 Samba connections + 10 P2P torrent processes :
//             2.6MB + 8MB + 24MB = 34.6MB
//
//    v1.02  12/31/2004
//      1) Bug fixed : 
//           In the btfiles.cpp file, routine btFiles::BuildFromMI() :
//             If torrent file does not contain the "files" parameters, and user specify
//              "-s" command line parameter, the following codes will be executed correctly.
//             1) fill the "saveas" into "m_directory" ==> directory name specified by user.
//             2) file the "path" into filename in torrent file for download.
//
//             The original codes does incorrectly: 
//                The files's name in torrent file will be replaced with "saveas".
//
//    v1.03  01/10/2005
//      1) Bug fixed :
//           Problem :
//             If run several P2P clients simultaneously & continuously to download 
//             different torrent files,  sometimes system crashed as "data bus error".
//           Reason :
//             In the btFiles::BuildFromMI() routine, it call strcpy() to copy string
//             into m_btfhead->bf_filename variable before allocate memory for it.
//           Solution :
//             In the btFiles::BuildFromMI() routine,  allocate memory for the
//             m_btfhead->bf_filename variable before strcpy() to it. 
//
//    v1.04  09/02/2005
//      1) Upgrade P2P BT client from "ctorrent v1.3.2" to "ctorrent v1.3.4-dnh1".
//         So, we modify by applying the patches of NH-220 to it.
//         - Change the maximum NO# of peers to 50
//         - Change the minimum NO# of peers to 10
//         - Change the size of memory used by a peer to reduce memory used by BT client.
//
//      2) Add the MultiTracker support.
//
//
// Notes :
//   08/23/2004
//     1) The BitField.NBits() routine return NO# of pieces in "size_t" (32-bit signed int).
//        So, the maximum NO# of pieces is : (2 ** 32)/2
//        The size of pieces is variable (for ex: 256KB).
//
//        The maximum size of files to download is :  512K * 1GB
//
//   09/02/2004
//     1) We cannot limit of maximu NO# of peers for download and upload separatedly.
//
//        Because P2P client is peer-to-peer, everyone can download/upload file(s) from/to
//        its peer(s).  The remote peer of the P2P client may contain the downloading and 
//        uploading behaviors both.
//        So, we cannot identify a peer as pure downloading only or pure uploading only.
//
// The remaining issues :
//   12/31/2004 :
//     1) The BT client supports only HTTP/TCP to authenticate with Tracker Server.
//        The BT client cannot authenticate with the following kind of Tracker Server(s) :
//           udp://tracker.bitcomet.net        , it use UDP        
//
//        And this Tracker Server uses additional UDP payload to do private actions.
//        Of course, this BT client cannot parse/add this according UDP payload.
//        So, you will see the Tracker Refuse Count increased in torrent status file.
//
//
//
//

#include <sys/types.h>

#include "./def.h"

#ifdef WINDOWS
#include <windows.h>
#else
#include <unistd.h>
#include <signal.h>
#endif

#include <sys/time.h>
#include <time.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "btconfig.h"
#include "btcontent.h"
#include "downloader.h"
#include "peerlist.h"
#include "tracker.h"

#include "./config.h"

#ifndef WINDOWS
#include "sigint.h"
#endif

#ifdef M_NH220
  #define STR_VERSION  "P2P Client for BitTorrent network  ver.1.04(09/02/2005)"

  // The name of status file for P2P to save the current active status
  char sz_status_file[MAXPATHLEN];
#endif

void usage();
int param_check(int argc, char **argv);


#ifdef WINDOWS

int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrzevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
}

#else

void Random_init()
{
  struct timeval tv; 
  gettimeofday(&tv,(struct timezone*) 0);
  unsigned int seed = tv.tv_usec + tv.tv_sec + getpid();
  return srandom(seed);
}

int main(int argc, char **argv)
{
#ifdef M_NH220
  // Display the version of P2P client
  printf("%s\n", STR_VERSION);
#endif
  Random_init();
  arg_user_agent = new char[MAX_PF_LEN];
  strcpy(arg_user_agent,PEER_PFX);
    
  if( argc < 2 || param_check(argc,argv) < 0 ){
    usage();
    exit(1);
  }

  if( arg_flg_make_torrent ){
    if( !arg_announce ){ fprintf(stderr,"please use -u to specify a announce url!\n"); exit(1);}
    if( !arg_save_as ){ fprintf(stderr,"please use -s to specify a metainfo file name!\n"); exit(1);}
    if( BTCONTENT.InitialFromFS(arg_metainfo_file, arg_announce,  arg_piece_length) < 0 ||
	BTCONTENT.CreateMetainfoFile(arg_save_as) < 0){
      fprintf(stderr,"create metainfo failed.\n");
      exit(1);
    }
    printf("create metainfo file %s successful.\n", arg_save_as);
    exit(0);
  }
  
  if( BTCONTENT.InitialFromMI(arg_metainfo_file, arg_save_as) < 0){
    fprintf(stderr,"error,initial meta info failed.\n");
    exit(1);
  }

  if( !arg_flg_exam_only && !arg_flg_check_only){
    if(WORLD.Initial_ListenPort() < 0){
      fprintf(stderr,"warn, you couldn't accept connection.\n");
    }else 
      printf("Listen on: %d\n",cfg_listen_port);

	Tracker.Initial();

	signal(SIGPIPE,SIG_IGN);
    signal(SIGINT,sig_catch);
    signal(SIGTERM,sig_catch);
    Downloader();
  }
  if( cfg_cache_size ) BTCONTENT.FlushCache();
  if( arg_bitfield_file ) BTCONTENT.pBF->WriteToFile(arg_bitfield_file);
  WORLD.CloseAll();

  exit(0);
}

#endif

int param_check(int argc, char **argv)
{
  int c, l;
#ifdef M_NH220
  while ( ( c = getopt(argc,argv,"b:c:e:fl:M:m:n:P:p:s:tu:Y:Z:vxhH")) != -1)
#else
  while ( ( c = getopt(argc,argv,"b:cC:D:e:fl:M:m:n:P:p:s:tu:U:vxhH")) != -1)
#endif
    switch( c ){
    case 'b':
      arg_bitfield_file = new char[strlen(optarg) + 1];
#ifndef WINDOWS
      if( !arg_bitfield_file ) return -1;
#endif
      strcpy(arg_bitfield_file, optarg);
      break;

    case 'p':			// listen on Port XXXX
      cfg_listen_port = atoi(optarg);
      break;

    case 's':			// Save as FILE/DIR NAME
      if( arg_save_as ) return -1;
      arg_save_as = new char[strlen(optarg) + 1];
#ifndef WINDOWS
      if( !arg_save_as ) return -1;
#endif
      strcpy(arg_save_as,optarg);
      break;

    case 'e':			// Exit while complete
      cfg_seed_hours = atoi(optarg);
      break;

    case 'c':			// Check exist only
      arg_flg_check_only = 1;
      break;

#ifdef M_NH220
#else
    case 'C':
      cfg_cache_size = atoi(optarg);
      break;
#endif
      
    case 'M':			// Max peers
      cfg_max_peers = atoi(optarg);
#ifdef M_NH220
      if (cfg_max_peers > MAX_OF_PEERS || cfg_max_peers < MIN_OF_PEERS) {
#else
      if( cfg_max_peers > 1000 || cfg_max_peers < 20){
#endif
	return -1;
      }
      break;
      
    case 'm':			// Min peers
      cfg_min_peers = atoi(optarg);
#ifdef M_NH220
      if (cfg_min_peers > MAX_OF_PEERS || cfg_min_peers < MIN_OF_PEERS) {
#else
      if( cfg_min_peers > 1000 || cfg_min_peers < 20){
#endif
	return -1;
      }
      break;

    case 'n':                  // Which file download
      arg_file_to_download = atoi(optarg);
    break;


    case 'f':			// force seed mode, skip sha1 check when startup.
      arg_flg_force_seed_mode = 1;
      break;
      
#ifdef M_NH220
    case 'Y':
      // Convert the specified Kbps to Bps for Rate::RateMeasure()
      cfg_max_bandwidth_down = (int)(strtod(optarg, NULL));
      if (cfg_max_bandwidth_down > MAX_OF_BANDWIDTH || cfg_max_bandwidth_down < MIN_OF_BANDWIDTH) {
        printf("Error: invalid input for maximum bandwidth for download.\n");
        return (-1);
      }
      cfg_max_bandwidth_down = (cfg_max_bandwidth_down * 1024) / 8;
#else
    case 'D':
      cfg_max_bandwidth_down = (int)(strtod(optarg, NULL) * 1024);
#endif
      break;

#ifdef M_NH220
    case 'Z':
      // Convert the specified Kbps to Bps for Rate::RateMeasure()
      cfg_max_bandwidth_up = (int)(strtod(optarg, NULL));
      if (cfg_max_bandwidth_up > MAX_OF_BANDWIDTH || cfg_max_bandwidth_up < MIN_OF_BANDWIDTH) {
        printf("Error: invalid input for maximum bandwidth for upload.\n");
        return (-1);
      }
      cfg_max_bandwidth_up = (cfg_max_bandwidth_up * 1024) / 8; 
#else
    case 'U':
      cfg_max_bandwidth_up = (int)(strtod(optarg, NULL) * 1024);
#endif
         break;

    case 'P':
		l = strlen(optarg);
		if (l > MAX_PF_LEN) {printf("-P arg must be 8 or less characters\n"); exit(1);}
		if (l == 1 && *optarg == '-') *arg_user_agent = (char) 0;
		else strcpy(arg_user_agent,optarg);
      break;

     // BELLOW OPTIONS USED FOR CREATE TORRENT.
    case 'u':			// Announce url
      if( arg_announce ) return -1;
      arg_announce = new char[strlen(optarg) + 1];
      strcpy(arg_announce, optarg);
      break;

    case 't':			// make Torrent
      arg_flg_make_torrent = 1;
      break;

    case 'l':			// piece Length (default 262144)
      arg_piece_length = atoi(optarg);
      if( arg_piece_length < 65536 ||
	  arg_piece_length > 1310720 ){
	// warn message:
	// piece length range is 65536 =>> 1310720
	return -1;
      }
      break;

    case 'x':
      arg_flg_exam_only = 1;
      break;

    case 'v':
      arg_verbose = 1;
      break;

    case 'h':
    case 'H':
    default:
      //unknow option.
      return -1;
    }

  argc -= optind; argv += optind;
  if( cfg_min_peers >= cfg_max_peers ) cfg_min_peers = cfg_max_peers - 1;
  if( argc != 1 ) return -1;
  arg_metainfo_file = new char[strlen(*argv) + 1];
  
#ifndef WINDOWS
  if( !arg_metainfo_file ) return -1;
#endif
  strcpy(arg_metainfo_file, *argv);

#ifdef M_NH220
  strcpy(sz_status_file, arg_metainfo_file);
  strcat(sz_status_file, ".status");    // status file name :   *.torrent.status
  printf("The status of P2P is stored in  %s  file.\n", sz_status_file);
#endif
  
  return 0;
}

void usage()
{
  fprintf(stderr,"%s		Copyright: YuHong(992126018601033)",PACKAGE_STRING);
  fprintf(stderr,"\nWARNING: THERE IS NO WARRANTY FOR CTorrent. USE AT YOUR OWN RISK!!!\n");
  fprintf(stderr,"\nGeneric Options:\n");
  fprintf(stderr,"-h/-H\t\tShow this message.\n");
  fprintf(stderr,"-x\t\tDecode metainfo(torrent) file only, don't download.\n");
  fprintf(stderr,"-c\t\tCheck exist only. don't download.\n");
  fprintf(stderr,"-v\t\tVerbose output (for debugging).\n");
  fprintf(stderr,"\nDownload Options:\n");
  fprintf(stderr,"-e int\t\tExit while seed <int> hours later. (default 72 hours)\n");
  fprintf(stderr,"-p port\t\tListen port. (default 2706 -> 2106)\n");
  fprintf(stderr,"-s save_as\tSave file/directory/metainfo as... \n");

#ifdef M_NH220
#else
  fprintf(stderr,"-C cache_size\tCache size,unit MB. (default 16MB)\n");
#endif

  fprintf(stderr,"-f\t\tForce seed mode. skip hash check at startup.\n");
  fprintf(stderr,"-b bf_filename\tBit field filename. (use it carefully)\n");

#ifdef M_NH220
  fprintf(stderr,"-M max_peers\tMax peers count. (%d ~ %d, default %d)\n",
			MIN_OF_PEERS, MAX_OF_PEERS, cfg_max_peers);
  fprintf(stderr,"-m min_peers\tMin peers count. (%d ~ %d, default %d)\n",
			MIN_OF_PEERS, MAX_OF_PEERS, cfg_min_peers);
#else
  fprintf(stderr,"-M max_peers\tMax peers count.\n");
  fprintf(stderr,"-m min_peers\tMin peers count.\n");
#endif
  fprintf(stderr,"-n file_number\tWhich file download.\n");
#ifdef M_NH220
  fprintf(stderr,"-Y rate\t\tMaximum bandwidth for download (%d ~ %d Kbps, default %d)\n",
                        MIN_OF_BANDWIDTH, MAX_OF_BANDWIDTH, cfg_max_bandwidth_down);
  fprintf(stderr,"-Z rate\t\tMaximum bandwidth for upload (%d ~ %d Kbps, default %d)\n",
                        MIN_OF_BANDWIDTH, MAX_OF_BANDWIDTH, cfg_max_bandwidth_up);
#else
  fprintf(stderr,"-D rate\t\tMax bandwidth down (unit KB/s)\n");
  fprintf(stderr,"-U rate\t\tMax bandwidth up (unit KB/s)\n");
#endif
  fprintf(stderr,"-P peer_id\tSet Peer ID ["PEER_PFX"]\n");
  fprintf(stderr,"\nMake metainfo(torrent) file Options:\n");
  fprintf(stderr,"-t\t\tWith make torrent. must specify this option.\n");
  fprintf(stderr,"-u url\t\tTracker's url.\n");
  fprintf(stderr,"-l piece_len\tPiece length.(default 262144)\n");
  fprintf(stderr,"\neg.\n");
  fprintf(stderr,"hong> ctorrent -s new_filename -e 12 -C 32 -p 6881 eg.torrent\n\n");
  fprintf(stderr,"home page: http://ctorrent.sourceforge.net/\n");
  fprintf(stderr,"bug report: %s\n\n",PACKAGE_BUGREPORT);
}
