/* 
   Unix SMB/Netbios implementation.
   Version 3.0
   handle NLTMSSP, server side

   Copyright (C) Andrew Tridgell      2001
   Copyright (C) Andrew Bartlett 2001-2003

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "includes.h"
#include "sp_auth_dbg.h"
extern char cl_NativeOS[128];
extern char cl_NativeLanMan[128];
extern char cl_PrimaryDomain[128];
auth_usersupplied_info cl_user_info;
char* cl_name="dh is here";
char cl_user[128];
char cl_domain[128];
char cl_workstation[128];
char cl_cl_domain[128];
char cl_in_uname[128];

			
/**
 * Return the challenge as determined by the authentication subsystem 
 * @return an 8 byte random challenge
 */

static const uint8 *auth_ntlmssp_get_challenge(const struct ntlmssp_state *ntlmssp_state)
{
	sp_suth_ntlmssp_dbg(0, ("%s: %s..:\n", __FILE__, __func__));
	AUTH_NTLMSSP_STATE *auth_ntlmssp_state = ntlmssp_state->auth_context;
	return auth_ntlmssp_state->auth_context->get_ntlm_challenge(auth_ntlmssp_state->auth_context);
}

/**
 * Some authentication methods 'fix' the challenge, so we may not be able to set it
 *
 * @return If the effective challenge used by the auth subsystem may be modified
 */
static BOOL auth_ntlmssp_may_set_challenge(const struct ntlmssp_state *ntlmssp_state)
{
	sp_suth_ntlmssp_dbg(0, ("%s: %s..:\n", __FILE__, __func__));
	AUTH_NTLMSSP_STATE *auth_ntlmssp_state = ntlmssp_state->auth_context;
	struct auth_context *auth_context = auth_ntlmssp_state->auth_context;

	return auth_context->challenge_may_be_modified;
}

/**
 * NTLM2 authentication modifies the effective challange, 
 * @param challenge The new challenge value
 */
static NTSTATUS auth_ntlmssp_set_challenge(struct ntlmssp_state *ntlmssp_state, DATA_BLOB *challenge)
{	sp_suth_ntlmssp_dbg(0, ("%s: %s..:\n", __FILE__, __func__));
	AUTH_NTLMSSP_STATE *auth_ntlmssp_state = ntlmssp_state->auth_context;
	struct auth_context *auth_context = auth_ntlmssp_state->auth_context;

	SMB_ASSERT(challenge->length == 8);

	auth_context->challenge = data_blob_talloc(auth_context->mem_ctx, 
						   challenge->data, challenge->length);

	auth_context->challenge_set_by = "NTLMSSP callback (NTLM2)";

	DEBUG(5, ("auth_context challenge set by %s\n", auth_context->challenge_set_by));
	DEBUG(5, ("challenge is: \n"));
	dump_data(5, (const char *)auth_context->challenge.data, auth_context->challenge.length);
	return NT_STATUS_OK;
}

/**
 * Check the password on an NTLMSSP login.  
 *
 * Return the session keys used on the connection.
 */
#ifdef M_BUG_XP_LOGON
static NTSTATUS auth_ntlmssp_check_password(struct ntlmssp_state *ntlmssp_state, DATA_BLOB *user_session_key, DATA_BLOB *lm_session_key, int suid) 
#else
static NTSTATUS auth_ntlmssp_check_password(struct ntlmssp_state *ntlmssp_state, DATA_BLOB *user_session_key, DATA_BLOB *lm_session_key) 
#endif
{	
	sp_suth_ntlmssp_dbg(0, ("%s: %s..:\n", __FILE__, __func__));
	AUTH_NTLMSSP_STATE *auth_ntlmssp_state = ntlmssp_state->auth_context;
	auth_usersupplied_info *user_info = NULL;
	NTSTATUS nt_status;

	/* the client has given us its machine name (which we otherwise would not get on port 445).
	   we need to possibly reload smb.conf if smb.conf includes depend on the machine name */

	set_remote_machine_name(auth_ntlmssp_state->ntlmssp_state->workstation, True);

	/* setup the string used by %U */
	/* sub_set_smb_name checks for weird internally */
	sub_set_smb_name(auth_ntlmssp_state->ntlmssp_state->user);

	reload_services(True);
#ifdef M_BUG_XP_LOGON
	
	nt_status = make_user_info_map(&user_info, 
			       auth_ntlmssp_state->ntlmssp_state->user, 
			       auth_ntlmssp_state->ntlmssp_state->domain, 
			       auth_ntlmssp_state->ntlmssp_state->workstation, 
                               auth_ntlmssp_state->ntlmssp_state->lm_resp.data ? &auth_ntlmssp_state->ntlmssp_state->lm_resp : NULL, 
                               auth_ntlmssp_state->ntlmssp_state->nt_resp.data ? &auth_ntlmssp_state->ntlmssp_state->nt_resp : NULL, 
			       NULL, NULL, NULL,
			       True);
	
	if(suid==0){
		sp_suth_ntlmssp_dbg(0, ("%s: %s..:if(first){..clear the user info\n", __FILE__, __func__));
					
		strcpy(cl_user, user_info->smb_name.str);
		strcpy(cl_domain, user_info->domain.str);
		strcpy(cl_workstation, user_info->wksta_name.str);
		strcpy(cl_cl_domain, user_info->client_domain.str);
		strcpy(cl_in_uname, user_info->internal_username.str);
		
		sp_suth_ntlmssp_dbg(0, ("%s: %s..:cl_user_info.smb_name.str=%s\n", __FILE__, __func__, cl_user_info.smb_name.str ));
		sp_suth_ntlmssp_dbg(0, ("%s: %s..:cl_user=%s cl_domain=%s cl_workstation=%s\n", __FILE__, __func__, cl_user, cl_domain, cl_workstation ));
		sp_suth_ntlmssp_dbg(0, ("%s: %s..:user_info->client_domain.str=%s user_info->internal_username.str=%s\n", __FILE__, __func__, user_info->client_domain.str, user_info->internal_username.str ));
		sp_suth_ntlmssp_dbg(0, ("%s: %s..:cl_user=%s cl_domain=%s cl_workstation=%s \n", __FILE__, __func__, cl_user, cl_domain, cl_workstation));
		sp_suth_ntlmssp_dbg(0, ("%s: %s..:user_info->client_domain.str=%s user_info->smb_name.str=%s user_info->wksta_name.str=%s user_info->domain.str=%s user_info->internal_username.str=%s\n", __FILE__, __func__, 
				user_info->client_domain.str, user_info->smb_name.str, user_info->wksta_name.str, user_info->domain.str, user_info->internal_username.str));		
	}
			
#else
	sp_suth_ntlmssp_dbg(0, ("%s: %s..:if(suid==%d){\n", __FILE__, __func__, suid));
	
	nt_status = make_user_info_map(&user_info, 
				       auth_ntlmssp_state->ntlmssp_state->user, 
				       auth_ntlmssp_state->ntlmssp_state->domain, 
				       auth_ntlmssp_state->ntlmssp_state->workstation, 
	                               auth_ntlmssp_state->ntlmssp_state->lm_resp.data ? &auth_ntlmssp_state->ntlmssp_state->lm_resp : NULL, 
	                               auth_ntlmssp_state->ntlmssp_state->nt_resp.data ? &auth_ntlmssp_state->ntlmssp_state->nt_resp : NULL, 
				       NULL, NULL, NULL,
				       True);
				       
#endif//#ifdef M_BUG_XP_LOGON
				       
	if (!NT_STATUS_IS_OK(nt_status)) {
		return nt_status;
	}
	
#ifdef M_BUG_XP_LOGON
	NTSTATUS nt_status1;
	sp_suth_ntlmssp_dbg(0, ("%s: %s..:before auth_ntlmssp_state->auth_context->check_ntlm_password\n", __FILE__, __func__));
	nt_status = auth_ntlmssp_state->auth_context->check_ntlm_password(auth_ntlmssp_state->auth_context, 
									  user_info, &auth_ntlmssp_state->server_info); 
	sp_suth_ntlmssp_dbg(0, ("%s: %s..:after auth_ntlmssp_state->auth_context->check_ntlm_password\n", __FILE__, __func__));
//sp/230	
	//while(1){}
/*
	sp_suth_ntlmssp_dbg(0, ("%s: %s..:2..auth_ntlmssp_state->server_info.guest=%d uid=%d gid=%d sam_account=%s unix_name=%s\n", __FILE__, __func__, 
		auth_ntlmssp_state->server_info->guest, 
		auth_ntlmssp_state->server_info->uid, 
		auth_ntlmssp_state->server_info->gid,
		auth_ntlmssp_state->server_info->sam_account,
		auth_ntlmssp_state->server_info->unix_name
	));
*/	
	sp_suth_ntlmssp_dbg(0, ("%s: %s..:a1\n", __FILE__, __func__));
	//while(1){}
	
	if( suid==0 ) {
#if defined(HAVE_IMMEDIATE_STRUCTURES)
		if(nt_status.v != 0) {
#else
		if(nt_status != 0) {
#endif//#if defined(HAVE_IMMEDIATE_STRUCTURES)

			user_info->client_domain.str[0]=0;
			user_info->smb_name.str[0]=0;
			user_info->wksta_name.str[0]=0;
			user_info->domain.str[0]=0;
			user_info->internal_username.str[0]=0;
			
			nt_status1 = auth_ntlmssp_state->auth_context->check_ntlm_password(auth_ntlmssp_state->auth_context, 
										  user_info, &auth_ntlmssp_state->server_info); 
										  
			sp_suth_ntlmssp_dbg(0, ("%s: %s..:3..auth_ntlmssp_state->server_info.guest=%d uid=%d gid=%d sam_account=%s unix_name=%s\n", __FILE__, __func__, 
					auth_ntlmssp_state->server_info->guest, 
					auth_ntlmssp_state->server_info->uid, 
					auth_ntlmssp_state->server_info->gid,
					auth_ntlmssp_state->server_info->sam_account,
					auth_ntlmssp_state->server_info->unix_name
			));
#if defined(HAVE_IMMEDIATE_STRUCTURES)
			sp_suth_ntlmssp_dbg(0, ("%s: %s..:nt_status.v=%d nt_status1.v=%d\n", __FILE__, __func__, nt_status.v, nt_status1.v));	
			nt_status.v= nt_status.v & nt_status1.v;
#else
			sp_suth_ntlmssp_dbg(0, ("%s: %s..:nt_status=%d nt_status1=%d\n", __FILE__, __func__, nt_status, nt_status1));	
			nt_status= nt_status & nt_status1;
#endif//#if defined(HAVE_IMMEDIATE_STRUCTURES)
			
		}
	}
	
#else
	nt_status = auth_ntlmssp_state->auth_context->check_ntlm_password(auth_ntlmssp_state->auth_context, 
									  user_info, &auth_ntlmssp_state->server_info); 
#endif//#ifdef M_BUG_XP_LOGON

									  
	sp_suth_ntlmssp_dbg(0, ("%s: %s..:2..nt_status=%d\n", __FILE__, __func__, nt_status));
	free_user_info(&user_info);

	if (!NT_STATUS_IS_OK(nt_status)) {
		return nt_status;
	}
	if (auth_ntlmssp_state->server_info->user_session_key.length) {
		DEBUG(10, ("Got NT session key of length %u\n", auth_ntlmssp_state->server_info->user_session_key.length));
		sp_suth_ntlmssp_dbg(0, ("%s: %s..:Got NT session key of length %u\n", __FILE__, __func__, auth_ntlmssp_state->server_info->user_session_key.length));
		*user_session_key = data_blob_talloc(auth_ntlmssp_state->mem_ctx, 
						   auth_ntlmssp_state->server_info->user_session_key.data,
						   auth_ntlmssp_state->server_info->user_session_key.length);
	}
	if (auth_ntlmssp_state->server_info->lm_session_key.length) {
		DEBUG(10, ("Got LM session key of length %u\n", auth_ntlmssp_state->server_info->lm_session_key.length));
		sp_suth_ntlmssp_dbg(0, ("%s: %s..:Got LM session key of length %u\n", __FILE__, __func__, auth_ntlmssp_state->server_info->lm_session_key.length));
		*lm_session_key = data_blob_talloc(auth_ntlmssp_state->mem_ctx, 
						   auth_ntlmssp_state->server_info->lm_session_key.data,
						   auth_ntlmssp_state->server_info->lm_session_key.length);
	}
	sp_suth_ntlmssp_dbg(0, ("%s: %s..:return nt_status=%d\n", __FILE__, __func__, nt_status));
	return nt_status;
}

NTSTATUS auth_ntlmssp_start(AUTH_NTLMSSP_STATE **auth_ntlmssp_state)
{	sp_suth_ntlmssp_dbg(0, ("%s: %s..:\n", __FILE__, __func__));
	NTSTATUS nt_status;
	TALLOC_CTX *mem_ctx;

	mem_ctx = talloc_init("AUTH NTLMSSP context");
	
	*auth_ntlmssp_state = talloc_zero(mem_ctx, sizeof(**auth_ntlmssp_state));
	if (!*auth_ntlmssp_state) {
		DEBUG(0,("auth_ntlmssp_start: talloc failed!\n"));
		talloc_destroy(mem_ctx);
		return NT_STATUS_NO_MEMORY;
	}

	ZERO_STRUCTP(*auth_ntlmssp_state);

	(*auth_ntlmssp_state)->mem_ctx = mem_ctx;

	if (!NT_STATUS_IS_OK(nt_status = ntlmssp_server_start(&(*auth_ntlmssp_state)->ntlmssp_state))) {
		sp_suth_ntlmssp_dbg(0, ("%s: %s..:1\n", __FILE__, __func__));
		return nt_status;
	}

	if (!NT_STATUS_IS_OK(nt_status = make_auth_context_subsystem(&(*auth_ntlmssp_state)->auth_context))) {
		sp_suth_ntlmssp_dbg(0, ("%s: %s..:2\n", __FILE__, __func__));
		return nt_status;
	}

	(*auth_ntlmssp_state)->ntlmssp_state->auth_context = (*auth_ntlmssp_state);
	(*auth_ntlmssp_state)->ntlmssp_state->get_challenge = auth_ntlmssp_get_challenge;
	(*auth_ntlmssp_state)->ntlmssp_state->may_set_challenge = auth_ntlmssp_may_set_challenge;
	(*auth_ntlmssp_state)->ntlmssp_state->set_challenge = auth_ntlmssp_set_challenge;
	(*auth_ntlmssp_state)->ntlmssp_state->check_password = auth_ntlmssp_check_password;
	(*auth_ntlmssp_state)->ntlmssp_state->server_role = lp_server_role();
	sp_suth_ntlmssp_dbg(0, ("%s: %s..:return NT_STATUS_OK;\n", __FILE__, __func__));
	return NT_STATUS_OK;
}

void auth_ntlmssp_end(AUTH_NTLMSSP_STATE **auth_ntlmssp_state)
{	sp_suth_ntlmssp_dbg(0, ("%s: %s..:\n", __FILE__, __func__));
	TALLOC_CTX *mem_ctx = (*auth_ntlmssp_state)->mem_ctx;

	if ((*auth_ntlmssp_state)->ntlmssp_state) {
		ntlmssp_end(&(*auth_ntlmssp_state)->ntlmssp_state);
	}
	if ((*auth_ntlmssp_state)->auth_context) {
		((*auth_ntlmssp_state)->auth_context->free)(&(*auth_ntlmssp_state)->auth_context);
	}
	if ((*auth_ntlmssp_state)->server_info) {
		free_server_info(&(*auth_ntlmssp_state)->server_info);
	}
	talloc_destroy(mem_ctx);
	*auth_ntlmssp_state = NULL;
}

#ifdef M_BUG_XP_LOGON
NTSTATUS auth_ntlmssp_update(AUTH_NTLMSSP_STATE *auth_ntlmssp_state, 
			     const DATA_BLOB request, DATA_BLOB *reply, int suid) 
#else
NTSTATUS auth_ntlmssp_update(AUTH_NTLMSSP_STATE *auth_ntlmssp_state, 
			     const DATA_BLOB request, DATA_BLOB *reply) 
#endif
{	
	sp_suth_ntlmssp_dbg(0, ("%s: %s..:\n", __FILE__, __func__));
#ifdef M_BUG_XP_LOGON
	return ntlmssp_update(auth_ntlmssp_state->ntlmssp_state, request, reply, suid);
#else
	return ntlmssp_update(auth_ntlmssp_state->ntlmssp_state, request, reply);
#endif
}
