/* 
   Unix SMB/CIFS implementation.

   Safe versions of getpw* calls

   Copyright (C) Andrew Bartlett 2002
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "includes.h"
#include "sp_lib_dbg.h"

static struct passwd *alloc_copy_passwd(const struct passwd *from) 
{	sp_util_pw_dbg(0, ("%s: %s..:\n", __FILE__, __func__));
	struct passwd *ret = smb_xmalloc(sizeof(struct passwd));
	ZERO_STRUCTP(ret);
	ret->pw_name = smb_xstrdup(from->pw_name);
	ret->pw_passwd = smb_xstrdup(from->pw_passwd);
	ret->pw_uid = from->pw_uid;
	ret->pw_gid = from->pw_gid;
	ret->pw_gecos = smb_xstrdup(from->pw_gecos);
	ret->pw_dir = smb_xstrdup(from->pw_dir);
	ret->pw_shell = smb_xstrdup(from->pw_shell);
	return ret;
}

void passwd_free (struct passwd **buf)
{
	if (!*buf) {
		DEBUG(0, ("attempted double-free of allocated passwd\n"));
		return;
	}

	SAFE_FREE((*buf)->pw_name);
	SAFE_FREE((*buf)->pw_passwd);
	SAFE_FREE((*buf)->pw_gecos);
	SAFE_FREE((*buf)->pw_dir);
	SAFE_FREE((*buf)->pw_shell);

	SAFE_FREE(*buf);
}

struct passwd *getpwnam_alloc(const char *name) 
{
	struct passwd *temp;
	//sp_util_pw_dbg(0, ("%s: %s..:\n", __FILE__, __func__));
	sp_util_pw_dbg(0, ("%s: %s..:name=%s\n", __FILE__, __func__, name));
	
	temp = sys_getpwnam(name);
	
	if (!temp) {
/*		
		sp_util_pw_dbg(0, ("%s: %s..:if (!temp) {.. errno=%d strerror(errno)=%s\n", __FILE__, __func__, errno, strerror(errno)));
		if(errno==0 || errno==ENOENT || errno==ESRCH || errno==EBADF || errno==EPERM )
			sp_util_pw_dbg(0, ("%s: %s..:if(errno==0 || errno==ENOENT || errno==ESRCH || errno==EBADF || errno==EPERM )\n", __FILE__, __func__));
		else if(errno==ENOMEM )
			sp_util_pw_dbg(0, ("%s: %s..:ENOMEM\n", __FILE__, __func__));
		else if(errno==EIO )
			sp_util_pw_dbg(0, ("%s: %s..:EIO\n", __FILE__, __func__));
		else if(errno==EINTR )
			sp_util_pw_dbg(0, ("%s: %s..:EINTR\n", __FILE__, __func__));
		else if(errno==EMFILE )
			sp_util_pw_dbg(0, ("%s: %s..:EMFILE\n", __FILE__, __func__));
		else if(errno==ENFILE )
			sp_util_pw_dbg(0, ("%s: %s..:ENFILE\n", __FILE__, __func__));
			

		temp = sys_getpwnam("all");
		if(temp){
			sp_util_pw_dbg(0, ("%s: %s..:temp->pw_name=%s\n", __FILE__, __func__, temp->pw_name));
			sp_util_pw_dbg(0, ("%s: %s..:temp->pw_passwd=%s\n", __FILE__, __func__, temp->pw_passwd));
			sp_util_pw_dbg(0, ("%s: %s..:temp->pw_uid=%d\n", __FILE__, __func__, temp->pw_uid));
			sp_util_pw_dbg(0, ("%s: %s..:temp->pw_gid=%d\n", __FILE__, __func__, temp->pw_gid));
			sp_util_pw_dbg(0, ("%s: %s..:temp->pw_gecos=%s\n", __FILE__, __func__, temp->pw_gecos));
			sp_util_pw_dbg(0, ("%s: %s..:temp->pw_dir=%s\n", __FILE__, __func__, temp->pw_dir));
			sp_util_pw_dbg(0, ("%s: %s..:temp->pw_shell=%s\n", __FILE__, __func__, temp->pw_shell));
			
			temp=NULL;
		}else{
			struct passwd temp1;
			FILE* fp; 
			
			if( (fp=fopen("/etc/passwd", "r")) !=NULL )
				sp_util_pw_dbg(0, ("%s: %s..:if( (fp=fopen /etc/passwd !=NULL )\n", __FILE__, __func__));
			else
				sp_util_pw_dbg(0, ("%s: %s..:if( (fp=fopen /etc/passwd ==NULL )\n", __FILE__, __func__));
			
			temp1.pw_name = "nobody";
			temp1.pw_passwd = "";
			temp1.pw_uid = 32767;
			temp1.pw_gid = 32767;
			temp1.pw_gecos = "Nobody";
			temp1.pw_dir = "/var";
			temp1.pw_shell = "/bin/sh";
			
			//return alloc_copy_passwd(&temp1);
		}
*/

#if 0
		if (errno == ENOMEM) {
			/* what now? */
		}
#endif
		return NULL;
	}

	return alloc_copy_passwd(temp);
}

struct passwd *getpwuid_alloc(uid_t uid) 
{
	struct passwd *temp;

	temp = sys_getpwuid(uid);
	
	if (!temp) {
#if 0
		if (errno == ENOMEM) {
			/* what now? */
		}
#endif
		return NULL;
	}

	return alloc_copy_passwd(temp);
}
