/***************************************************************************************/  
/*  NTX_IGD v2 UPnP device by using Intel UPnP SDK v1.0.4                           */
/*                                                                                     */
/*  Revision 0.4.0                                                                     */
/*                                                                                     */
/*  Copyright (c) 2001, 2002 REDSonic Inc. All rights reserved.                        */
/***************************************************************************************/

#include "upnp_soho.h"
//sp
#include "common_util.h"
#include "sp_types.h"
#include "sp_inet.h"
#include "sp_upnp_dbg.h"

char *SohoDeviceType[] = {
  "urn:schemas-upnp-org:device:InternetGatewayDevice:1",
  "urn:schemas-upnp-org:device:WANDevice:1",
  "urn:schemas-upnp-org:device:WANConnectionDevice:1"
};

float stat_initial_val[4] = {0,0,0,0};
time_t stat_initial_time = 0;
time_t pm_time = 0;
long checkpoint = 0;

char *SohoDeviceUDN[] = {
  "uuid:upnp-InternetGatewayDevice-1_0-0006250411r0",
  "uuid:upnp-WANDevice-1_0-0006250411r0",
  "uuid:upnp-WANConnectionDevice-1_0-0006250411r0"
};

char *SohoServiceType[] = {
  "urn:schemas-upnp-org:service:Layer3Forwarding:1",
  "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
  "urn:schemas-upnp-org:service:WANIPConnection:1"
};

char *SohoServiceID[] = {
  "urn:upnp-org:serviceId:L3Forwarding1",
  "urn:upnp-org:serviceId:WANCommonInterfaceConfig",
  "urn:upnp-org:serviceId:WANIPConnection"
};

char *soho_l3f_varname[] = {"DefaultConnectionService"};
char soho_l3f_varval[SOHO_L3F_VARCOUNT][SOHO_MAX_VAL_LEN]; 
char *soho_l3f_def_val[] = {"WANIPConnection"};

char *soho_cic_varname[] = {
  "WANAccessType",
  "Layer1UpstreamMaxBitRate",
  "Layer1DownstreamMaxBitRate",
  "PhysicalLinkStatus",
  "MaximumActiveConnections",
  "TotalBytesSent",
  "TotalBytesReceived",
  "TotalPacketsSent",
  "TotalPacketsReceived",
  "PersonalFirewallEnabled",
  "X_Uptime"
};
char soho_cic_varval[SOHO_CIC_VARCOUNT][SOHO_MAX_VAL_LEN]; 
char *soho_cic_def_val[] = {"Ethernet","10000000","10000000","Up","254","0","0","0","0","0","0" };

char *soho_ipc_varname[] = {
  "ConnectionType",
  "PossibleConnectionTypes",
  "ConnectionStatus",
  "Uptime",
  "LastConnectionError",
  "RSIPAvailable",
  "NATEnabled",
  "ExternalIPAddress",
  "PortMappingNumberOfEntries",
  "X_Name"
};
char soho_ipc_varval[SOHO_IPC_VARCOUNT][SOHO_MAX_VAL_LEN]; 
char *soho_ipc_def_val[] = { "IP_Routed","IP_Routed","Connected","0","ERROR_NONE","1","1","192.168.1.1","0","Local Area Connection"};

struct SohoService soho_service_table[3];  

struct SohoPortMapping soho_portmapping_table[MAX_PORT_MAPPING_NUM];

int soho_portmapping_number = 0;
int default_advr_expire = 1800;
UpnpDevice_Handle device_handle = -1;
pthread_mutex_t SohoDevMutex = PTHREAD_MUTEX_INITIALIZER;
int upnp_update_enable = 1; /* by default enable update */

char *lanip = NULL;

int SohoDeviceConfigIP2(void) /* get the current running LAN ip */
{
  FILE *fp;
  char buf[MAX_LINE_LENGTH+1], *key, *value, *value2;
  int i=0,c;
  const char delimiters[] = "=\r\n\t";
	
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  if ((fp = fopen(UPNP_SOHO_IP2_FILE, "r")) == NULL) {    
    return -1;
  }  
  
  	while (!feof(fp)) {
    	c = fgetc(fp);     
    	if ( (!isspace(c)) && i == 0) {
      		if (i < MAX_LINE_LENGTH)
				buf[i++] = c;
    	} else if ((i > 0) && (c == '\n')) {
      		buf[i]='\0';
      		key = strtok(buf, delimiters);
//sp      		      		
      		//value = strtok(NULL, delimiters);
      		value = key + strlen(key) + 1;
      		//sp_upnp_sohoTrace("%s: %s..:key=%s value=%s\n", __FILE__, __func__, key, value);
      		value = strtok(value,"\t\r\n");    
      		//sp_upnp_sohoTrace("%s: %s..:key=%s value=%s\n", __FILE__, __func__, key, value);
      		
      		if ((key != NULL)&&(key[0] != '#')) {
				if (value !=NULL) {
	  				if ((value2 = strdup(value)) == NULL) 
	    				CommonUtil_MallocError();
				} else {
	  				value2 = NULL;
				}
					if (!strcmp(key, UPNP_SOHO_IP2_ITEM)) 
					{ 
						char value3[16];
						
						sp_upnp_sohoTrace("%s: %s..:key=%s value=%s\n", __FILE__, __func__, key, value);
						memset(value3, 0, 16);
						memcpy(value3, value2+1, strlen(value2)-2);
						//lanip = value2;
						lanip = strdup(value3);
						sp_upnp_sohoTrace("%s: %s..:key=%s lanip=%s\n", __FILE__, __func__, key, lanip);
					}
      		}
      		i = 0;
    	} else if (!isspace(c)) {
      		if (i < MAX_LINE_LENGTH)
				buf[i++] = c;
    	}
  	}
  	
  	fclose(fp);
  
  	return(0);
}

int SohoDeviceUpdateEnable(void) /* get if allow to update configuration via control point */
{
  FILE *fp;
  char buf[MAX_LINE_LENGTH+1], *key, *value, *value2,*ena;
  int i=0,c;
  const char delimiters[] = "=\r\n\t";
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  if ((fp = fopen(RCT_CONF_FILE, "r")) == NULL) {    
    return 0;
  }  
  ena=NULL;
  while (!feof(fp)) {
    c = fgetc(fp);     
    if ( (!isspace(c)) && i == 0) {
      if (i < MAX_LINE_LENGTH)
	buf[i++] = c;
    } else if ((i > 0) && (c == '\n')) {
      buf[i]='\0';
      key = strtok(buf, delimiters);
      value = strtok(NULL, delimiters);
      if ((key != NULL)&&(key[0] != '#')) {
	if (value !=NULL) {
	  if ((value2 = strdup(value)) == NULL) 
	    CommonUtil_MallocError();
	} else {
	  value2 = NULL;
	}
	if (!strcmp(key, "UPNP_UPDATE_ENABLE")) {
	  ena = value2;
	  break;
	} 
      }
      i = 0;
    } else if (!isspace(c)) {
      if (i < MAX_LINE_LENGTH)
	buf[i++] = c;
    }
  }
  fclose(fp);
//  fprintf(stderr,"2:UPNP_UPDATE_ENABLE=%s\n",ena);
//  fflush(stderr);

  if(ena != NULL && index(ena,'1')) return 1;

  return(0);
}

int SohoDescriptionFileParse(int port)
{
  FILE *fp,*fd;
  int c;
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  	if((!lanip)||(port < 1025))
  	{
  		sp_upnp_sohoTrace("%s: %s..:if((!lanip)||(port < 1025))\n", __FILE__, __func__);
    	return 1;
  	}
#ifdef NOVAC
	if ((fp = fopen("/etc/xml/desc_base_novac.xml", "r")) == NULL) 
#else
  	if ((fp = fopen("/etc/xml/desc_base_0.xml", "r")) == NULL) 
#endif  	
    {
  		sp_upnp_sohoTrace("%s: %s..:if ((fp = fopen../etc/xml/desc_base_0.xml \n", __FILE__, __func__);
    	return 1;
  	}
  
  //because this is a symbolic link
  //unlink("/xml/descrip.xml");  
#ifdef M_NH200  
  	if ((fd = fopen("/etc/xml/descrip.xml", "w")) == NULL) 
#else  	
  	//if ((fd = fopen("/xml/descrip.xml", "w")) == NULL) 
#endif  	
    {
  		sp_upnp_sohoTrace("%s: %s..:if ((fd = fopen../etc/xml/descrip.xml\n", __FILE__, __func__);
  		perror("fopen /etc/xml/descrip.xml fail");
    	return 1;
  	}
//sp  
  //fprintf(fd,"<?xml version=\"1.0\"?>\n<root xmlns=\"urn:schemas-upnp-org:device-1-0\">\n\t<specVersion>\n\t\t<major>1</major>\n\t\t<minor>0</minor>\n\t</specVersion>\n\t<URLBase>http://%s:%d</URLBase>\n", lanip, port);
  fprintf(fd,"<?xml version=\"1.0\"?>\n<root xmlns=\"urn:schemas-upnp-org:device-1-0\">\n\t<specVersion>\n\t\t<major>1</major>\n\t\t<minor>0</minor>\n\t</specVersion>\n\t<URLBase>http://%s:%d</URLBase>\n", lanip, 80);
  
  while (!feof(fp)) {
    c = fgetc(fp);
    if(c > 0 && c < 128) 
      fputc(c,fd);
  }
  fclose(fp);
  fclose(fd);
  sp_upnp_sohoTrace("%s: %s..:return(0);\n", __FILE__, __func__);
  return(0);
}

char* SohoParsePacketByte(int field)
{
  char *statval;
  char resultcode[100];
  float val[10];
  FILE *fp;
  char buf[300], *key, *value, *value2;
  int i=0;
  unsigned int c;
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  value2 = NULL;

  if(strcmp(soho_service_table[SOHO_SERVICE_CIC].VariableStrVal[3],"Down")
     && strcmp(soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[2],"Disconnected")) {  
    if ((fp = fopen("/proc/net/dev", "r")) == NULL)
      return NULL;
    
    while(!feof(fp)) {
      c = fgetc(fp);     
      if((i == 0) && (!isspace(c))) {
	buf[i++] = c;
      } else if((i>0)&&(c == '\n')) {
	buf[i]='\0';
	key = strdup(buf);
	value=key;
	while(key&&isspace(*key)) key++;
	if(key && (strncmp(key,"eth0:",5)==0)) {
	  key += 5;
	  value2 = strdup(key);
	  free(value);
	  break;
	}
	free(value);
	i = 0;
      } else {
	buf[i++] = c;
      }
    }
    fclose(fp);
    
    if(!value2) return NULL;
    
    sscanf(value2,"%f %f %f %f %f %f %f %f %f %f",&val[0],&val[1],&val[2],&val[3],&val[4],&val[5],&val[6],&val[7],&val[8],&val[9]);
    
    switch (field) {
    case TOTAL_BYTES_SENT:      
      sprintf(resultcode,"<NewTotalBytesSent>%.0f</NewTotalBytesSent>",(val[8] - stat_initial_val[0]));
      break;
    case TOTAL_BYTES_RECEIVED:
      sprintf(resultcode,"<NewTotalBytesReceived>%.0f</NewTotalBytesReceived>",(val[0] - stat_initial_val[1]));
      break;
    case TOTAL_PACKETS_SENT:
      sprintf(resultcode,"<NewTotalPacketsSent>%.0f</NewTotalPacketsSent>",(val[9] - stat_initial_val[2]));
      break;
    case TOTAL_PACKETS_RECEIVED:
      sprintf(resultcode,"<NewTotalPacketsReceived>%.0f</NewTotalPacketsReceived>",(val[1] - stat_initial_val[3]));
      break;
    default:
      return NULL;
    }      
    statval = strdup(resultcode); 
    return statval;
  }
  return NULL;
}

int SohoDeviceStateTableInit(char* DescDocURL)
{
  Upnp_Document DescDoc=NULL;
  int ret = UPNP_E_SUCCESS;
  char *servid_l3f=NULL, *evnturl_l3f=NULL, *ctrlurl_l3f=NULL;
  char *servid_cic=NULL, *evnturl_cic=NULL, *ctrlurl_cic=NULL;
  char *servid_ipc=NULL, *evnturl_ipc=NULL, *ctrlurl_ipc=NULL;
  float val[10];  
  FILE *fp;
  char buf[300], *key, *value, *value2;
  int i=0,c;
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  if (UpnpDownloadXmlDoc(DescDocURL, &DescDoc) != UPNP_E_SUCCESS) {
    fprintf(stderr,"SohoDeviceStateTableInit -- Error Parsing %s\n", DescDocURL);
    fflush(stderr);
    ret = UPNP_E_INVALID_DESC;
  }
  
  if (!CommonUtil_FindAndParseService(DescDoc, DescDocURL, SohoServiceType[SOHO_SERVICE_L3F], 
				      &servid_l3f, &evnturl_l3f, &ctrlurl_l3f,1)) {    
    fprintf(stderr,"SohoDeviceStateTableInit -- Error: Could not find Service: %s\n", SohoServiceType[0]);
    fflush(stderr);
    if (servid_l3f) free(servid_l3f);
    if (evnturl_l3f) free(evnturl_l3f);
    if (ctrlurl_l3f) free(ctrlurl_l3f);
    return(UPNP_E_INVALID_DESC);
  }

  if (!CommonUtil_FindAndParseService(DescDoc, DescDocURL, SohoServiceType[1], 
				      &servid_cic, &evnturl_cic, &ctrlurl_cic,2)) {
    fprintf(stderr,"SohoDeviceStateTableInit -- Error: Could not find Service: %s\n", SohoServiceType[SOHO_SERVICE_CIC]);
    fflush(stderr);
    if (servid_l3f) free(servid_l3f);
    if (evnturl_l3f) free(evnturl_l3f);
    if (ctrlurl_l3f) free(ctrlurl_l3f);
    if (servid_cic) free(servid_cic);
    if (evnturl_cic) free(evnturl_cic);
    if (ctrlurl_cic) free(ctrlurl_cic);
    return(UPNP_E_INVALID_DESC);
  }  

  if (!CommonUtil_FindAndParseService(DescDoc, DescDocURL, SohoServiceType[2], 
				      &servid_ipc, &evnturl_ipc, &ctrlurl_ipc,3)) {
    fprintf(stderr,"SohoDeviceStateTableInit -- Error: Could not find Service: %s\n", SohoServiceType[SOHO_SERVICE_IPC]);
    fflush(stderr);
    if (servid_l3f) free(servid_l3f);
    if (evnturl_l3f) free(evnturl_l3f);
    if (ctrlurl_l3f) free(ctrlurl_l3f);
    if (servid_cic) free(servid_cic);
    if (evnturl_cic) free(evnturl_cic);
    if (ctrlurl_cic) free(ctrlurl_cic);
    if (servid_ipc) free(servid_ipc);
    if (evnturl_ipc) free(evnturl_ipc);
    if (ctrlurl_ipc) free(ctrlurl_ipc);
    return(UPNP_E_INVALID_DESC);
  }
  
  value2 = NULL;
  if ((fp = fopen("/proc/net/dev", "r")) != NULL) {
    while(!feof(fp)) {
      c = (int) fgetc(fp);     
      if((i == 0) && (!isspace(c))) {
	buf[i++] = c;
      } else if((i>0)&&(c == '\n')) {
	buf[i]='\0';
	key = strdup(buf);
	value=key;
	while(key&&isspace(*key)) key++;
	if(key && (strncmp(key,"eth0:",5)==0)) {
	  key += 5;
	  value2 = strdup(key);
	  free(value);
	  break;
	  }
	free(value);
	i = 0;
      } else {
	buf[i++] = c;
      }
    }
    fclose(fp);
    
    if(value2) { 
      sscanf(value2,"%f %f %f %f %f %f %f %f %f %f",&val[0],&val[1],&val[2],&val[3],&val[4],&val[5],&val[6],&val[7],&val[8],&val[9]);
      free(value2);
      stat_initial_val[0] = val[8];
      stat_initial_val[1] = val[0];
      stat_initial_val[2] = val[9];
      stat_initial_val[3] = val[1];
    }
  }
  
  time(&stat_initial_time);
  
  strcpy(soho_service_table[SOHO_SERVICE_L3F].UDN, SohoDeviceUDN[SOHO_DEVICE_IGD]); 
  strcpy(soho_service_table[SOHO_SERVICE_L3F].ServiceId, servid_l3f); 
  strcpy(soho_service_table[SOHO_SERVICE_L3F].ServiceType, SohoServiceType[SOHO_SERVICE_L3F]); 
  soho_service_table[SOHO_SERVICE_L3F].VariableCount=SOHO_L3F_VARCOUNT; 
  soho_service_table[SOHO_SERVICE_L3F].VariableName[0] = soho_l3f_varname[0];
  soho_service_table[SOHO_SERVICE_L3F].VariableStrVal[0] = soho_l3f_varval[0];
  strcpy(soho_service_table[SOHO_SERVICE_L3F].VariableStrVal[0],soho_l3f_def_val[0]);
#ifdef CONSOLE_CMD
  printf("\tSOHO_SERVICE_L3F initialized ok.\n"); 
#endif
  strcpy(soho_service_table[SOHO_SERVICE_CIC].UDN, SohoDeviceUDN[SOHO_DEVICE_WAN]); 
  strcpy(soho_service_table[SOHO_SERVICE_CIC].ServiceId, servid_cic);
  strcpy(soho_service_table[SOHO_SERVICE_CIC].ServiceType, SohoServiceType[SOHO_SERVICE_CIC]);
  soho_service_table[SOHO_SERVICE_CIC].VariableCount=SOHO_CIC_VARCOUNT;
  for (i=0; i<soho_service_table[SOHO_SERVICE_CIC].VariableCount; i++) {
    soho_service_table[SOHO_SERVICE_CIC].VariableName[i] = soho_cic_varname[i];
    soho_service_table[SOHO_SERVICE_CIC].VariableStrVal[i] = soho_cic_varval[i];
    strcpy(soho_service_table[SOHO_SERVICE_CIC].VariableStrVal[i], soho_cic_def_val[i]);
  }
#ifdef CONSOLE_CMD
  printf("\tSOHO_SERVICE_CIC initialized ok.\n");
#endif  
  strcpy(soho_service_table[SOHO_SERVICE_IPC].UDN, SohoDeviceUDN[SOHO_DEVICE_WCD]);
  strcpy(soho_service_table[SOHO_SERVICE_IPC].ServiceId, servid_ipc);
  strcpy(soho_service_table[SOHO_SERVICE_IPC].ServiceType, SohoServiceType[SOHO_SERVICE_IPC]);
  soho_service_table[SOHO_SERVICE_IPC].VariableCount=SOHO_IPC_VARCOUNT;
  for (i=0; i<soho_service_table[SOHO_SERVICE_IPC].VariableCount; i++) {
    soho_service_table[SOHO_SERVICE_IPC].VariableName[i] = soho_ipc_varname[i];
    soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[i] = soho_ipc_varval[i];
    strcpy(soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[i], soho_ipc_def_val[i]);
  }
  /* export the external ip  */
  strcpy(soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[7], lanip);
  soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[7][strlen(lanip)]='\0';
#ifdef CONSOLE_CMD
  printf("\tSOHO_SERVICE_IPC initialized ok.\n"); 
#endif
  if (servid_l3f) free(servid_l3f);
  if (evnturl_l3f) free(evnturl_l3f);
  if (ctrlurl_l3f) free(ctrlurl_l3f);  
  if (servid_cic) free(servid_cic);
  if (evnturl_cic) free(evnturl_cic);
  if (ctrlurl_cic) free(ctrlurl_cic);
  if (servid_ipc) free(servid_ipc);
  if (evnturl_ipc) free(evnturl_ipc);
  if (ctrlurl_ipc) free(ctrlurl_ipc);

  ret = SohoDeviceStateTableValueInit();

  return(ret);
}             

int SohoDeviceStateTableValueInit()
{
  /* Any value need to be gathered when the device is first up goes this section.              */
  /* Some of these values are already set as default value and may not need to be setup again. */ 

  /* ? physical connection status: default is up         */
  /* ? connection speeds:                                */
  /* ? personal firewall enabled: default is 0           */
  /* ? NATEnabled & RSIPAvailable:  by default are 1 & 1 */ 


  /* Since the Windows XP control point really doesn't care lots of information (variables)
     in the IGD spec, so I just ignore them. Later whoever take care of this part can add them 
     in if really needed. The items in above list are needed as default. And portmapping is below.
  */
     
  /* ! upnp_upadte_enable                                     */  
  /* ! port mapping:                                         */  

  int fd;
  void* ptr;
  char *rec = NULL, *rec1, *rec2, *rec3;
  struct stat sb;
  const char delimiters[] = "\r\n\t"; 
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  upnp_update_enable = SohoDeviceUpdateEnable();
#ifdef DEBUG
fprintf(stderr,"upnp_update_enable=%d\n",upnp_update_enable);
fflush(stderr);
#endif

  if ( stat(UPNP_SOHO_CONF_FILE, &sb ) < 0 ) {
      soho_portmapping_number = 0;
      return (UPNP_E_SUCCESS);
  }
#ifdef DEBUG
fprintf(stderr,"open %s\n",UPNP_SOHO_CONF_FILE); 
fflush(stderr);
#endif
  	if (S_ISDIR( sb.st_mode ) )	
  	{
  		sp_upnp_sohoTrace("%s: %s..:if (S_ISDIR( sb.st_mode ) )\n", __FILE__, __func__);	  
    	return UPNP_E_INIT_FAILED;
	}

  pm_time = sb.st_mtime;
  
#ifdef DEBUG
fprintf(stderr,"open %s\n",UPNP_SOHO_CONF_FILE); 
fflush(stderr);
#endif
  fd = open(UPNP_SOHO_CONF_FILE, O_RDONLY);
  	if ( fd < 0 )
    {
  		sp_upnp_sohoTrace("%s: %s..:if ( fd < 0 )\n", __FILE__, __func__);	  
    	return UPNP_E_INIT_FAILED;
	}
  
  if ( sb.st_size > 0 ) {   /* avoid zero-length mmap */
    ptr = mmap( 0, sb.st_size, PROT_READ, MAP_PRIVATE, fd, 0 );
    if ( ptr != (void*) -1 ) {
      rec = strdup ((char*)ptr);   
      (void) munmap( ptr, sb.st_size );
    }
  }
  (void) close( fd );
  	if(rec == NULL) 
    {
  		sp_upnp_sohoTrace("%s: %s..:if(rec == NULL) \n", __FILE__, __func__);	  
    	return UPNP_E_INIT_FAILED;
	}

  soho_portmapping_number = 0;
  rec2 = rec;
  rec1 = strtok (rec, delimiters); 
 
  if(rec1 && strlen(rec1)>10 && rec1[0] != '#')  {  
    rec3 = strdup(rec1);  
#ifdef DEBUG
printf("%s\n",rec3);  
#endif
    SohoDevicePortMappingInit(rec3,&soho_portmapping_number);
    free(rec3);  
  } 
  
  while(rec1 = strtok (NULL, delimiters)) {
    if(strlen(rec1)>10 && rec1[0] != '#') {
      rec3 = strdup(rec1);
#ifdef DEBUG
printf("%s\n",rec3);
#endif
      SohoDevicePortMappingInit(rec3,&soho_portmapping_number);
      free(rec3);
    }
  }
  
  free(rec2); 
#ifdef CONSOLE_CMD
  printf("\tPortMapping Records: %d\n",soho_portmapping_number);
#endif  
  sprintf(soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[8], "%d",soho_portmapping_number);

  return (UPNP_E_SUCCESS);
}

int SohoDevicePortMappingInit(const char *rec, int *num)
{
  char *p1,*p2,*p3,*subnet;
  unsigned int n,n1;
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  if( *num >= MAX_PORT_MAPPING_NUM )
    return (1);

  // TCP:23:0:0:23:Telnet
  p1 = (char*) strndup(rec,(size_t) 3);
  if(strcmp(p1,"TCP")&&strcmp(p1,"UDP")) {
    free(p1);
    return (1);
  }  
  sprintf(soho_portmapping_table[*num].protocol,"%s",p1);
  free(p1);

  p1 = rec + 4;
  p2 = index(p1,':');
  n = strlen(p1) - strlen(p2);
  p3 = strndup(p1,(size_t)n);
  if(atoi(p3)<=0 ||atoi(p3)>65535) {
    free(p3);
    return (1);
  }  
  n += 5;
  soho_portmapping_table[*num].external_port = atoi(p3);
  free(p3);

  p1 = rec + n;
  p2 = index(p1,':');
  n1 = strlen(p1) - strlen(p2);
  p3 = strndup(p1,(size_t) n1);
  n += n1 + 1;
  soho_portmapping_table[*num].enabled = atoi(p3)?1:0;
  free(p3);

  p2 = rindex(lanip,'.');
  subnet = strndup(lanip,(size_t)(strlen(lanip)-strlen(p2)));

  p1 = rec + n;
  p2 = index(p1,':');
  n1 = strlen(p1) - strlen(p2);
  p3 = strndup(p1,(size_t) n1);
  if(atoi(p3)<0 ||atoi(p3)>254) {
    free(p3);
    return (1);
  }  
  n += n1 + 1;
  sprintf(soho_portmapping_table[*num].internal_client,"%s.%s",subnet,p3);
  free(p3);
  free(subnet);

  p1 = rec + n;
  p2 = index(p1,':');
  n1 = strlen(p1) - strlen(p2);
  p3 = strndup(p1,(size_t) n1);
  if(atoi(p3)<1 ||atoi(p3)>65535) {
    free(p3);	
    return (1);
  }  
  soho_portmapping_table[*num].internal_port = atoi(p3);
  free(p3);

  p1 = rindex(rec,':');
  p1++;
  sprintf(soho_portmapping_table[*num].description,"%s",p1);
  
  soho_portmapping_table[*num].lease_duration = 0;
  soho_portmapping_table[*num].index = *num;
  *num += 1;

  return (0);
}

int SohoDeviceCallbackEventHandler(Upnp_EventType EventType, void *Event, void *Cookie)
{
#ifdef DEBUG
  /* Print a summary of the event received */
  //CommonUtil_PrintEvent(EventType, Event);
#endif
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  switch ( EventType) {    
  case UPNP_EVENT_SUBSCRIPTION_REQUEST:
    SohoDeviceHandleSubscriptionRequest((struct Upnp_Subscription_Request *) Event);
    break;    
  case UPNP_CONTROL_GET_VAR_REQUEST:
    SohoDeviceHandleGetVarRequest((struct Upnp_State_Var_Request *) Event);
    break;
  case UPNP_CONTROL_ACTION_REQUEST:
    SohoDeviceHandleActionRequest((struct Upnp_Action_Request *) Event);
    break;
    /* ignore these cases, since this is not a control point */
  case UPNP_DISCOVERY_ADVERTISEMENT_ALIVE:
  case UPNP_DISCOVERY_SEARCH_RESULT:
  case UPNP_DISCOVERY_SEARCH_TIMEOUT:
  case UPNP_DISCOVERY_ADVERTISEMENT_BYEBYE:
  case UPNP_CONTROL_ACTION_COMPLETE:
  case UPNP_CONTROL_GET_VAR_COMPLETE:
  case UPNP_EVENT_RECEIVED:
  case UPNP_EVENT_RENEWAL_COMPLETE:
  case UPNP_EVENT_SUBSCRIBE_COMPLETE:
  case UPNP_EVENT_UNSUBSCRIBE_COMPLETE:
    break;
  default:

#ifdef DEBUG
    fprintf(stderr,"Error in SohoDeviceCallbackEventHandler: unknown event type %d\n", EventType);
    fflush(stderr);
#endif

	break;
  }
  return(0);
}


int SohoDeviceHandleGetVarRequest(struct Upnp_State_Var_Request *cgv_event) 
{
  int i, j;
  int getvar_succeeded = 0;
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  cgv_event->CurrentVal = NULL;
  
  pthread_mutex_lock(&SohoDevMutex);

  for (i=0; i<SOHO_SERVICE_SERVCOUNT; i++) {
    if ((strcmp(cgv_event->DevUDN,soho_service_table[i].UDN)==0) &&
	(strcmp(cgv_event->ServiceID,soho_service_table[i].ServiceId)==0)) {
      for (j=0; j< soho_service_table[i].VariableCount; j++) {
	if (strcmp(cgv_event->StateVarName, soho_service_table[i].VariableName[j])==0) {
	  getvar_succeeded = 1;
	  cgv_event->CurrentVal = (Upnp_DOMString) malloc(sizeof(soho_service_table[i].VariableStrVal[j]));
	  strcpy(cgv_event->CurrentVal, soho_service_table[i].VariableStrVal[j]);
	  break;
	}
      } 
    }
  }
       
  if (getvar_succeeded) {
    cgv_event->ErrCode = UPNP_E_SUCCESS;
  } else {

#ifdef DEBUG
    fprintf(stderr,"Error in UPNP_CONTROL_GET_VAR_REQUEST callback:\n"); 
    fprintf(stderr,"   Unknown variable name = %s\n",  cgv_event->StateVarName); 
    fflush(stderr);
#endif
    cgv_event->ErrCode = 404;
    strcpy(cgv_event->ErrStr, "Invalid Variable");
  }

  pthread_mutex_unlock(&SohoDevMutex);
  
  return(cgv_event->ErrCode == UPNP_E_SUCCESS);
}

int SohoDeviceHandleActionRequest(struct Upnp_Action_Request *ca_event) 
{
  int action_succeeded = -1;
  	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  ca_event->ErrCode = 0;
  ca_event->ActionResult = NULL;
  
  if ((strcmp(ca_event->DevUDN, SohoDeviceUDN[SOHO_DEVICE_WCD])==0) 
      && (strcmp(ca_event->ServiceID, SohoServiceID[SOHO_SERVICE_IPC])==0)) {
    
    if (strcmp(ca_event->ActionName, "GetConnectionTypeInfo") == 0) {
      action_succeeded = GetConnectionTypeInfo(ca_event);
    } else if (strcmp(ca_event->ActionName, "GetNATRSIPStatus") == 0) {
      action_succeeded = GetNATRSIPStatus(ca_event);
    } else if (strcmp(ca_event->ActionName, "GetStatusInfo") == 0) {
      action_succeeded = GetStatusInfo(ca_event);
    } else if (strcmp(ca_event->ActionName, "GetExternalIPAddress") == 0) {
      action_succeeded = GetExternalIPAddress(ca_event);
    } else if (strcmp(ca_event->ActionName, "GetGenericPortMappingEntry") == 0) {
      action_succeeded = GetGenericPortMappingEntry(ca_event);
    } else if (strcmp(ca_event->ActionName, "GetSpecificPortMappingEntry") == 0) {
      action_succeeded = GetSpecificPortMappingEntry(ca_event);
    } else if ((strcmp(ca_event->ActionName, "ForceTermination") == 0)
	       && upnp_update_enable ==1) {
      action_succeeded = ForceTermination(ca_event);
    } else if ((strcmp(ca_event->ActionName, "RequestConnection") == 0) 
	       && upnp_update_enable ==1) {
      action_succeeded = RequestConnection(ca_event); 
    } else if ((strcmp(ca_event->ActionName, "DeletePortMapping") == 0) 
	       && upnp_update_enable ==1) {
      action_succeeded = DeletePortMapping(ca_event);
    } else if ((strcmp(ca_event->ActionName, "AddPortMapping") == 0) 
	       && upnp_update_enable ==1) {
      action_succeeded = AddPortMapping(ca_event);
    } 

  } else if ((strcmp(ca_event->DevUDN, SohoDeviceUDN[SOHO_DEVICE_WAN])==0) 
	     && (strcmp(ca_event->ServiceID, SohoServiceID[SOHO_SERVICE_CIC])==0)) {
    
    if (strcmp(ca_event->ActionName, "GetTotalBytesSent") == 0) {
      action_succeeded = GetPacketByteStat(TOTAL_BYTES_SENT, ca_event);
    } else if (strcmp(ca_event->ActionName, "GetTotalBytesReceived") == 0) {
      action_succeeded = GetPacketByteStat(TOTAL_BYTES_RECEIVED, ca_event);
    } else if (strcmp(ca_event->ActionName, "GetTotalPacketsSent") == 0) {
      action_succeeded = GetPacketByteStat(TOTAL_PACKETS_SENT, ca_event);
    } else if (strcmp(ca_event->ActionName, "GetTotalPacketsReceived") == 0) {
      action_succeeded = GetPacketByteStat(TOTAL_PACKETS_RECEIVED, ca_event);
    } else if (strcmp(ca_event->ActionName, "GetCommonLinkProperties") == 0) {
      action_succeeded = GetCommonLinkProperties(ca_event);
    }      
  }

  if(action_succeeded <=0) {
    if (action_succeeded == -1) {      
#ifdef DEBUG
      fprintf(stderr,"Error in UPNP_CONTROL_ACTION_REQUEST callback:\n"); 
      fprintf(stderr,"   Unknown ActionName = %s\n",  ca_event->ActionName); 
      fprintf(stderr,"   or unknown UDN = %s or ServiceId = %s\n",  ca_event->DevUDN, ca_event->ServiceID ); 
      fflush(stderr);
#endif
      ca_event->ErrCode = 401;
      strcpy(ca_event->ErrStr, "Invalid Action");
      ca_event->ActionResult = NULL;      
    } else {      
#ifdef DEBUG
      fprintf(stderr,"Error in UPNP_CONTROL_ACTION_REQUEST callback:\n"); 
      fprintf(stderr,"   Failure while running %s\n",  ca_event->ActionName); 
      fflush(stderr);
#endif
      ca_event->ErrCode = action_succeeded ;
      strcpy(ca_event->ErrStr, "Invalid Args");
      ca_event->ActionResult = NULL;
    }
  }  
  return(ca_event->ErrCode);
}

int SohoDeviceHandleSubscriptionRequest(struct Upnp_Subscription_Request *sr_event) 
{
  int i,j; 
  Upnp_Document PropSet; 
  	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  pthread_mutex_lock(&SohoDevMutex);
  
  for (i=0; i<SOHO_SERVICE_SERVCOUNT; i++) {
    if ((strcmp(sr_event->UDN,soho_service_table[i].UDN) == 0) && 
	(strcmp(sr_event->ServiceId,soho_service_table[i].ServiceId) == 0)) {    
      
      for (j=0; j< soho_service_table[i].VariableCount; j++) 
	UpnpAddToPropertySet(&PropSet, soho_service_table[i].VariableName[j],soho_service_table[i].VariableStrVal[j]); 
      
      UpnpAcceptSubscriptionExt(device_handle, sr_event->UDN, sr_event->ServiceId,PropSet, sr_event->Sid); 
      UpnpDocument_free(PropSet);       
    }
  }

  pthread_mutex_unlock(&SohoDevMutex);	
  return(1);
}

int GetConnectionTypeInfo(struct Upnp_Action_Request * ca_event)
{
  char result_str[500];
  char resultcode[]="<NewConnectionType>IP_Routed</NewConnectionType><NewPossibleConnectionTypes>IP_Routed</NewPossibleConnectionTypes>";
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  pthread_mutex_lock(&SohoDevMutex);

  ca_event->ErrCode = UPNP_E_SUCCESS;
  
  sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">%s</u:%sResponse>", 
	  ca_event->ActionName,
	  SohoServiceType[SOHO_SERVICE_IPC],  
	  resultcode,
	  ca_event->ActionName);

  //printf("%s\n",result_str);

  ca_event->ActionResult = UpnpParse_Buffer(result_str);

  pthread_mutex_unlock(&SohoDevMutex); 

  return(1);
}

int GetNATRSIPStatus (struct Upnp_Action_Request * ca_event)
{
  char result_str[500];
  char resultcode[200];
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  pthread_mutex_lock(&SohoDevMutex);

  ca_event->ErrCode = UPNP_E_SUCCESS;

  sprintf(resultcode, "<NewRSIPAvailable>%s</NewRSIPAvailable><NewNATEnabled>%s</NewNATEnabled>",
	  soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[5],
	  soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[6]);

  sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">%s</u:%sResponse>", 
	  ca_event->ActionName,
	  SohoServiceType[SOHO_SERVICE_IPC],  
	  resultcode,
	  ca_event->ActionName);

  //printf("%s\n",result_str);

  ca_event->ActionResult = UpnpParse_Buffer(result_str);

  pthread_mutex_unlock(&SohoDevMutex); 

  return(1);
}

int GetCommonLinkProperties(struct Upnp_Action_Request * ca_event)
{
  char result_str[500];
  char resultcode[300];
  struct stat sb;
  	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  pthread_mutex_lock(&SohoDevMutex);

  ca_event->ErrCode = UPNP_E_SUCCESS;

  sprintf(resultcode,"<NewWANAccessType>%s</NewWANAccessType><NewLayer1UpstreamMaxBitRate>%s</NewLayer1UpstreamMaxBitRate><NewLayer1DownstreamMaxBitRate>%s</NewLayer1DownstreamMaxBitRate><NewPhysicalLinkStatus>%s</NewPhysicalLinkStatus>",
	  soho_service_table[SOHO_SERVICE_CIC].VariableStrVal[0],
	  soho_service_table[SOHO_SERVICE_CIC].VariableStrVal[1],
	  soho_service_table[SOHO_SERVICE_CIC].VariableStrVal[2],
	  soho_service_table[SOHO_SERVICE_CIC].VariableStrVal[3]);

  sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">%s</u:%sResponse>", 
	  ca_event->ActionName,
	  SohoServiceType[SOHO_SERVICE_IPC],  
	  resultcode,
	  ca_event->ActionName);

  //printf("%s\n",result_str);

  ca_event->ActionResult = UpnpParse_Buffer(result_str);
  
  pthread_mutex_unlock(&SohoDevMutex); 

  return(1);
}


int GetStatusInfo(struct Upnp_Action_Request * ca_event)
{
  char result_str[500];
  char resultcode[200];
  time_t t = 0;
#ifdef DEBUG
printf("GetStatusInfo:ENTER\n");
#endif
  pthread_mutex_lock(&SohoDevMutex);
#ifdef DEBUG
printf("GetStatusInfo:GET LOCK\n");
#endif
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  if(strcmp(soho_service_table[SOHO_SERVICE_CIC].VariableStrVal[3],"Down")
     && strcmp(soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[2],"Disconnected")) {

    time(&t);
    t = (t >= stat_initial_time)?(t - stat_initial_time):0;
  }

  ca_event->ErrCode = UPNP_E_SUCCESS;
  
  sprintf(resultcode,"<NewConnectionStatus>%s</NewConnectionStatus><NewLastConnectionError>ERROR_NONE</NewLastConnectionError><NewUptime>%d</NewUptime>", soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[2],(int)t);

  sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">%s</u:%sResponse>", 
	  ca_event->ActionName,
	  SohoServiceType[SOHO_SERVICE_IPC],  
	  resultcode,
	  ca_event->ActionName);

  ca_event->ActionResult = UpnpParse_Buffer(result_str);

  pthread_mutex_unlock(&SohoDevMutex); 

  return(1);
}


int GetPacketByteStat(int statfield, struct Upnp_Action_Request *ca_event)
{
  char result_str[500];
  char *data;
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  pthread_mutex_lock(&SohoDevMutex);

  data = SohoParsePacketByte(statfield);

  ca_event->ErrCode = UPNP_E_SUCCESS;
 
  sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">%s</u:%sResponse>", 
	  ca_event->ActionName,
	  SohoServiceType[SOHO_SERVICE_IPC],  
	  data,
	  ca_event->ActionName);

  //printf("%s\n",result_str);

  ca_event->ActionResult = UpnpParse_Buffer(result_str);

  if(data) free(data);

  pthread_mutex_unlock(&SohoDevMutex); 

  return(1);
}


int GetGenericPortMappingEntry(struct Upnp_Action_Request *ca_event)
{
  char result_str[800];
  char resultcode[500];
  char *value;
  struct stat sb;
  int i;
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  pthread_mutex_lock(&SohoDevMutex);

  ca_event->ErrCode = UPNP_E_SUCCESS;

  if ((value = CommonUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewPortMappingIndex")) == NULL) {
    pthread_mutex_unlock(&SohoDevMutex); 
    return( UPNP_E_INVALID_ARGUMENT );
  }
  
  i = atoi(value);
#ifdef DEBUG
printf("soho_portmapping_number=%d\n",soho_portmapping_number);  
#endif
  if ((i<0) || (i >= soho_portmapping_number)) {
    pthread_mutex_unlock(&SohoDevMutex); 
    return( UPNP_E_INVALID_ARGUMENT );
  }
  
  sprintf(resultcode,"<NewRemoteHost></NewRemoteHost><NewExternalPort>%d</NewExternalPort><NewProtocol>%s</NewProtocol><NewInternalPort>%d</NewInternalPort><NewInternalClient>%s</NewInternalClient><NewEnabled>%d</NewEnabled><NewPortMappingDescription>%s</NewPortMappingDescription><NewLeaseDuration>%.0f</NewLeaseDuration>",
	  soho_portmapping_table[i].external_port,
	  soho_portmapping_table[i].protocol,	  
	  soho_portmapping_table[i].internal_port,
	  soho_portmapping_table[i].internal_client,
	  soho_portmapping_table[i].enabled,
	  soho_portmapping_table[i].description,
	  (float)soho_portmapping_table[i].lease_duration );
  
  sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">%s</u:%sResponse>", 
	  ca_event->ActionName,
	  SohoServiceType[SOHO_SERVICE_IPC],  
	  resultcode,
	  ca_event->ActionName);

  //  printf("%s\n",result_str);

  ca_event->ActionResult = UpnpParse_Buffer(result_str);

  pthread_mutex_unlock(&SohoDevMutex); 

  return(1);
}

int GetSpecificPortMappingEntry(struct Upnp_Action_Request *ca_event)
{
  char result_str[500];
  char resultcode[320];
  char *value1;
  char *value2;
  int i,j = -1;
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  pthread_mutex_lock(&SohoDevMutex);

  ca_event->ErrCode = UPNP_E_SUCCESS;

  if ((value1 = CommonUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewExternalPort")) == NULL) {
    pthread_mutex_unlock(&SohoDevMutex); 
    return(UPNP_E_INVALID_ARGUMENT );
  }

  if ((value2 = CommonUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewProtocol")) == NULL) {
    pthread_mutex_unlock(&SohoDevMutex); 
    return( UPNP_E_INVALID_ARGUMENT);
  }
  
  for(i=0; i<soho_portmapping_number; i++) {
    if ((soho_portmapping_table[i].external_port == atoi(value1)) 
	&& (strcmp(soho_portmapping_table[i].protocol,value2)==0))
      j = i;
  }

  if( j < 0 ) {
    pthread_mutex_unlock(&SohoDevMutex); 
    return( UPNP_E_INVALID_ARGUMENT );
  }
  
  sprintf(resultcode,"<NewInternalPort>%d</NewInternalPort><NewInternalClient>%s</NewInternalClient><NewEnabled>%d</NewEnabled><NewPortMappingDescription>%s</NewPortMappingDescription><NewLeaseDuration>%.0f</NewLeaseDuration>",
	  soho_portmapping_table[j].internal_port,
	  soho_portmapping_table[j].internal_client,
	  soho_portmapping_table[j].enabled,
	  soho_portmapping_table[j].description,
	  (float)soho_portmapping_table[j].lease_duration );
  
  sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">%s</u:%sResponse>", 
	  ca_event->ActionName,
	  SohoServiceType[SOHO_SERVICE_IPC],  
	  resultcode,
	  ca_event->ActionName);

  // printf("%s\n",result_str);

  ca_event->ActionResult = UpnpParse_Buffer(result_str);

  pthread_mutex_unlock(&SohoDevMutex); 

  return(1);
}

int AddPortMapping(struct Upnp_Action_Request *ca_event)
{
  char result_str[200];
  char *v1,*v2,*v3,*v4,*v5,*v6,*v7,*subnet;
  int i,j = -1,sync=0,k; 
  int pid,ret;
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  pthread_mutex_lock(&SohoDevMutex);

  if ((v1 = CommonUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewExternalPort")) == NULL) {
    pthread_mutex_unlock(&SohoDevMutex); 
    return(UPNP_E_INVALID_ARGUMENT );
  }

  if ((v2 = CommonUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewProtocol")) == NULL) {
    pthread_mutex_unlock(&SohoDevMutex); 
    return(UPNP_E_INVALID_ARGUMENT );
  }

  if ((v3 = CommonUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewInternalPort")) == NULL) {
    pthread_mutex_unlock(&SohoDevMutex); 
    return(UPNP_E_INVALID_ARGUMENT );
  }

  if ((v4 = CommonUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewInternalClient")) == NULL) {
    pthread_mutex_unlock(&SohoDevMutex); 
    return(UPNP_E_INVALID_ARGUMENT );
  }

  if ((v5 = CommonUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewEnabled")) == NULL) {
    pthread_mutex_unlock(&SohoDevMutex); 
    return(UPNP_E_INVALID_ARGUMENT );
  }

  if ((v6 = CommonUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewPortMappingDescription")) == NULL) {
    pthread_mutex_unlock(&SohoDevMutex); 
    return(UPNP_E_INVALID_ARGUMENT );
  }

  if ((v7 = CommonUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewLeaseDuration")) == NULL) {
    pthread_mutex_unlock(&SohoDevMutex); 
    return(UPNP_E_INVALID_ARGUMENT );
  }

  for(i=0; i < soho_portmapping_number; i++) { /* if an existing record */ 
    if ((soho_portmapping_table[i].external_port == atoi(v1)) 
	&& (strcasecmp(soho_portmapping_table[i].protocol,v2)==0))
      j = i;
  }

  k=(j>=0)?j:soho_portmapping_number;

  subnet = strndup(lanip,(size_t)(strlen(lanip)-strlen(rindex(lanip,'.'))+1));   

  if(j<0) {
    if(soho_portmapping_number >= MAX_PORT_MAPPING_NUM ) {
      pthread_mutex_unlock(&SohoDevMutex); 
      return(UPNP_E_INVALID_ARGUMENT );
    }
    soho_portmapping_table[k].enabled = (int)strtol(v5,NULL,10);
    soho_portmapping_table[k].lease_duration = strtol(v7,NULL,10);
    soho_portmapping_table[k].external_port = (int)strtol(v1,NULL,10);
    soho_portmapping_table[k].internal_port = (int)strtol(v3,NULL,10);
    sprintf(soho_portmapping_table[k].protocol,"%s",v2);
    sprintf(soho_portmapping_table[k].description,"%s",v6);  
    soho_portmapping_table[k].index = k;
    sync = soho_portmapping_table[k].enabled;

    if(strncmp(v4,subnet,strlen(subnet))) {
      soho_portmapping_table[k].enabled = sync = 0;
      sprintf(soho_portmapping_table[k].internal_client,"%s0",subnet);
    } else {
      sprintf(soho_portmapping_table[k].internal_client,"%s",v4);
    }
    soho_portmapping_number++;
    sprintf(soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[8], "%d", soho_portmapping_number);

  } else {

    if(soho_portmapping_table[k].enabled != (int)strtol(v5,NULL,10)) {
      soho_portmapping_table[k].enabled = (int)strtol(v5,NULL,10);
      sync = 1;
    }
    if(soho_portmapping_table[k].external_port != (int)strtol(v1,NULL,10)) {
      soho_portmapping_table[k].external_port = (int)strtol(v1,NULL,10);
      sync =1;
    }
    if(soho_portmapping_table[k].internal_port != (int)strtol(v3,NULL,10)) {
      soho_portmapping_table[k].internal_port = (int)strtol(v3,NULL,10);
      sync =1;
    }
    if(strncmp(v4,subnet,strlen(subnet))) {
      soho_portmapping_table[k].enabled = sync = 0;
      sprintf(soho_portmapping_table[k].internal_client,"%s0",subnet);
    } else { 
      if(strcmp(soho_portmapping_table[k].internal_client,v4))
	sprintf(soho_portmapping_table[k].internal_client,"%s",v4);
    }
    sprintf(soho_portmapping_table[k].description,"%s",v6);  
    soho_portmapping_table[k].index = k;
  }

  free(subnet);

  /* If the record to be deleted is an enabled record, then we need to sync with /conf/config file */
  /* and restart the firewall to make the change affective.                                        */
  /* Otherwise we can just dump all records to /etc/sysconfig/portmapping.conf and we don't touch  */
  /* the firewall.                                                                                 */ 
  /* so far the internal client does not support doamin name, and should be easy solved            */

  PortMappingDump();
  if(sync) PortMappingSync();  

  ca_event->ErrCode = UPNP_E_SUCCESS;

  sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\"></u:%sResponse>", 
	  ca_event->ActionName,
	  SohoServiceType[SOHO_SERVICE_IPC],  
	  ca_event->ActionName);

 // printf("%s\n",result_str);

  ca_event->ActionResult = UpnpParse_Buffer(result_str);
 
  UpnpNotify(device_handle,
	     soho_service_table[SOHO_SERVICE_IPC].UDN,
	     soho_service_table[SOHO_SERVICE_IPC].ServiceId,
	     soho_service_table[SOHO_SERVICE_IPC].VariableName,
	     soho_service_table[SOHO_SERVICE_IPC].VariableStrVal,
	     SOHO_IPC_VARCOUNT);
  
  if(sync) {
#ifdef USE_FORK
    pid = fork();
    if (pid == 0) {
      // This may be broken on some OS. Please keep in mind.
      // -wycc
printf("Start to sync the system\n");
      //setsid();
      for(i=3;i<256;i++)
	close(i);
      //ret = execl("/etc/init.d/firewall","firewall","restart",NULL); 
      _exit(ret);
printf("End to sync the system\n");
    } else if (pid>0) {
printf("WEait for the child to be finished\n");
#ifdef WIATPID_WAS_OK
      waitpid(-1,&ret,0);
#endif
printf("child is finished\n");
    }
#else
    system("/etc/init.d/firewall restart_portmap");
#endif
  }
  pthread_mutex_unlock(&SohoDevMutex); 
  return(1);
}

int DeletePortMapping(struct Upnp_Action_Request *ca_event)
{
  char result_str[200];
  char *value1, *value2, *subnet;
  int i,j = -1,sync=0,pid,ret;
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  pthread_mutex_lock(&SohoDevMutex);

  if ((value1 = CommonUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewExternalPort")) == NULL) {
    pthread_mutex_unlock(&SohoDevMutex); 
    return(UPNP_E_INVALID_ARGUMENT );
  }
  
  if ((value2 = CommonUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewProtocol")) == NULL) {
    pthread_mutex_unlock(&SohoDevMutex); 
    return( UPNP_E_INVALID_ARGUMENT);
  }
  
  for(i=0; i<soho_portmapping_number; i++) {
    if ((soho_portmapping_table[i].external_port == atoi(value1)) 
	&& (strcasecmp(soho_portmapping_table[i].protocol,value2)==0))
      j = i;
  }

  /* If the record to be deleted is an enabled record, then we need to sync with /conf/config file */
  /* and restart the firewall to make the change affective.                                        */
  /* Otherwise we can just dump all records to /etc/sysconfig/portmapping.conf and we don't touch  */
  /* the firewall.                                                                                 */ 
  
  if (j > -1) {  /* delete the portmapping recording */
    subnet = strndup(lanip,(size_t)(strlen(lanip)-strlen(rindex(lanip,'.'))+1));   
    
    sync = soho_portmapping_table[j].enabled;
    if(strncmp(soho_portmapping_table[j].internal_client,subnet,strlen(subnet))) sync =0;    
    free(subnet);

    if (j < soho_portmapping_number - 1) {
      for(i=j; i<=soho_portmapping_number-1;i++) {
	  soho_portmapping_table[i] = soho_portmapping_table[i+1];
	  soho_portmapping_table[i].index = i; 
	}
    }
    soho_portmapping_number --;
    sprintf(soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[8], "%d", soho_portmapping_number);

    PortMappingDump();
    if(sync) PortMappingSync();
  }

  ca_event->ErrCode = UPNP_E_SUCCESS;

  sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\"></u:%sResponse>", 
	  ca_event->ActionName,
	  SohoServiceType[SOHO_SERVICE_IPC],  
	  ca_event->ActionName);

 // printf("%s\n",result_str);

  UpnpNotify(device_handle,
	     soho_service_table[SOHO_SERVICE_IPC].UDN,
	     soho_service_table[SOHO_SERVICE_IPC].ServiceId,
	     soho_service_table[SOHO_SERVICE_IPC].VariableName,
	     soho_service_table[SOHO_SERVICE_IPC].VariableStrVal,
	     SOHO_IPC_VARCOUNT);

  ca_event->ActionResult = UpnpParse_Buffer(result_str);

  if(sync) {
#ifdef USE_FORK
    pid = fork();
    if (pid == 0) {
      if (fork()==0) {
	setsid();
	for(i=3;i<256;i++)
	  close(i);
	ret = execl("/etc/init.d/firewall","firewall","restart_", NULL); 
	_exit(ret);
      } else {
	_exit(ret);
      }
    } else if (pid>0) {
#ifdef WAITPID_WAS_OK
      waitpid(-1,&ret,0);
#endif
    }
#else
      system("/etc/init.d/firewall restart_portmap");
#endif
  }
  
  pthread_mutex_unlock(&SohoDevMutex); 
  return(1);
}

int PortMappingDump()
{
  FILE *fp;
  int i;
  char *part,*subnet;
  struct stat sb;
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  if ((fp = fopen(UPNP_SOHO_CONF_FILE, "w")) == NULL) {    
    return -1;
  }
  //  flock (fp,LOCK_EX);
  subnet = strndup(lanip,(size_t)(strlen(lanip)-strlen(rindex(lanip,'.'))+1));  

  fprintf(fp,"# This file holds the portmapping records which are maintained by both upnp_soho & rct.\n# Format: external port:protocol:status:lan host ip last part:internal port:description\n# Copyright (C) REDSonic Inc. 2001, 2002 All rights reserved.\n####\n");

  for(i=0; i < soho_portmapping_number; i++) {
    if(strncmp(soho_portmapping_table[i].internal_client,subnet,strlen(subnet))) {
      fprintf(fp,"%s:%d:%d:0:%d:%s\n",
	      soho_portmapping_table[i].protocol,
	      soho_portmapping_table[i].external_port,
	      soho_portmapping_table[i].enabled,
	      soho_portmapping_table[i].internal_port,
	      soho_portmapping_table[i].description);
    } else {
      part = rindex( soho_portmapping_table[i].internal_client,'.');
      part++;
      fprintf(fp,"%s:%d:%d:%s:%d:%s\n",
	      soho_portmapping_table[i].protocol,
	      soho_portmapping_table[i].external_port,
	      soho_portmapping_table[i].enabled,
	      part,
	      soho_portmapping_table[i].internal_port,
	      soho_portmapping_table[i].description);
    }
  }

  free(subnet);
  //  flock (fp,LOCK_EX);
  fclose(fp);

 if ( stat(UPNP_SOHO_CONF_FILE, &sb ) < 0 ) 
   return -1;
 pm_time = sb.st_mtime;

  return(0);
}

int PortMappingSync()
{
  FILE *fp, *fd;
  int i,k=0,c;
  char *part,*subnet;
  char buf[MAX_LINE_LENGTH+1],buf1[MAX_LINE_LENGTH+1];
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  subnet = strndup(lanip,(size_t)(strlen(lanip)-strlen(rindex(lanip,'.'))+1));  

  for(i=0; i < soho_portmapping_number; i++) {
    if((!strncmp(soho_portmapping_table[i].internal_client,subnet,strlen(subnet)))
       && (soho_portmapping_table[i].enabled)) {
      part = rindex( soho_portmapping_table[i].internal_client,'.');
      part++;
      sprintf(buf1,":%c%d,%s,%d",
	      soho_portmapping_table[i].protocol[0],
	      soho_portmapping_table[i].external_port,
	      part,
	      soho_portmapping_table[i].internal_port);
      if(k)
	strcat(buf,buf1);
      else{
	k=1;
	strcpy(buf,buf1);
      }
    }
  }

  if(k) part = &buf[1];
  
  i=0;

  if ((fp = fopen(RCT_CONF_FILE, "r")) == NULL) {    
    return -1;
  }  
  if ((fd = fopen(SYNC_TMP_FILE, "w")) == NULL) {    
    return -1;
  }
  
  while (!feof(fp)) {
    c = fgetc(fp);     
    if ( (!isspace(c)) && i == 0) {
      if (i < MAX_LINE_LENGTH)
	buf1[i++] = c;
    } else if ((i > 0) && (c == '\n')) {
      buf1[i]='\0';
      
      if(strncmp(buf1,"FIREWALL_FORWARD=",17))
	fprintf(fd,"%s\n",buf1);
      else
	fprintf(fd,"FIREWALL_FORWARD=\'%s\'\n",part);
      i = 0;
    } else if (!isspace(c)) {
      if (i < MAX_LINE_LENGTH)
	buf1[i++] = c;
    }
  }
  fclose(fp);  
  fclose(fd);

  if ((fd = fopen(RCT_CONF_FILE, "w")) == NULL) {    
    return -1;
  }    
  if ((fp = fopen(SYNC_TMP_FILE, "r")) == NULL) {    
    return -1;
  }
  while (!feof(fp)) {
    c = fgetc(fp);
    if(c > 0 && c < 128) 
      fputc(c,fd);
  }

  fclose(fp);
  fclose(fd);
  unlink(SYNC_TMP_FILE);

  return(0);
}

int GetExternalIPAddress(struct Upnp_Action_Request *ca_event)
{
  char result_str[300];
  char resultcode[80];
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  pthread_mutex_lock(&SohoDevMutex);

  ca_event->ErrCode = UPNP_E_SUCCESS;

  sprintf(resultcode,"<NewExternalIPAddress>%s</NewExternalIPAddress>",
	  soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[7]);

  sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">%s</u:%sResponse>", 
	  ca_event->ActionName,
	  SohoServiceType[SOHO_SERVICE_IPC],  
	  resultcode,
	  ca_event->ActionName);

 // printf("%s\n",result_str);

  ca_event->ActionResult = UpnpParse_Buffer(result_str);

  pthread_mutex_unlock(&SohoDevMutex); 

  return(1);
}


int ForceTermination(struct Upnp_Action_Request *ca_event)
{
  char result_str[200];
  int pid,ret=0,i=0;
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  pthread_mutex_lock(&SohoDevMutex);

  ca_event->ErrCode = UPNP_E_SUCCESS;

  soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[5][0]='0';
  soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[6][0]='0';  

  strcpy(soho_service_table[SOHO_SERVICE_CIC].VariableStrVal[3],"Down");
  soho_service_table[SOHO_SERVICE_CIC].VariableStrVal[3][4]='\0';

  strcpy(soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[2],"Disconnected");
  soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[2][12]='\0';
  
  sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\"></u:%sResponse>", 
	  ca_event->ActionName,
	  SohoServiceType[SOHO_SERVICE_IPC],  
	  ca_event->ActionName);

 // printf("%s\n",result_str);
  ca_event->ActionResult = UpnpParse_Buffer(result_str);

  UpnpNotify(device_handle,
	     soho_service_table[SOHO_SERVICE_IPC].UDN,
	     soho_service_table[SOHO_SERVICE_IPC].ServiceId,
	     soho_service_table[SOHO_SERVICE_IPC].VariableName,
	     soho_service_table[SOHO_SERVICE_IPC].VariableStrVal,
	     SOHO_IPC_VARCOUNT);
  
  pid = fork();
  if (pid == 0) {
    if (fork()==0) {
      setsid();
      for(i=3;i<256;i++)
	close(i);
      ret = execl("/etc/init.d/network","network","stopwan", NULL); 
      _exit(ret);
    } else {
      _exit(ret);
    }
  } else if (pid>0) {
#ifdef WAITPID_WAS_OK
    waitpid(-1,&ret,0);
#endif
  }
   
  pthread_mutex_unlock(&SohoDevMutex); 

  return(1);
}

int RequestConnection(struct Upnp_Action_Request *ca_event)
{
  char result_str[200];
  int ret;
  int pid;
  int i;
  	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  pthread_mutex_lock(&SohoDevMutex);


/*   ca_event->ErrCode = UPNP_E_SUCCESS; */
/*   soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[5][0]='1'; */
/*   soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[6][0]='1';   */

/*   strcpy(soho_service_table[SOHO_SERVICE_CIC].VariableStrVal[3],"Up"); */
/*   soho_service_table[SOHO_SERVICE_CIC].VariableStrVal[3][2]='\0'; */

/*   strcpy(soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[2],"Connected"); */
/*   soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[2][9]='\0'; */

/*   sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\"></u:%sResponse>",  */
/* 	  ca_event->ActionName, */
/* 	  SohoServiceType[SOHO_SERVICE_IPC],   */
/* 	  ca_event->ActionName); */

/*   ca_event->ActionResult = UpnpParse_Buffer(result_str); */

/*   UpnpNotify(device_handle, */
/* 	     soho_service_table[SOHO_SERVICE_IPC].UDN, */
/* 	     soho_service_table[SOHO_SERVICE_IPC].ServiceId, */
/* 	     soho_service_table[SOHO_SERVICE_IPC].VariableName, */
/* 	     soho_service_table[SOHO_SERVICE_IPC].VariableStrVal, */
/* 	     SOHO_IPC_VARCOUNT); */

  ca_event->ErrCode = UPNP_E_SUCCESS;

  soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[5][0]='0';
  soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[6][0]='0';  

  strcpy(soho_service_table[SOHO_SERVICE_CIC].VariableStrVal[3],"Down");
  soho_service_table[SOHO_SERVICE_CIC].VariableStrVal[3][4]='\0';

  strcpy(soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[2],"Disconnected");
  soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[2][12]='\0';
  
  sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\"></u:%sResponse>", 
	  ca_event->ActionName,
	  SohoServiceType[SOHO_SERVICE_IPC],  
	  ca_event->ActionName);

  ca_event->ActionResult = UpnpParse_Buffer(result_str);

  pid = fork();
  if (pid == 0) {
    if (fork()==0) {
      setsid();
      for(i=3;i<256;i++)
	close(i);
      ret = execl("/etc/init.d/network","network","startwan",NULL);  
      _exit(ret);
    } else {
      _exit(ret);
    }
  } else if (pid>0) {
#ifdef WAITPID_WAS_OK
    waitpid(-1,&ret,WNOHANG);
#endif
    // printf("Parent %d of pid %d continues...\n",getpid(),pid);
    UpnpUnRegisterRootDevice(device_handle);
    UpnpFinish();
  }

  pthread_mutex_unlock(&SohoDevMutex); 
  
  return(1);
}

#ifdef CONSOLE_CMD
void * SohoDeviceCommandLoop(void *args)
{
  int stoploop=0;
  char cmdline[100];
  char cmd[100];
  int i=0;
  sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  for(i=0;i<atoi(soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[8]); i++) {
	sp_upnp_sohoTrace("\trec[%d]=%s:%d:%s:%s:%d (%s)\n",
	       i,
	       soho_portmapping_table[i].description,
	       soho_portmapping_table[i].external_port,
	       soho_portmapping_table[i].protocol,
	       soho_portmapping_table[i].internal_client,
	       soho_portmapping_table[i].internal_port,
	       soho_portmapping_table[i].enabled?"enabled":"disabled");
      }
      
  while (!stoploop) {
    sprintf(cmdline, " ");
    sprintf(cmd, " ");
//sp    
    //printf("\n>> ");

    // Get a command line
    fgets(cmdline, 100, stdin);

    sscanf(cmdline, "%s", cmd);

    if (strcasecmp(cmd, "exit") == 0) {
      printf("Shutting down...\n");
      UpnpUnRegisterRootDevice(device_handle);
      UpnpFinish();
      exit(0);
    } else if (strcasecmp(cmd, "list") == 0) {
      printf("   Current Value List: \n");
      printf("   PortMapping: %s\n",soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[8]);
      for(i=0;i<atoi(soho_service_table[SOHO_SERVICE_IPC].VariableStrVal[8]); i++) {
	printf("\trec[%d]=%s:%d:%s:%s:%d (%s)\n",
	       i,
	       soho_portmapping_table[i].description,
	       soho_portmapping_table[i].external_port,
	       soho_portmapping_table[i].protocol,
	       soho_portmapping_table[i].internal_client,
	       soho_portmapping_table[i].internal_port,
	       soho_portmapping_table[i].enabled?"enabled":"disabled");
      }
    } else {
//sp/
/*    	
      printf("\n   Unknown command: %s\n\n", cmd);
      printf("   Valid Commands:\n");
      printf("     List\n");
      printf("     Exit\n\n");
*/     
    }
  }
  return NULL;
}
#endif

void SohoDeviceUsage(void)
{
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  fprintf(stderr,"Usage: upnp_soho [-p portnumber]\n");
  fprintf(stderr,"The portnumber must larger than 1024. If the -p option is not specified, the default portnumber 8091 will be used.\n\n");
  fflush(stderr);
  return; 
}

int main(int argc, char** argv)
{
  int ret = 1;
  int port = -1;
  char desc_doc_url[200];
  int sig;
  sigset_t sigs_to_catch;  
  struct stat sb;
  FILE *fp;

#ifdef CONSOLE_CMD
  pthread_t cmdloop_thread;
  int code;
#endif
	sp_upnp_sohoTrace("%s: %s..:\n", __FILE__, __func__);
  if(argc == 1){
    port = UPNP_SOHO_DEFAULT_PORT;
  } else if((argc == 3) && ((strcmp( argv[1], "-p" ) == 0) 
			    && (strlen(argv[2]) == strspn(argv[2], "0123456789")))) {
    port = atoi(argv[2]);
//sp    
    
    if((port < 1025) || (port>65535)) {
      SohoDeviceUsage();
      exit(1);	
    }
    
  } else {    
    SohoDeviceUsage();
    exit(1);
  }
  
  SohoDeviceConfigIP2();  

  if (lanip==NULL) {
    fprintf(stderr,"Unable to figure out LAN interface IP. Check if you have the %s file and the IP2 value is valid.\n", UPNP_SOHO_IP2_FILE);
    fflush(stderr);
    exit(1);
  }

  if ( stat(UPNP_SOHO_PID_FILE, &sb ) >=0 ) {
    if (S_ISDIR( sb.st_mode ) ) {       
      fprintf(stderr,"The %s is not a file.\n",UPNP_SOHO_PID_FILE);
      fflush(stderr);
      exit(1);
    } else {
      unlink(UPNP_SOHO_PID_FILE);
      sp_upnp_sohoTrace("%s: %s..:unlink(UPNP_SOHO_PID_FILE);\n", __FILE__, __func__);
    }
  }

  if ((fp = fopen(UPNP_SOHO_PID_FILE, "w")) == NULL) {
    fprintf(stderr,"The %s is not a file.\n",UPNP_SOHO_PID_FILE);
    fflush(stderr);
    exit(1);
  }  
  fprintf(fp,"%d",getpid());
  fclose(fp);
//sp
  sprintf(desc_doc_url, "http://%s:%d/descrip.xml", lanip, port);
  

  SohoDescriptionFileParse(port);
#ifdef CONSOLE_CMD    
  printf("Intializing UPnP \n\twith desc_doc_url=%s\n\t", desc_doc_url);
  printf("\tipaddress=%s port=%d\n", lanip, port);
#endif
  if ((ret = UpnpInit(lanip, port)) != UPNP_E_SUCCESS) {
    fprintf(stderr,"Error with UpnpInit -- %s\n", UpnpGetErrorMessage(ret));
    UpnpFinish();
    fflush(stderr);
    exit(1);
  }
#ifdef CONSOLE_CMD
  printf("UPnP Initialized\n");
  printf("Specifying the webserver root directory -- /etc/xml/\n");
#endif
  //if ((ret = UpnpSetWebServerRootDir("/upnp_soho/xml/")) != UPNP_E_SUCCESS) {
  if ((ret = UpnpSetWebServerRootDir("/etc/xml/")) != UPNP_E_SUCCESS) {
    fprintf(stderr,"Error: %s\n", UpnpGetErrorMessage(ret));
    UpnpFinish();
    exit(1);
  }
#ifdef CONSOLE_CMD  
  printf("Registering the RootDevice\n");
#endif
  if ((ret = UpnpRegisterRootDevice(desc_doc_url, SohoDeviceCallbackEventHandler, 
				    &device_handle, &device_handle)) != UPNP_E_SUCCESS) {
    fprintf(stderr,"Error: %s\n", UpnpGetErrorMessage(ret));
    fflush(stderr);
    UpnpFinish();
    exit(1);
  } else {
#ifdef CONSOLE_CMD
    printf("RootDevice Registered\n");
    printf("Initializing State Table\n");
#endif
    if ((ret=SohoDeviceStateTableInit(desc_doc_url))!=UPNP_E_SUCCESS) {
      fprintf(stderr,"Error: Device Initialization fail(ret=%d)\n",ret);
      fflush(stderr);    
      UpnpFinish();
      exit(1);
    }
      
#ifdef CONSOLE_CMD
    printf("State Table Initialized\n");
    printf("Sending Advertisements\n");
#endif
    if ((ret = UpnpSendAdvertisement(device_handle, default_advr_expire)) != UPNP_E_SUCCESS) {
      fprintf(stderr,"Error: %s\n", UpnpGetErrorMessage(ret));
      fflush(stderr);    
      UpnpFinish();
      exit(1);
    }

#ifdef CONSOLE_CMD
    printf("Advertisements Sent\n");
#endif
  }

#ifdef CONSOLE_CMD
  /* start a command loop thread */
  code = pthread_create( &cmdloop_thread, NULL, SohoDeviceCommandLoop, NULL );
#endif
  
  /* Catch Ctrl-C and properly shutdown Or Catch Signal 1 To Re-initialize */
  sigemptyset(&sigs_to_catch);
  sigaddset(&sigs_to_catch, 1);
  sigaddset(&sigs_to_catch, SIGINT);
  sigaddset(&sigs_to_catch, SIGTERM);
  //  sigaddset(&sigs_to_catch, SIGTSTP);
  //  sigaddset(&sigs_to_catch, SIGHUP);
  
  for(;;) {
    sigwait(&sigs_to_catch, &sig);
    if(sig == 1) {
      pthread_mutex_lock(&SohoDevMutex); 
      SohoDeviceStateTableValueInit();
      UpnpNotify(device_handle,
		 soho_service_table[SOHO_SERVICE_IPC].UDN,
		 soho_service_table[SOHO_SERVICE_IPC].ServiceId,
		 soho_service_table[SOHO_SERVICE_IPC].VariableName,
		 soho_service_table[SOHO_SERVICE_IPC].VariableStrVal,
		 SOHO_IPC_VARCOUNT);      
      pthread_mutex_unlock(&SohoDevMutex); 
    } else {
      break;
    }
  }

  fprintf(stderr,"Shutting down on signal %d...\n", sig);
  fflush(stderr);  
  UpnpUnRegisterRootDevice(device_handle);
  UpnpFinish();
  
  exit(0);
}

