/**********************************************************************************  
 *  SecureSOHO v2 UPnP device by using Intel UPnP SDK v1.0.4
 *
 *  Revision 0.4.0
 *
 *  Copyright (c) 2001, 2002 REDSonic Inc. All rights reserved. 
 **********************************************************************************/

#ifndef UPNP_SOHO_H
#define UPNP_SOHO_H

#include <stdio.h>
#include <signal.h>
#include <pthread.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/wait.h>
//#include <sys/file.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include "./intel_upnpsdk-1.0.4/inc/upnp.h"
#include "common_util.h"
#define UPNP_SOHO_DEFAULT_PORT   8091


#ifndef M_NH200
#define M_NH200
#endif

//#define UPNP_SOHO_IP2_FILE "/etc/sysconfig/ip"
#ifdef M_NH200
#define UPNP_SOHO_IP2_FILE "/etc/config"
#define UPNP_SOHO_IP2_ITEM "IP"

#else
#define UPNP_SOHO_IP2_FILE "/var/conf/config"
#define UPNP_SOHO_IP2_ITEM "LAN_IP"

#endif

//#define UPNP_SOHO_CONF_FILE "/conf/portmapping.conf"
#define UPNP_SOHO_CONF_FILE "/var/conf/portmapping.conf"
#define UPNP_SOHO_PID_FILE "/var/run/upnp_soho.pid"
//#define RCT_CONF_FILE "/conf/config"
#define RCT_CONF_FILE "/var/conf/config"
#define SYNC_TMP_FILE "/tmp/pm.tmp"

#define SOHO_SERVICE_SERVCOUNT 3

#define SOHO_SERVICE_L3F  0 
#define SOHO_SERVICE_CIC  1
#define SOHO_SERVICE_IPC  2

#define SOHO_DEVICE_IGD   0 
#define SOHO_DEVICE_WAN   1
#define SOHO_DEVICE_WCD   2

#define SOHO_L3F_VARCOUNT 1
#define SOHO_CIC_VARCOUNT 11
#define SOHO_IPC_VARCOUNT 10  
#define SOHO_MAXVARS  SOHO_CIC_VARCOUNT

#define SOHO_MAX_VAL_LEN 64  
#define MAX_PORT_MAPPING_NUM 128
#define MAX_LINE_LENGTH 40959
#define TOTAL_BYTES_SENT 0
#define TOTAL_BYTES_RECEIVED 1
#define TOTAL_PACKETS_SENT 2
#define TOTAL_PACKETS_RECEIVED 3


extern UpnpDevice_Handle device_handle;
extern char *SohoDeviceType[];
extern char *SohoServiceType[];
extern pthread_mutex_t SohoDevMutex;

struct SohoService {
  char UDN[NAME_SIZE]; 
  char ServiceId[NAME_SIZE];
  char ServiceType[NAME_SIZE];
  char *VariableName[SOHO_MAXVARS];
  char *VariableStrVal[SOHO_MAXVARS];
  int  VariableCount;
};                                    

struct SohoPortMapping {
  int enabled;
  long lease_duration;
  /*  char remote_host[32]; */
  int external_port;
  int internal_port;
  char protocol[6];
  char internal_client[32];
  char description[32];
  int index;
};

void SohoDeviceUsage(void);
int SohoDeviceConfigIP2(void);
int SohoDeviceUpdateEnable(void);
int SohoDescriptionFileParse(int);
char* SohoParsePacketByte(int);

#ifdef CONSOLE_CMD
void* SohoDeviceCommandLoop(void *args);
#endif

int SohoDeviceStateTableInit(char*); 
int SohoDeviceStateTableValueInit();
int SohoDevicePortMappingInit(const char*, int*);

int SohoDeviceCallbackEventHandler(Upnp_EventType, void*, void*);
int SohoDeviceHandleSubscriptionRequest(struct Upnp_Subscription_Request *);
int SohoDeviceHandleGetVarRequest(struct Upnp_State_Var_Request *);
int SohoDeviceHandleActionRequest(struct Upnp_Action_Request *);

/* 
   //  funtions: actions defined in L3Frwd.xml
   int SetDefaultConnectionService();
   int GetDefaultConnectionService();
   //  funtions: actions defined in WANIPCn.xml   
   int SetConnectionType();
   
*/

int GetCommonLinkProperties(struct Upnp_Action_Request *);
int GetConnectionTypeInfo(struct Upnp_Action_Request *);
int GetNATRSIPStatus(struct Upnp_Action_Request *);
int GetPacketByteStat(int, struct Upnp_Action_Request *);
int GetStatusInfo(struct Upnp_Action_Request *);
int ForceTermination(struct Upnp_Action_Request *);
int RequestConnection(struct Upnp_Action_Request *);

int GetGenericPortMappingEntry(struct Upnp_Action_Request *);
int GetSpecificPortMappingEntry(struct Upnp_Action_Request *);
int AddPortMapping(struct Upnp_Action_Request *);
int DeletePortMapping(struct Upnp_Action_Request *);
int GetExternalIPAddress(struct Upnp_Action_Request *);

int PortMappingDump();
int PortMappingSync();

#endif
