/* httpd.c:  A very simple http server
 *
 * Copyright (c) 1997-1999 D. Jeff Dionne <jeff@rt-control.com>
 * Copyright (c) 1998      Kenneth Albanowski <kjahds@kjahds.com>
 * Copyright (c) 1999      Nick Brok <nick@nbrok.iaehv.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

//#define HTTPD_DEBUG
#if      defined(HTTPD_DEBUG)
# define  httpdTrace(fmt,arg...)                 printf(fmt,##arg)
# define  httpdAssert(cond)                      assert(cond)
	
#else
# define  httpdTrace(format, args...)
# define  httpdAssert(cond)
#endif


#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <sys/stat.h>
#include <dirent.h>
#include <signal.h>
#include <unistd.h>
#include <ctype.h>
#include <time.h>

#ifndef SOCK_STREAM
#define SOCK_STREAM       2
#endif

extern int copy(FILE *, FILE *);
int ReadConfiguration();
void unescape(char *string);


#undef DEBUG

int TIMEOUT = 30;

#ifndef O_BINARY
#define O_BINARY 0
#endif

char referrer[128];
int content_length;
char FOLDER[100];

#define SERVER_PORT 80

int PrintHeader(FILE * f, int content_type)
{
    alarm(TIMEOUT);
    fprintf(f, "HTTP/1.0 200 OK\n");
    switch (content_type) {
    case 't':
	fprintf(f, "Content-type: text/plain\n");
	break;
    case 'g':
	fprintf(f, "Content-type: image/gif\n");
	break;
    case 'j':
	fprintf(f, "Content-type: image/jpeg\n");
	break;
    case 'h':
	fprintf(f, "Content-type: text/html\n");
	break;
    }
    fprintf(f, "Server: uClinux-httpd 0.2.2\n");
    fprintf(f, "Expires: 0\n");
    fprintf(f, "\n");
    alarm(0);
    return (0);
}

int DoJpeg(FILE * f, char *name)
{
/*   char *buf; */
    FILE *infile;
/*   int count; */

    if (!(infile = fopen(name, "r"))) {
	alarm(TIMEOUT);
	fprintf(stderr, "Unable to open JPEG file %s, %d\n", name, errno);
	fflush(f);
	alarm(0);
	return -1;
    }

    PrintHeader(f, 'j');

    copy(infile, f);		/* prints the page */

    alarm(TIMEOUT);
    fclose(infile);
    alarm(0);

    return 0;
}

int DoGif(FILE * f, char *name)
{
/*   char *buf; */
    FILE *infile;
/*   int count; */

    if (!(infile = fopen(name, "r"))) {
	alarm(TIMEOUT);
	fprintf(stderr, "Unable to open GIF file %s, %d\n", name, errno);
	fflush(f);
	alarm(0);
	return -1;
    }

    PrintHeader(f, 'g');

    copy(infile, f);		/* prints the page */

    alarm(TIMEOUT);
    fclose(infile);
    alarm(0);

    return 0;
}

//=========================================================================

//=========================================================================
int DoDir(FILE * f, char *name)
{
	
//+++++++++++++++++++++++++++++++++++++
		struct stat sb;
		char *timestr;
		time_t now;
		char *PATH;
		int count;
//+++++++++++++++++++++++++++++++++++++
	
/*   char *buf; */
    DIR *dir;
    struct dirent *dirent;
	
	char *OldPath,*NewPath,root_folder[50];
	FILE *fp;
	OldPath = getcwd (0, 1024);
	chdir(name);
	NewPath = getcwd (0, 1024);
	//printf("the oldpath=%s, the new path is %s, FOLDER=%s\n",OldPath,NewPath,FOLDER);
	chdir(OldPath);
	
	if(!strstr(NewPath,OldPath))
	{
		fprintf(f,"<html><head><title>404 File Not Found</title></head>\n");
		fprintf(f,"<body>The requested URL was not found on this server</body></html>\n");
		return -1;
	}
	
    if ((dir = opendir(name)) == 0) {
	fprintf(stderr, "Unable to open directory %s, %d\n", name, errno);
	fflush(f);
	return -1;
    }

    PrintHeader(f, 'h');

    alarm(TIMEOUT);
    fprintf(f, "<html><head><title></title>\n");
	#if !defined(PCI) && !defined(LONGSHINE)
	    fprintf(f, "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\"></head>\n");     
	#endif
    fprintf(f, "<PRE><H1>Index of %s</H1>\n", name);
    alarm(0);

    if (name[strlen(name) - 1] != '/') {
	strcat(name, "/");
    }

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//	Read Dir
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    while ((dirent = readdir(dir)) != NULL) {
		
//=====================================================
		if(strcmp(dirent->d_name,".") == 0)  // Don't show '.'
		{	
			continue;
		}
		
		if(strcmp(dirent->d_name,"..") == 0)
		{
			if(strcmp(name,"./") == 0)  // You has no up directory !!
			{
				fprintf(f, "<FONT color=\"green\"><B> &laquo </B></FONT>\n");
			}else
			{
				fprintf(f, "<FONT color=\"green\"><B> &laquo </B></FONT><a href=\"/%s%s\">%s</a>\n", name,
					dirent->d_name, "Parent Directory");
			}	
			
			fprintf(f,"<br><HR>");

			continue;
		}
		
//===============================================
	
		count = sizeof(name)+sizeof(dirent->d_name)+5;
		PATH=malloc(count);
		sprintf(PATH,"%s%s",name,dirent->d_name);
//		fprintf(f,"(%d)%s",count,PATH);

		//---------------------------------------
		
			if( stat(PATH,&sb) != 0) 
			{
				fprintf(f,"stat error !! \n");
			}

//-------------------------------------------------------
			if(sb.st_mode == 16877)  // You are a folder !!
				fprintf(f,"<FONT color=\"#FF8040\"><B> &#43   </B></FONT>");	
			else  // You are a file !!
				fprintf(f,"<FONT color=\"wingdings\"><B> &curren   </B></FONT>");	
				
//-------------------------------------------------------
			if(strcmp(name,"./") == 0){	   //patch for IE
				fprintf(f, "<a href=\"/%s\">%-20.50s</a>  ",
				dirent->d_name, dirent->d_name);
			}else{
				fprintf(f, "<a href=\"/%s%s\">%-20.50s</a>  ", name,
				dirent->d_name, dirent->d_name);
			}
					
//-------------------------------------------------------				
			//timestr = malloc(50);
	
			 /* Get time string. */
			now = time((time_t *) 0);
			
			fprintf(f,"%10ld  ",sb.st_size);   //ok
//			fprintf(f,"sb.st_mtime (%ld) \n",sb.st_mtime); //ok
			
//			fprintf(f,"ctime (%s) \n",(char *)ctime((time_t *)&sb.st_mtime));
			timestr = ctime((time_t *)&sb.st_mtime);
		
			timestr[0] = timestr[4];
			timestr[1] = timestr[5];
			timestr[2] = timestr[6];
			timestr[3] = ' ';
			timestr[4] = timestr[8];
			timestr[5] = timestr[9];
			timestr[6] = ' ';
			if (now - sb.st_mtime > 60 * 60 * 24 * 182) {	/* 1/2 year */
			  timestr[7] = ' ';
			  timestr[8] = timestr[20];
			  timestr[9] = timestr[21];
			  timestr[10] = timestr[22];
			  timestr[11] = timestr[23];
			} else {
			  timestr[7] = timestr[11];
			  timestr[8] = timestr[12];
			  timestr[9] = ':';
			  timestr[10] = timestr[14];
			  timestr[11] = timestr[15];
			}
			  timestr[12] = '\0';
		
			
			fprintf(f,"%-30.50s",timestr);
	
			
		
			
			free(PATH);		
			free(timestr);
			
//===============================================
		
//===============================================			
		alarm(TIMEOUT);
	
//		if(strcmp(name,"./") == 0){	   //patch for IE
//			fprintf(f, "<p><a href=\"/%s\">%s</a></p>\n",
//			dirent->d_name, dirent->d_name);
//		}else{
//			fprintf(f, "<p><a href=\"/%s%s\">%s</a></p>\n", name,
//			dirent->d_name, dirent->d_name);
//		}
		
		alarm(0);
		
		fprintf(f,"<BR>");
    }

	fprintf(f,"\n<HR>\n\n</PRE></html>");
	
    closedir(dir);
    return 0;
}

int DoHTML(FILE * f, char *name)
{
/*   char *buf; */
    FILE *infile;
/*   int count; */
/*   char * dir = 0; */
	httpdTrace("%s: %s..:\n", __FILE__, __func__);

	
    if (!(infile = fopen(name, "r"))) {
    	httpdTrace("%s: %s..:if (!(infile = fopen(name, \"r\")))\n", __FILE__, __func__);
    	
		alarm(TIMEOUT);
		fprintf(stderr, "Unable to open HTML file %s, %d\n", name, errno);
		fflush(f);
		alarm(0);
		return -1;
    }
    
	httpdTrace("%s: %s..:after\n", __FILE__, __func__);
	
    PrintHeader(f, 'h');
    copy(infile, f);		/* prints the page */

    alarm(TIMEOUT);
    fclose(infile);
    alarm(0);

    return 0;
}

int DoText(FILE * f, char *name)
{
/*   char *buf; */
    FILE *infile;
/*   int count; */

    if (!(infile = fopen(name, "r"))) {
	alarm(TIMEOUT);
	fprintf(stderr, "Unable to open text file %s, %d\n", name, errno);
	fflush(f);
	alarm(0);
	return -1;
    }

    PrintHeader(f, 't');
    copy(infile, f);		/* prints the page */

    alarm(TIMEOUT);
    fclose(infile);
    alarm(0);

    return 0;
}

/* Given a www-form encoded string, restore the original: */
void unescape(char *string)
{
    char *src = string, *dest = string;
    int digit1, digit2;

    while (*src!='\0'){
        switch (*src){
            case '+':
                *dest = ' ';
                break;
            case '%':
                src ++;
                if (*src >= '0' && *src <= '9') digit1 = *src - '0';
                else if (*src >= 'A' && *src <= 'F') digit1 = (*src - 'A') + 0xA;
                else if (*src >= 'a' && *src <= 'f') digit1 = (*src - 'a') + 0xA;
                else digit1 = 0;
                src ++;
                if (*src >= '0' && *src <= '9') digit2 = *src - '0';
                else if (*src >= 'A' && *src <= 'F') digit2 = (*src - 'A') + 0xA;
                else if (*src >= 'a' && *src <= 'f') digit2 = (*src - 'a') + 0xA;
                else digit2 = 0;
                *dest = (digit1 * 0x10) + digit2;
                break;
            default:
                *dest = *src;
        }
        src ++;
        dest ++;
    }
    *dest = '\0';
    return;
}

int ParseReq(FILE * f, char *r)
{
    char *bp;
    struct stat stbuf;
    char *arg;
    char *c;
    char directory[100];
/*   int e; */
/*   int raw; */

#ifdef DEBUG
    printf("ParseReq: req is '%s'\n", r);
#endif

    while (*(++r) != ' ');	/*skip non-white space */
    while (isspace(*r))
	r++;

    while (*r == '/')
	r++;

    bp = r;
  /**r = '.';*/

    while (*r && (*(r) != ' ') && (*(r) != '?'))
	r++;

#ifdef DEBUG
    printf("ParseReq: bp='%s', r='%s'\n", bp, r);
#endif

    if (*r == '?') {
	char *e;
	*r = 0;
	arg = r + 1;
	if ((e = strchr(arg, ' ')) != NULL) {
	    *e = '\0';
	}
    } else {
	arg = 0;
	*r = 0;
    }

    c = bp;

    if (c[0] == '\0')
	    strcat(c, ".");
	//strcat(c, "/mnt/");
	
	    //strcat(c,"/");
	ReadConfiguration();
	strcpy(directory,"/mnt/");
	strcat(directory,FOLDER);
	//printf("directory = %s\n",directory);
	chdir(directory);


    unescape(c);
	
	httpdTrace("%s: %s..:c=%s\n", __FILE__, __func__, c);
#ifdef DEBUG
    printf("%s: %s..:c=%s\n", __FILE__, __func__, c);
#endif
	
    if (c && !stat(c, &stbuf) && !strstr(c,"/etc") && !strstr(c,"..")) {
	if (S_ISDIR(stbuf.st_mode)) {
	    char *end = c + strlen(c);
	    //ReadConfiguration();
	    //strcat(c,"/");
	    //strcat(c,FOLDER);
	    strcat(c, "/index.html");
	    if (!stat(c, &stbuf)) {
		    DoHTML(f, c);
	    } else {
		    *end = '\0';
		    DoDir(f, c);
	    }
	} else if (!strcmp(r - 4, ".gif"))
	    DoGif(f, c);
	else if (!strcasecmp(r - 4, ".jpg") || !strcmp(r - 5, ".jpeg"))
	    DoJpeg(f, c);
	else if (!strcmp(r - 4, ".htm") || !strcmp(r - 5, ".html"))
	    DoHTML(f, c);
	else
	    DoText(f, c);
    } else {
    	
  #ifdef DEBUG
    printf("%s: %s..:else\n", __FILE__, __func__);
#endif  	
//sp's test code for open a chinese file; since stat() fails to recognize it's a file/dir.
/*    	
    	if (!strcmp(r - 4, ".htm") || !strcmp(r - 5, ".html")){
    		httpdTrace("%s: %s..:if..c=%s sizeof(c)=%d\n", __FILE__, __func__, c, sizeof(c));
	        DoHTML(f, c);
	        return 0;
	    }else{
	    	httpdTrace("%s: %s..:else..c=%s sizeof(c)=%d\n", __FILE__, __func__, c, sizeof(c));
	    	
	    }
*/
	    	
	PrintHeader(f, 'h');
	alarm(TIMEOUT);
	fprintf(f,
		"<html><head><title>404 File Not Found</title></head>\n");
	fprintf(f,
		"<body>The requested URL was not found on this server</body></html>\n");
	alarm(0);
    }
    return 0;
}

void sigalrm(int signo)
{
    /* got an alarm, exit & recycle */
    exit(0);
}

int HandleConnect(int fd)
{
    FILE *f;

    char buf[160];
    char buf1[160];

    f = fdopen(fd, "r+");
    if (!f) {
	fprintf(stderr, "httpd: Unable to open httpd input fd, error %d\n",
		errno);
	alarm(TIMEOUT);
	close(fd);
	alarm(0);
	return 0;
    }
    setbuf(f, 0);

    alarm(TIMEOUT);

    if (!fgets(buf, 150, f)) {
	fprintf(stderr, "httpd: Error reading connection, error %d\n",
		errno);
	fclose(f);
	alarm(0);
	return 0;
    }
#ifdef DEBUG
    printf("buf = '%s'\n", buf);
#endif

    alarm(0);

    referrer[0] = '\0';
    content_length = -1;

    alarm(TIMEOUT);
    while (fgets(buf1, 150, f) && (strlen(buf1) > 2)) {
	alarm(TIMEOUT);
#ifdef DEBUG
	printf("Got buf1 '%s'\n", buf1);
#endif
	if (!strncasecmp(buf1, "Referer:", 8)) {
	    char *c = buf1 + 8;
	    while (isspace(*c))
		c++;
	    strcpy(referrer, c);
	} else if (!strncasecmp(buf1, "Referrer:", 9)) {
	    char *c = buf1 + 9;
	    while (isspace(*c))
		c++;
	    strcpy(referrer, c);
	} else if (!strncasecmp(buf1, "Content-length:", 15)) {
	    content_length = atoi(buf1 + 15);
	}
    }
    alarm(0);


    if (ferror(f)) {
	fprintf(stderr,
		"http: Error continuing reading connection, error %d\n",
		errno);
	fclose(f);
	return 0;
    }

    ParseReq(f, buf);

    alarm(TIMEOUT);
    fflush(f);
    fclose(f);
    alarm(0);

    return 1;
}


int ReadConfiguration(void){
    
    FILE *fip;
    char LineBuf[256];
    char *ptr, *end;
    int i=0;
    
    chdir("/etc");
    if((fip = fopen("/etc/config","r")) == NULL){
    	printf("%s : Line %d\n",__func__,__LINE__);  
        return 0;
    }

    while(fgets(LineBuf,255,fip) != NULL){
        for(i=0;i<255;i++){
            if((LineBuf[i] == ' ') || (LineBuf[i] == '\n') || (LineBuf[i] == '\r')){
	            LineBuf[i] = '\0';
	            i=256;
	        }
	    }	   
	    if((ptr = strstr(LineBuf,"WEB_FOLDER='")) != NULL){
	        ptr+=12;
	        end = strchr(ptr,'\'');
	    	*end = '\0';
	        strcpy(FOLDER, ptr);
	    }	    
	}
	fclose(fip);
    return 1;
}

int main(int argc, char *argv[])
{
    int fd, s;
    int len;
    int true = 1;
    struct sockaddr_in ec;
    struct sockaddr_in server_sockaddr;

    signal(SIGCHLD, SIG_IGN);
    signal(SIGPIPE, SIG_IGN);
    signal(SIGALRM, sigalrm);    // exit(0)

    //chroot(HTTPD_DOCUMENT_ROOT);
    chdir("/");

    if (argc > 1 && !strcmp(argv[1], "-i")) {
	/* I'm running from inetd, handle the request on stdin */
	fclose(stderr);
	HandleConnect(0);
	exit(0);
    }

    if ((s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
	perror("Unable to obtain network");
	exit(1);
    }

    if ((setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (void *) &true,
		    sizeof(true))) == -1) {
	perror("setsockopt failed");
	exit(1);
    }

    server_sockaddr.sin_family = AF_INET;
    server_sockaddr.sin_port = htons(SERVER_PORT);
    server_sockaddr.sin_addr.s_addr = htonl(INADDR_ANY);

    if (bind(s, (struct sockaddr *) &server_sockaddr,
	     sizeof(server_sockaddr)) == -1) {
	perror("Unable to bind socket");
	exit(1);
    }

    if (listen(s, 8 * 3) == -1) {	/* Arbitrary, 8 files/page, 3 clients */
	perror("Unable to listen");
	exit(4);
    }

    while (1) {
	len = sizeof(ec);

	if ((fd = accept(s, (void *) &ec, &len)) == -1) {
	    exit(5);
	    close(s);
	}
	HandleConnect(fd);
    }
}

